/*
 * gui_lib.c for RR182x Linux driver
 * Copyright (c) 2002-2004 HighPoint Technologies, Inc. All rights reserved.
 */
 
#include "hptintf.h"

#if BITS_PER_LONG > 32
#define VDEV_TO_ID(pVDev) (DEVICEID)(ULONG_PTR)(pVDev)
#define ID_TO_VDEV(id) (PVDevice)(((ULONG_PTR)gIal_Adapter & 0xffffffff00000000) | (id))
#else 
#define VDEV_TO_ID(pVDev) (DEVICEID)(pVDev)
#define ID_TO_VDEV(id) (PVDevice)(id)
#endif

#define INVALID_DEVICEID		(DEVICEID)(-1)
#define INVALID_STRIPSIZE		(-1)

#define shortswap(w) ((WORD)((w)>>8 | ((w) & 0xFF)<<8))

#ifndef MinBlockSizeShift
#define MinBlockSizeShift 5
#define MaxBlockSizeShift 12
#endif

#pragma pack(1)

typedef struct _HPT_IOCTL_TRANSFER_PARAM
{
	ULONG nInBufferSize;
	ULONG nOutBufferSize;
	UCHAR buffer[0];
}HPT_IOCTL_TRANSFER_PARAM, *PHPT_IOCTL_TRANSFER_PARAM;

typedef struct _HPT_SET_STATE_PARAM
{
	DEVICEID idArray;
	DWORD state;
} HPT_SET_STATE_PARAM, *PHPT_SET_STATE_PARAM;

typedef struct _HPT_SET_ARRAY_INFO
{
	DEVICEID idArray;
	ALTERABLE_ARRAY_INFO Info;
} HPT_SET_ARRAY_INFO, *PHPT_SET_ARRAY_INFO;

typedef struct _HPT_SET_DEVICE_INFO
{
	DEVICEID idDisk;
	ALTERABLE_DEVICE_INFO Info;
} HPT_SET_DEVICE_INFO, *PHPT_SET_DEVICE_INFO;

typedef struct _HPT_SET_DEVICE_INFO_V2
{
	DEVICEID idDisk;
	ALTERABLE_DEVICE_INFO_V2 Info;
} HPT_SET_DEVICE_INFO_V2, *PHPT_SET_DEVICE_INFO_V2;

typedef struct _HPT_ADD_DISK_TO_ARRAY
{
	DEVICEID idArray;
	DEVICEID idDisk;
} HPT_ADD_DISK_TO_ARRAY, *PHPT_ADD_DISK_TO_ARRAY;

typedef struct _HPT_DEVICE_IO
{
	DEVICEID	id;
	int			cmd;
	ULONG		lba;
	DWORD		nSector;
	UCHAR		buffer[0];
} HPT_DEVICE_IO, *PHPT_DEVICE_IO;

#ifdef SUPPORT_HPT601
typedef struct _HPT_SET_601_INFO
{
	DEVICEID idDisk;
	HPT601_INFO Info;
} HPT_SET_601_INFO, *PHPT_SET_601_INFO;
#endif

#pragma pack()

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,4)
#define __BDEV_RAW
#else 
#define __BDEV_RAW , BDEV_RAW
#endif
static struct block_device * get_bdev(struct Scsi_Host *host, int id)
{
	static int major[] = { SCSI_DISK0_MAJOR, SCSI_DISK1_MAJOR, SCSI_DISK2_MAJOR, SCSI_DISK3_MAJOR, 
				SCSI_DISK4_MAJOR, SCSI_DISK5_MAJOR, SCSI_DISK6_MAJOR, SCSI_DISK7_MAJOR, 
				SCSI_DISK8_MAJOR, SCSI_DISK9_MAJOR, SCSI_DISK10_MAJOR, SCSI_DISK11_MAJOR, 
				SCSI_DISK12_MAJOR, SCSI_DISK13_MAJOR, SCSI_DISK14_MAJOR, SCSI_DISK15_MAJOR, 0 };
	Scsi_Device *SDptr;
	shost_for_each_device(SDptr, host) {
		if (SDptr->id==id) {
			scsi_device_put(SDptr);
			break;
		}
	}
	if (SDptr) {
		int i, minor;
		for (i=0; major[i]; i++) {
			for (minor=0; minor<=240; minor+=16) {
				struct block_device *bdev = bdget(MKDEV(major[i], minor));
				if (bdev && blkdev_get(bdev, FMODE_READ, 0 __BDEV_RAW)==0) {
					if (bdev->bd_disk && bdev->bd_disk->driverfs_dev==&SDptr->sdev_gendev) {
						return bdev;
					}
					blkdev_put(bdev __BDEV_RAW);
				}
			}
		}
	}
	return 0;
}
static int sd_inuse(struct Scsi_Host *host, int id)
{
	struct block_device * bdev = get_bdev(host, id);
	int inuse = 0;
	if (bdev) {
		if (bdev->bd_openers>1) inuse=1;
		blkdev_put(bdev __BDEV_RAW);
	}
	return inuse;
}

#else 

static int sd_inuse(struct Scsi_Host *host, int id)
{
	Scsi_Device *SDptr = host->host_queue;
	while (SDptr) {
		if ((SDptr->id == id) && SDptr->access_count)
			return 1;
		SDptr = SDptr->next;
	}
	return 0;
}
#endif

static void cmd_timeout(unsigned long data)
{
	PCommand pCmd = (PCommand)data;
	pCmd->Result = RETURN_SELECTION_TIMEOUT;
	up(pCmd->pOrgCommand);
}

static void timer_start_for_cmd(PCommand pCmd, int second)
{
	unsigned long timeout;
	struct timer_list timer;

	timeout = jiffies + second * HZ;
	init_timer(&timer);
	timer.expires = timeout;
	timer.data = (ULONG_PTR)pCmd;
	timer.function = cmd_timeout;
	add_timer(&timer);
	 
	if (down_interruptible(pCmd->pOrgCommand))
		down(pCmd->pOrgCommand);

	del_timer(&timer);
}

static int check_VDevice_valid(PVDevice p)
{
	int i;
	PVDevice pVDevice;
	PVBus    _vbus_p;
	IAL_ADAPTER_T *pAdapter = gIal_Adapter;
	
	while(pAdapter != 0)
	{
		for (i = 0; i < MV_SATA_CHANNELS_NUM; i++)
			if(&(pAdapter->VDevices[i]) == p)  return 0;
		pAdapter = pAdapter->next;
	}

#ifdef SUPPORT_ARRAY
	pAdapter = gIal_Adapter;
	while(pAdapter != 0)
	{
		_vbus_p = &pAdapter->VBus;
		for (i=0;i<MAX_ARRAY_PER_VBUS;i++) 
		{
			pVDevice=ArrayTables(i);
			if ((pVDevice->u.array.dArStamp != 0) && (pVDevice == p))
				return 0;
		}
		pAdapter = pAdapter->next;
	}
#endif

	return -1;
}

#ifdef SUPPORT_ARRAY

static UCHAR get_vdev_type(PVDevice pVDevice)
{
	switch (pVDevice->VDeviceType) {
		case VD_RAID_0: return AT_RAID0;
		case VD_RAID_1: return AT_RAID1;
		case VD_JBOD:   return AT_JBOD;
		case VD_RAID_5: return AT_RAID5;
		default:        return AT_UNKNOWN;
	}
}

static DWORD get_array_flag(PVDevice pVDevice)
{
	int i;
	DWORD f = 0;
	
	/* The array is disabled */
	if(!pVDevice->vf_online) {
		f |= ARRAY_FLAG_DISABLED;
		/* Ignore other info */
		return f;
	}

	/* array need synchronizing */
	if(pVDevice->u.array.rf_need_rebuild && !pVDevice->u.array.rf_duplicate_and_create)
		f |= ARRAY_FLAG_NEEDBUILDING;

	/* array is in rebuilding process */
	if(pVDevice->u.array.rf_rebuilding)
		f |= ARRAY_FLAG_REBUILDING;

	/* array is being verified */
	if(pVDevice->u.array.rf_verifying)
		f |= ARRAY_FLAG_VERIFYING;

	/* array is being initialized */
	if(pVDevice->u.array.rf_initializing)
		f |= ARRAY_FLAG_INITIALIZING;

	/* broken but may still working */
	if(pVDevice->u.array.rf_broken)
		f |= ARRAY_FLAG_BROKEN;

	/* array has a active partition */
	if(pVDevice->vf_bootable)
		f |= ARRAY_FLAG_BOOTDISK;

	/* a newly created array */
	if(pVDevice->u.array.rf_newly_created)
		f |= ARRAY_FLAG_NEWLY_CREATED;

	/* array has boot mark set */
	if(pVDevice->vf_bootmark)
		f |= ARRAY_FLAG_BOOTMARK;

	/* auto-rebuild should start */
	if(pVDevice->u.array.rf_auto_rebuild)
		f |= ARRAY_FLAG_NEED_AUTOREBUILD;
		
	for(i = 0; i < pVDevice->u.array.bArnMember; i++)
	{
		PVDevice pMember = pVDevice->u.array.pMember[i];
		if (!pMember || !pMember->vf_online || (pMember->VDeviceType==VD_SINGLE_DISK))
			continue;

		/* array need synchronizing */
		if(pMember->u.array.rf_need_rebuild && 
		   !pMember->u.array.rf_duplicate_and_create)
			f |= ARRAY_FLAG_NEEDBUILDING;

		/* array is in rebuilding process */
		if(pMember->u.array.rf_rebuilding)
			f |= ARRAY_FLAG_REBUILDING;
		
		/* array is being verified */	
		if(pMember->u.array.rf_verifying)
			f |= ARRAY_FLAG_VERIFYING;
			
		/* array is being initialized */
		if(pMember->u.array.rf_initializing)
			f |= ARRAY_FLAG_INITIALIZING;

		/* broken but may still working */
		if(pMember->u.array.rf_broken)
			f |= ARRAY_FLAG_BROKEN;

		/* a newly created array */
		if(pMember->u.array.rf_newly_created)
			f |= ARRAY_FLAG_NEWLY_CREATED;

		/* auto-rebuild should start */
		if(pMember->u.array.rf_auto_rebuild)
			f |= ARRAY_FLAG_NEED_AUTOREBUILD;
	}

	return f;
}

static DWORD calc_rebuild_progress(PVDevice pVDevice)
{
	int i;
	DWORD result = ((ULONG)(pVDevice->u.array.RebuildSectors>>11)*1000 / 
		(ULONG)(pVDevice->VDeviceCapacity>>11) * (pVDevice->u.array.bArnMember-1)) * 10;

	for(i = 0; i < pVDevice->u.array.bArnMember; i++)
	{
		PVDevice pMember = pVDevice->u.array.pMember[i];
		if (!pMember || !pMember->vf_online || (pMember->VDeviceType==VD_SINGLE_DISK))
			continue;

		/* for RAID1/0 case */
		if (pMember->u.array.rf_rebuilding || 
			pMember->u.array.rf_verifying ||
			pMember->u.array.rf_initializing)
		{
			DWORD percent = ((ULONG)(pMember->u.array.RebuildSectors>>11)*1000 / 
				(ULONG)(pMember->VDeviceCapacity>>11) * (pMember->u.array.bArnMember-1)) * 10;
			if (result==0 || result>percent)
				result = percent;
		}
	}
	
	if (result>10000) result = 10000;
	return result;
}

static void get_array_info(PVDevice pVDevice, PHPT_ARRAY_INFO pArrayInfo)
{
	int	i;

	memcpy(pArrayInfo->Name, pVDevice->u.array.ArrayName, MAX_ARRAY_NAME);
	pArrayInfo->ArrayType = get_vdev_type(pVDevice);
	pArrayInfo->BlockSizeShift = pVDevice->u.array.bArBlockSizeShift;
	pArrayInfo->RebuiltSectors = pVDevice->u.array.RebuildSectors;	
	pArrayInfo->Flags = get_array_flag(pVDevice);
	pArrayInfo->RebuildingProgress = calc_rebuild_progress(pVDevice);

	pArrayInfo->nDisk = 0;
	
	for(i = 0; i < pVDevice->u.array.bArnMember; i++)
		if(pVDevice->u.array.pMember[i] != NULL)
			pArrayInfo->Members[pArrayInfo->nDisk++] = VDEV_TO_ID(pVDevice->u.array.pMember[i]);

	for(i=pArrayInfo->nDisk; i<MAX_ARRAY_MEMBERS; i++)
		pArrayInfo->Members[i] = INVALID_DEVICEID;
}

static void get_array_info_v2(PVDevice pVDevice, PHPT_ARRAY_INFO_V2 pArrayInfo)
{
	int	i;

	memcpy(pArrayInfo->Name, pVDevice->u.array.ArrayName, MAX_ARRAYNAME_LEN);
	pArrayInfo->ArrayType = get_vdev_type(pVDevice);
	pArrayInfo->BlockSizeShift = pVDevice->u.array.bArBlockSizeShift;
	pArrayInfo->RebuiltSectors.lo32 = pVDevice->u.array.RebuildSectors;	
	pArrayInfo->RebuiltSectors.hi32 = sizeof(LBA_T)>4? (pVDevice->u.array.RebuildSectors>>32) : 0;
	pArrayInfo->Flags = get_array_flag(pVDevice);
	pArrayInfo->RebuildingProgress = calc_rebuild_progress(pVDevice);

	pArrayInfo->nDisk = 0;
	
	for(i = 0; i < pVDevice->u.array.bArnMember; i++)
		if(pVDevice->u.array.pMember[i] != NULL)
			pArrayInfo->Members[pArrayInfo->nDisk++] = VDEV_TO_ID(pVDevice->u.array.pMember[i]);

	for(i=pArrayInfo->nDisk; i<MAX_ARRAY_MEMBERS_V2; i++)
		pArrayInfo->Members[i] = INVALID_DEVICEID;
}
#endif

static int get_disk_info(PVDevice pVDevice, PDEVICE_INFO pDiskInfo)
{
	MV_SATA_ADAPTER *pSataAdapter;
	MV_SATA_CHANNEL *pSataChannel;
	IAL_ADAPTER_T   *pAdapter;
	MV_CHANNEL		*channelInfo;
	char *p;
	int i;

	/* device location */
	pSataChannel = pVDevice->u.disk.mv;
	if(pSataChannel == NULL)	return -1;	
	pDiskInfo->TargetId = 0;
	pSataAdapter = pSataChannel->mvSataAdapter;
	if(pSataAdapter == NULL)	return -1;

	pAdapter = pSataAdapter->IALData;

	pDiskInfo->PathId = pSataChannel->channelNumber;
	pDiskInfo->ControllerId = (UCHAR)pSataAdapter->adapterId;

	if (pVDevice->u.disk.df_user_mode_set)
		pDiskInfo->DeviceModeSetting = pVDevice->u.disk.bDeUserSelectMode;
	else if ((cpu_to_le16(((PIDENTIFY_DATA)pVDevice->u.disk.mv->identifyDevice)->SataCapability) & 3)==2) 
		pDiskInfo->DeviceModeSetting = 15;
	else {
		p = (char *)&((PIDENTIFY_DATA)pVDevice->u.disk.mv->identifyDevice)->ModelNumber;
		if (*(WORD*)p==le16_to_cpu(0x5354) /*'ST'*/ &&
			(*(WORD*)(p+8)==le16_to_cpu(0x4153)/*'AS'*/ || (p[8]=='A' && p[11]=='S')))
			pDiskInfo->DeviceModeSetting = 15;
		else
			pDiskInfo->DeviceModeSetting = pVDevice->u.disk.bDeModeSetting;
	}
		
	pDiskInfo->UsableMode = pVDevice->u.disk.bDeUsable_Mode;

	pDiskInfo->DeviceType = PDT_HARDDISK;

	pDiskInfo->Flags = 0x0;

	/* device is disabled */
	if(!pVDevice->u.disk.df_on_line)
		pDiskInfo->Flags |= DEVICE_FLAG_DISABLED;

	/* disk has a active partition */
	if(pVDevice->vf_bootable)
		pDiskInfo->Flags |= DEVICE_FLAG_BOOTDISK;

	/* disk has boot mark set */
	if(pVDevice->vf_bootmark)
		pDiskInfo->Flags |= DEVICE_FLAG_BOOTMARK;

	pDiskInfo->Flags |= DEVICE_FLAG_SATA;

	/* is a spare disk */
	if(pVDevice->VDeviceType == VD_SPARE)
		pDiskInfo->Flags |= DEVICE_FLAG_IS_SPARE;

	memcpy(&(pDiskInfo->IdentifyData), (pSataChannel->identifyDevice), sizeof(IDENTIFY_DATA2));
	p = (char *)&pDiskInfo->IdentifyData.ModelNumber;
	for (i = 0; i < 20; i++)
		((WORD*)p)[i] = shortswap(pSataChannel->identifyDevice[IDEN_MODEL_OFFSET+i]);
	p[39] = '\0';

	channelInfo = &pAdapter->mvChannel[pSataChannel->channelNumber];
	pDiskInfo->ReadAheadSupported = channelInfo->readAheadSupported;
	pDiskInfo->ReadAheadEnabled = channelInfo->readAheadEnabled;
	pDiskInfo->WriteCacheSupported = channelInfo->writeCacheSupported;
	pDiskInfo->WriteCacheEnabled = channelInfo->writeCacheEnabled;
	pDiskInfo->TCQSupported = (pSataChannel->identifyDevice[IDEN_SUPPORTED_COMMANDS2] & le16_to_cpu(0x2))!=0;
	pDiskInfo->TCQEnabled = pSataChannel->queuedDMA==MV_EDMA_MODE_QUEUED;
	pDiskInfo->NCQSupported = MV_SATA_GEN_2(pSataAdapter) &&
									(pSataChannel->identifyDevice[IDEN_SATA_CAPABILITIES] & le16_to_cpu(0x0100));
	pDiskInfo->NCQEnabled = pSataChannel->queuedDMA==MV_EDMA_MODE_NATIVE_QUEUING;
	return 0;
}

int hpt_get_driver_capabilities(PDRIVER_CAPABILITIES cap)
{
	ZeroMemory(cap, sizeof(DRIVER_CAPABILITIES));
	cap->dwSize = sizeof(DRIVER_CAPABILITIES);
	cap->MaximumControllers = MAX_VBUS;

	/* cap->SupportCrossControllerRAID = 0; */
	/* take care for various OSes! */
	cap->SupportCrossControllerRAID = 0;


	cap->MinimumBlockSizeShift = MinBlockSizeShift;
	cap->MaximumBlockSizeShift = MaxBlockSizeShift;
	cap->SupportDiskModeSetting = 0;
	cap->SupportSparePool = 1;		
	cap->MaximumArrayNameLength = MAX_ARRAY_NAME - 1;
	cap->SupportDedicatedSpare = 0;
	

#ifdef SUPPORT_ARRAY
	/* Stripe */
	cap->SupportedRAIDTypes[0] = AT_RAID0;
	cap->MaximumArrayMembers[0] = MAX_MEMBERS;
	/* Mirror */
	cap->SupportedRAIDTypes[1] = AT_RAID1;
	cap->MaximumArrayMembers[1] = 2;
	/* Mirror + Stripe */
	cap->SupportedRAIDTypes[2] = (AT_RAID0<<4)|AT_RAID1; /* RAID1/0 */
	cap->MaximumArrayMembers[2] = MAX_MEMBERS;
	/* Jbod */
	cap->SupportedRAIDTypes[3] = AT_JBOD;
	cap->MaximumArrayMembers[3] = MAX_MEMBERS;
	/* RAID5 */
#if SUPPORT_RAID5
	cap->SupportedRAIDTypes[4] = AT_RAID5;
	cap->MaximumArrayMembers[4] = MAX_MEMBERS;
#endif
#endif
	return 0;
}

int hpt_get_controller_count(void)
{
	IAL_ADAPTER_T    *pAdapTemp = gIal_Adapter;
	int iControllerCount = 0;
	
	while(pAdapTemp != 0)
	{		 
		iControllerCount++;
		pAdapTemp = pAdapTemp->next;
	}
	
	return iControllerCount;
}

int hpt_get_controller_info(int id, PCONTROLLER_INFO pInfo)
{
	IAL_ADAPTER_T    *pAdapTemp;
	int iControllerCount = 0;

	for (pAdapTemp = gIal_Adapter; pAdapTemp; pAdapTemp = pAdapTemp->next) {
		if (iControllerCount++==id) {
			pInfo->InterruptLevel = 0;
			pInfo->ChipType = 0;
			pInfo->ChipFlags = CHIP_SUPPORT_ULTRA_100|CHIP_SUPPORT_ULTRA_133|CHIP_SUPPORT_ULTRA_150;
			pInfo->NumBuses = 8;
			strcpy( pInfo->szVendorID, "HighPoint Technologies, Inc.");
			strcpy(pInfo->szProductID, "RocketRAID 182x SATA Controller");			
			return 0;
		}
	}
	return -1;
}


int hpt_get_channel_info(int id, int bus, PCHANNEL_INFO pInfo)
{
	IAL_ADAPTER_T    *pAdapTemp = gIal_Adapter;
	int i,iControllerCount = 0;

	while(pAdapTemp != 0)
	{
		if (iControllerCount++==id) 
			goto found;
		pAdapTemp = pAdapTemp->next;
	}
	return -1;

found:
	
	pInfo->IoPort = 0;
	pInfo->ControlPort = 0;
	
	for (i=0; i<2 ;i++)
	{
		pInfo->Devices[i] = INVALID_DEVICEID;
	}

	if (pAdapTemp->mvChannel[bus].online == MV_TRUE)
		pInfo->Devices[0] = VDEV_TO_ID(&pAdapTemp->VDevices[bus]);
	else
		pInfo->Devices[0] = INVALID_DEVICEID;

	return 0;
	

}

int hpt_get_logical_devices(DEVICEID * pIds, int nMaxCount)
{
	int count = 0;
	int	i,j;
	PVDevice pPhysical, pLogical;
	IAL_ADAPTER_T    *pAdapTemp;

	for(i = 0; i < nMaxCount; i++)
		pIds[i] = INVALID_DEVICEID;

	/* append the arrays not registered on VBus */
	for (pAdapTemp = gIal_Adapter; pAdapTemp; pAdapTemp = pAdapTemp->next) {
		for(i = 0; i < MV_SATA_CHANNELS_NUM; i++)
		{
			pPhysical = &pAdapTemp->VDevices[i];
			pLogical = pPhysical;
			
			while (pLogical->pParent) pLogical = pLogical->pParent;
			if (pLogical->VDeviceType==VD_SPARE)
				continue;
			
			for (j=0; j<count; j++)
				if (pIds[j]==VDEV_TO_ID(pLogical)) goto next;
			pIds[count++] = VDEV_TO_ID(pLogical);
			if (count>=nMaxCount) goto done;
			next:;
		}
	}

done:
	return count;
}

int hpt_get_device_info(DEVICEID id, PLOGICAL_DEVICE_INFO pInfo)
{
	PVDevice pVDevice = ID_TO_VDEV(id);

	if((id == 0) || check_VDevice_valid(pVDevice))
		return -1;

#ifdef SUPPORT_ARRAY
	if (mIsArray(pVDevice)) {
		pInfo->Type = LDT_ARRAY;
		pInfo->Capacity = pVDevice->VDeviceCapacity;
		pInfo->ParentArray = VDEV_TO_ID(pVDevice->pParent);
		get_array_info(pVDevice, &pInfo->u.array);
		return 0;
	}
#endif

	pInfo->Type = LDT_DEVICE;
	pInfo->ParentArray = pVDevice->pParent? VDEV_TO_ID(pVDevice->pParent) : INVALID_DEVICEID;
	/* report real capacity to be compatible with old arrays */
	pInfo->Capacity = pVDevice->u.disk.dDeRealCapacity;
	return get_disk_info(pVDevice, &pInfo->u.device);
}

int hpt_get_device_info_v2(DEVICEID id, PLOGICAL_DEVICE_INFO_V2 pInfo)
{
	PVDevice pVDevice = ID_TO_VDEV(id);

	if((id == 0) || check_VDevice_valid(pVDevice))
		return -1;

#ifdef SUPPORT_ARRAY
	if (mIsArray(pVDevice)) {
		pInfo->Type = LDT_ARRAY;
		pInfo->Capacity.lo32 = pVDevice->VDeviceCapacity;
		pInfo->Capacity.hi32 = sizeof(LBA_T)>4? (pVDevice->VDeviceCapacity>>32) : 0;
		pInfo->ParentArray = VDEV_TO_ID(pVDevice->pParent);
		get_array_info_v2(pVDevice, &pInfo->u.array);
		return 0;
	}
#endif

	pInfo->Type = LDT_DEVICE;
	pInfo->ParentArray = pVDevice->pParent? VDEV_TO_ID(pVDevice->pParent) : INVALID_DEVICEID;
	/* report real capacity to be compatible with old arrays */
	pInfo->Capacity.lo32 = pVDevice->u.disk.dDeRealCapacity;
	pInfo->Capacity.hi32 = 0;
	return get_disk_info(pVDevice, &pInfo->u.device);
}

#ifdef SUPPORT_ARRAY
DEVICEID hpt_create_array_v2(_VBUS_ARG PCREATE_ARRAY_PARAMS_V2 pParam)
{
	ULONG Stamp = GetStamp();
	int	i,j;
	LBA_T capacity = MAX_LBA_T;
	PVDevice pArray,pChild;
	int		Loca = -1;
	
	if (pParam->nDisk > MAX_MEMBERS)
		return INVALID_DEVICEID;
/* check in verify_vd
	for(i = 0; i < pParam->nDisk; i++)
	{
		PVDevice pVDev = ID_TO_VDEV(pParam->Members[i]);
		if (check_VDevice_valid(pVDev)) return INVALID_DEVICEID;
		if (mIsArray(pVDev)) return INVALID_DEVICEID;
		if (!pVDev->vf_online) return INVALID_DEVICEID;
		if (!_vbus_p)
			_vbus_p = pVDev->u.disk.pVBus;
		else if (_vbus_p != pVDev->u.disk.pVBus)
			return INVALID_DEVICEID;
	}
*/
	_vbus_p = (ID_TO_VDEV(pParam->Members[0]))->u.disk.pVBus;
	if (!_vbus_p) return INVALID_DEVICEID;

	mArGetArrayTable(pArray);
	if(!pArray)	return INVALID_DEVICEID;

	switch (pParam->ArrayType)
	{
		case AT_JBOD:
			pArray->VDeviceType = VD_JBOD;
			goto simple;

		case AT_RAID0:
			if((pParam->BlockSizeShift < MinBlockSizeShift) || (pParam->BlockSizeShift > MaxBlockSizeShift))
				goto error;
			pArray->VDeviceType = VD_RAID_0;
			goto simple;

		case AT_RAID5:
			if((pParam->BlockSizeShift < MinBlockSizeShift) || (pParam->BlockSizeShift > MaxBlockSizeShift))
				goto error;
			pArray->VDeviceType = VD_RAID_5;
			/* only "no build" R5 is not critical after creation. */
			if ((pParam->CreateFlags & CAF_CREATE_R5_NO_BUILD)==0)
				pArray->u.array.rf_need_rebuild = 1;
			goto simple;

		case AT_RAID1:
			if(pParam->nDisk <= 2)
			{
				pArray->VDeviceType = VD_RAID_1;
simple:
				pArray->u.array.bArnMember = pParam->nDisk;
				pArray->u.array.bArRealnMember = pParam->nDisk;
				pArray->u.array.bArBlockSizeShift = pParam->BlockSizeShift;
				pArray->u.array.bStripeWitch = (1 << pParam->BlockSizeShift);
				pArray->u.array.dArStamp = Stamp;

				pArray->u.array.rf_need_sync = 1;
				pArray->u.array.rf_newly_created = 1;

				if ((pParam->CreateFlags & CAF_CREATE_AND_DUPLICATE) && 
					(pArray->VDeviceType == VD_RAID_1))
				{
					pArray->u.array.rf_newly_created = 0; /* R1 shall still be accessible */
					pArray->u.array.rf_need_rebuild = 1;
					pArray->u.array.rf_auto_rebuild = 1;
					pArray->u.array.rf_duplicate_and_create = 1;

					for(i = 0; i < MAX_VDEVICE_PER_VBUS; i++)
						if (_vbus_p->pVDevice[i] == ID_TO_VDEV(pParam->Members[0]))
							Loca = i;
				}
				
				pArray->u.array.RebuildSectors = pArray->u.array.rf_need_rebuild? 0 : MAX_LBA_T;

				memcpy(pArray->u.array.ArrayName, pParam->ArrayName, MAX_ARRAY_NAME);

				for(i = 0; i < pParam->nDisk; i++)
				{
					pArray->u.array.pMember[i] = ID_TO_VDEV(pParam->Members[i]);
					pArray->u.array.pMember[i]->bSerialNumber = i;
					pArray->u.array.pMember[i]->pParent = pArray;

					/*wy @ don't unregister source disk for duplicate RAID1  */
					if (i ||
						pArray->VDeviceType!=VD_RAID_1 ||
						(pParam->CreateFlags & CAF_CREATE_AND_DUPLICATE)==0)
						UnregisterVDevice(pArray->u.array.pMember[i]);

					if(pArray->VDeviceType == VD_RAID_5)
						pArray->u.array.pMember[i]->vf_cache_disk = 1;
				}
			}
			else
			{
				for(i = 0; i < (pParam->nDisk / 2); i++)
				{
					mArGetArrayTable(pChild);
					pChild->VDeviceType = VD_RAID_1;

					pChild->u.array.bArnMember = 2;
					pChild->u.array.bArRealnMember = 2;
					pChild->u.array.bArBlockSizeShift = pParam->BlockSizeShift;
					pChild->u.array.bStripeWitch = (1 << pParam->BlockSizeShift);
					pChild->u.array.dArStamp = Stamp;

					pChild->u.array.rf_need_sync = 1;
					pChild->u.array.rf_newly_created = 1;

					pChild->u.array.RebuildSectors = MAX_LBA_T;	
					
					memcpy(pChild->u.array.ArrayName, pParam->ArrayName, MAX_ARRAY_NAME);

					for(j = 0; j < 2; j++)
					{
						pChild->u.array.pMember[j] = ID_TO_VDEV(pParam->Members[i*2 + j]);
						pChild->u.array.pMember[j]->bSerialNumber = j;
						pChild->u.array.pMember[j]->pParent = pChild;
						pChild->u.array.pMember[j]->pfnDeviceFailed = pfnDeviceFailed[pChild->VDeviceType];
						UnregisterVDevice(pChild->u.array.pMember[j]);
					}

					pArray->u.array.pMember[i] = pChild;

					pChild->vf_online = 1;
					pChild->bSerialNumber = i;
					pChild->pParent = pArray;
					pChild->VDeviceCapacity = MIN(pChild->u.array.pMember[0]->VDeviceCapacity,
						pChild->u.array.pMember[1]->VDeviceCapacity);

					pChild->pfnSendCommand = pfnSendCommand[pChild->VDeviceType];
					pChild->pfnDeviceFailed = pfnDeviceFailed[VD_RAID_0];
				}

				pArray->VDeviceType = VD_RAID_0;

				pArray->u.array.bArnMember = pParam->nDisk / 2;
				pArray->u.array.bArRealnMember = pParam->nDisk / 2;
				pArray->u.array.bArBlockSizeShift = pParam->BlockSizeShift;
				pArray->u.array.bStripeWitch = (1 << pParam->BlockSizeShift);
				pArray->u.array.dArStamp = Stamp;

				pArray->u.array.rf_need_sync = 1;
				pArray->u.array.rf_newly_created = 1;

				memcpy(pArray->u.array.ArrayName, pParam->ArrayName, MAX_ARRAY_NAME);
			}
			break;

		default:
			goto error;
	}

	for(i = 0; i < pArray->u.array.bArnMember; i++)
		pArray->u.array.pMember[i]->pfnDeviceFailed = pfnDeviceFailed[pArray->VDeviceType];

	if ((pParam->CreateFlags & CAF_CREATE_AND_DUPLICATE) && 
		(pArray->VDeviceType == VD_RAID_1))
	{
		pArray->vf_bootmark = pArray->u.array.pMember[0]->vf_bootmark;
		pArray->vf_bootable = pArray->u.array.pMember[0]->vf_bootable;
		pArray->u.array.pMember[0]->vf_bootable = 0;
		pArray->u.array.pMember[0]->vf_bootmark = 0;
		if (Loca>=0) {
			_vbus_p->pVDevice[Loca] = pArray;
			/* to comfort OS */
			pArray->u.array.rf_duplicate_and_created = 1;
			pArray->pVBus = _vbus_p;
		}
	}
	else {
		UCHAR TempBuffer[512];
		ZeroMemory(TempBuffer, 512);
		for(i = 0; i < pParam->nDisk; i++)
		{
			PVDevice	pDisk = ID_TO_VDEV(pParam->Members[i]);
			pDisk->vf_bootmark = pDisk->vf_bootable = 0;
			fDeReadWrite(&pDisk->u.disk, 0, IDE_COMMAND_WRITE, TempBuffer);
		}
	}

	pArray->vf_online = 1;
	pArray->pParent = NULL;

	switch(pArray->VDeviceType)
	{
		case VD_RAID_0:
			for(i = 0; i < pArray->u.array.bArnMember; i++)
				if(pArray->u.array.pMember[i]->VDeviceCapacity < capacity)
					capacity = pArray->u.array.pMember[i]->VDeviceCapacity;
			capacity &= ~(pArray->u.array.bStripeWitch - 1);
			/* shrink member capacity for RAID 1/0 */
			for(i = 0; i < pArray->u.array.bArnMember; i++)
				if (mIsArray(pArray->u.array.pMember[i]))
					pArray->u.array.pMember[i]->VDeviceCapacity = capacity;
			pArray->VDeviceCapacity = capacity * pArray->u.array.bArnMember;
			break;

		case VD_RAID_1:
			pArray->VDeviceCapacity = MIN(pArray->u.array.pMember[0]->VDeviceCapacity,
						pArray->u.array.pMember[1]->VDeviceCapacity);
			break;

		case VD_JBOD:
			for(i = 0; i < pArray->u.array.bArnMember; i++)
				pArray->VDeviceCapacity += pArray->u.array.pMember[i]->VDeviceCapacity;
			break;

		case VD_RAID_5:
			for(i = 0; i < pArray->u.array.bArnMember; i++)
				if(pArray->u.array.pMember[i]->VDeviceCapacity < capacity)
					capacity = pArray->u.array.pMember[i]->VDeviceCapacity;
			pArray->VDeviceCapacity = (capacity & ~(pArray->u.array.bStripeWitch - 1))
				* (pArray->u.array.bArnMember - 1);
			break;

		default:
			goto error;
	}

	pArray->pfnSendCommand = pfnSendCommand[pArray->VDeviceType];
	pArray->pfnDeviceFailed = fOsDiskFailed;
	SyncArrayInfo(pArray);
	/*wy ? */
	if (!pArray->u.array.rf_duplicate_and_created)
		RegisterVDevice(pArray);
	return VDEV_TO_ID(pArray);

error:
	for(i = 0; i < pArray->u.array.bArnMember; i++)
	{
		pChild = pArray->u.array.pMember[i];
		if((pChild != NULL) && (pChild->VDeviceType != VD_SINGLE_DISK))
			mArFreeArrayTable(pChild);
	}
	mArFreeArrayTable(pArray);
	return INVALID_DEVICEID;
}

DEVICEID hpt_create_array(_VBUS_ARG PCREATE_ARRAY_PARAMS pParam)
{
	CREATE_ARRAY_PARAMS_V2 param2;
	param2.ArrayType = pParam->ArrayType;
	param2.nDisk = pParam->nDisk;
	param2.BlockSizeShift = pParam->BlockSizeShift;
	param2.CreateFlags = pParam->CreateFlags;
	param2.CreateTime = pParam->CreateTime;
	memcpy(param2.ArrayName, pParam->ArrayName, sizeof(param2.ArrayName));
	memcpy(param2.Description, pParam->Description, sizeof(param2.Description));
	memcpy(param2.CreateManager, pParam->CreateManager, sizeof(param2.CreateManager));
	param2.Capacity.lo32 = param2.Capacity.hi32 = 0;
	memcpy(param2.Members, pParam->Members, sizeof(pParam->Members));
	return hpt_create_array_v2(_VBUS_P &param2);
}

int hpt_add_disk_to_array(_VBUS_ARG DEVICEID idArray, DEVICEID idDisk)
{
	int	i;

	LBA_T Capacity;
	PVDevice pArray = ID_TO_VDEV(idArray);
	PVDevice pDisk	= ID_TO_VDEV(idDisk);

	if((idArray == 0) || (idDisk == 0))	return -1;
	if(check_VDevice_valid(pArray) || check_VDevice_valid(pDisk))	return -1;
	if(!pArray->u.array.rf_broken)	return -1;

	if(pArray->VDeviceType != VD_RAID_1 && pArray->VDeviceType != VD_RAID_5)
		return -1;
	if((pDisk->VDeviceType != VD_SINGLE_DISK) && (pDisk->VDeviceType != VD_SPARE))
		return -1;

	Capacity = pArray->VDeviceCapacity;
	do_div(Capacity, (pArray->u.array.bArnMember - 1));

	if (pArray->vf_format_v2) {
		if(pDisk->u.disk.dDeRealCapacity < Capacity) return -1;
	}
	else
		if(pDisk->VDeviceCapacity < Capacity) return -1;
	
	if (pArray->pVBus!=_vbus_p) { HPT_ASSERT(0); return -1;}

	for(i = 0; i < pArray->u.array.bArnMember; i++)
		if((pArray->u.array.pMember[i] == 0) || !pArray->u.array.pMember[i]->vf_online)
		{
			if(pArray->u.array.pMember[i] != NULL)
				pArray->u.array.pMember[i]->pParent = NULL;
			pArray->u.array.pMember[i] = pDisk;
			goto find;
		}
	return -1;

find:
	UnregisterVDevice(pDisk);
	pDisk->VDeviceType = VD_SINGLE_DISK;
	pDisk->bSerialNumber = i;
	pDisk->pParent = pArray;
	if (pArray->VDeviceType==VD_RAID_5) pDisk->vf_cache_disk = 1;
	pDisk->pfnDeviceFailed = pfnDeviceFailed[pArray->VDeviceType];
	if (pArray->vf_format_v2) {
		pDisk->vf_format_v2 = 1;
		pDisk->VDeviceCapacity = pDisk->u.disk.dDeRealCapacity;
	}

	pArray->u.array.bArRealnMember++;
	if(pArray->u.array.bArnMember == pArray->u.array.bArRealnMember)
	{
		pArray->u.array.rf_need_rebuild = 1;
		pArray->u.array.RebuildSectors = 0;
		pArray->u.array.rf_auto_rebuild = 1;
		pArray->u.array.rf_broken = 0;
	}
	pArray->u.array.RebuildSectors = 0;

	/* sync the whole array */
	while (pArray->pParent) pArray = pArray->pParent;
	pArray->u.array.dArStamp = GetStamp();
	SyncArrayInfo(pArray);
	return 0;
}

int hpt_add_spare_disk(_VBUS_ARG DEVICEID idDisk)
{
	PVDevice pVDevice = ID_TO_VDEV(idDisk);
	DECLARE_BUFFER(PUCHAR, pbuffer);

	if(idDisk == 0 || check_VDevice_valid(pVDevice))	return -1;
	if (pVDevice->VDeviceType != VD_SINGLE_DISK || pVDevice->pParent)
		return -1;

	if (pVDevice->u.disk.pVBus!=_vbus_p) return -1;

	UnregisterVDevice(pVDevice);
	pVDevice->VDeviceType = VD_SPARE;
	pVDevice->vf_bootmark = 0;

	ZeroMemory((char *)pbuffer, 512);
	fDeReadWrite(&pVDevice->u.disk, 0, IDE_COMMAND_WRITE, pbuffer);
	SyncArrayInfo(pVDevice);
	return 0;
}

int hpt_remove_spare_disk(_VBUS_ARG DEVICEID idDisk)
{
	PVDevice pVDevice = ID_TO_VDEV(idDisk);

	if(idDisk == 0 || check_VDevice_valid(pVDevice))	return -1;

	if (pVDevice->u.disk.pVBus!=_vbus_p) return -1;

	pVDevice->VDeviceType = VD_SINGLE_DISK;

	SyncArrayInfo(pVDevice);
	RegisterVDevice(pVDevice);
	return 0;
}

int hpt_set_array_info(_VBUS_ARG DEVICEID idArray, PALTERABLE_ARRAY_INFO pInfo)
{
	PVDevice pVDevice = ID_TO_VDEV(idArray);

	if(idArray == 0 || check_VDevice_valid(pVDevice)) return -1;
	if (!mIsArray(pVDevice)) return -1;

	/* if the pVDevice isn't a top level, return -1; */
	if(pVDevice->pParent != NULL) return -1;

	if (pVDevice->pVBus!=_vbus_p) { HPT_ASSERT(0); return -1;}

	if (pInfo->ValidFields & AAIF_NAME) {
		memset(pVDevice->u.array.ArrayName, 0, MAX_ARRAY_NAME);
		memcpy(pVDevice->u.array.ArrayName, pInfo->Name, sizeof(pInfo->Name));
		pVDevice->u.array.rf_need_sync = 1;
	}

	if (pInfo->ValidFields & AAIF_DESCRIPTION) {
		memcpy(pVDevice->u.array.Description, pInfo->Description, sizeof(pInfo->Description));
		pVDevice->u.array.rf_need_sync = 1;
	}

	if (pVDevice->u.array.rf_need_sync)
		SyncArrayInfo(pVDevice);
	return 0;
}

int hpt_set_device_info(_VBUS_ARG DEVICEID idDisk, PALTERABLE_DEVICE_INFO pInfo)
{
	PVDevice pVDevice = ID_TO_VDEV(idDisk);

	if(idDisk == 0 || check_VDevice_valid(pVDevice)) return -1;
	if (mIsArray(pVDevice))
		return -1;

	if (pVDevice->u.disk.pVBus!=_vbus_p) return -1;

	/* TODO */
	return 0;
}

int hpt_set_device_info_v2(_VBUS_ARG DEVICEID idDisk, PALTERABLE_DEVICE_INFO_V2 pInfo)
{
	PVDevice pVDevice = ID_TO_VDEV(idDisk);
	int sync = 0;
	
	if(idDisk==0 || check_VDevice_valid(pVDevice)) return -1;
	if (mIsArray(pVDevice))
		return -1;

	if (pVDevice->u.disk.pVBus!=_vbus_p) return -1;

	if (pInfo->ValidFields & ADIF_MODE) {
		pVDevice->u.disk.bDeModeSetting = pInfo->DeviceModeSetting;
		pVDevice->u.disk.bDeUserSelectMode = pInfo->DeviceModeSetting;
		pVDevice->u.disk.df_user_mode_set = 1;
		fDeSelectMode((PDevice)&(pVDevice->u.disk), (UCHAR)pInfo->DeviceModeSetting);
		sync = 1;
	}

	if (pInfo->ValidFields & ADIF_TCQ) {
		if (fDeSetTCQ(&pVDevice->u.disk, pInfo->TCQEnabled, 0)) {
			pVDevice->u.disk.df_tcq_set = 1;
			pVDevice->u.disk.df_tcq = pInfo->TCQEnabled!=0;
			sync = 1;
		}
	}

	if (pInfo->ValidFields & ADIF_NCQ) {
		if (fDeSetNCQ(&pVDevice->u.disk, pInfo->NCQEnabled, 0)) {
			pVDevice->u.disk.df_ncq_set = 1;
			pVDevice->u.disk.df_ncq = pInfo->NCQEnabled!=0;
			sync = 1;
		}
	}

	if (pInfo->ValidFields & ADIF_WRITE_CACHE) {
		if (fDeSetWriteCache(&pVDevice->u.disk, pInfo->WriteCacheEnabled)) {
			pVDevice->u.disk.df_write_cache_set = 1;
			pVDevice->u.disk.df_write_cache = pInfo->WriteCacheEnabled!=0;
			sync = 1;
		}
	}

	if (pInfo->ValidFields & ADIF_READ_AHEAD) {
		if (fDeSetReadAhead(&pVDevice->u.disk, pInfo->ReadAheadEnabled)) {
			pVDevice->u.disk.df_read_ahead_set = 1;
			pVDevice->u.disk.df_read_ahead = pInfo->ReadAheadEnabled!=0;
			sync = 1;
		}
	}

	if (sync)
		SyncArrayInfo(pVDevice);
	return 0;
}

#endif

/* hpt_default_ioctl()
 *  This is a default implementation. The platform dependent part
 *  may reuse this function and/or use it own implementation for
 *  each ioctl function.
 */
int hpt_default_ioctl(_VBUS_ARG
						DWORD dwIoControlCode,       	/* operation control code */
						PVOID lpInBuffer,            	/* input data buffer */
						DWORD nInBufferSize,         	/* size of input data buffer */
						PVOID lpOutBuffer,           	/* output data buffer */
						DWORD nOutBufferSize,        	/* size of output data buffer */
						PDWORD lpBytesReturned      	/* byte count */
					)
{
	switch(dwIoControlCode)	{

	case HPT_IOCTL_GET_VERSION:

		if (nInBufferSize != 0) return -1;
		if (nOutBufferSize != sizeof(DWORD)) return -1;
		*((DWORD*)lpOutBuffer) = HPT_INTERFACE_VERSION;
		break;

	case HPT_IOCTL_GET_CONTROLLER_COUNT:

		if (nOutBufferSize!=sizeof(DWORD)) return -1;
		*(PDWORD)lpOutBuffer = hpt_get_controller_count();
		break;

	case HPT_IOCTL_GET_CONTROLLER_INFO:
		{
			int id;
			PCONTROLLER_INFO pInfo;
	
			if (nInBufferSize!=sizeof(DWORD)) return -1;
			if (nOutBufferSize!=sizeof(CONTROLLER_INFO)) return -1;
	
			id = *(DWORD *)lpInBuffer;
			pInfo = (PCONTROLLER_INFO)lpOutBuffer;
			if (hpt_get_controller_info(id, pInfo)!=0)
				return -1;
		}
		break;

	case HPT_IOCTL_GET_CHANNEL_INFO:
		{
			int id, bus;
			PCHANNEL_INFO pInfo;

			if (nInBufferSize!=8) return -1;
			if (nOutBufferSize!=sizeof(CHANNEL_INFO)) return -1;

			id = *(DWORD *)lpInBuffer;
			bus = ((DWORD *)lpInBuffer)[1];
			pInfo = (PCHANNEL_INFO)lpOutBuffer;

			if (hpt_get_channel_info(id, bus, pInfo)!=0)
				return -1;
		}
		break;

	case HPT_IOCTL_GET_LOGICAL_DEVICES:
		{
			DWORD nMax;
			DEVICEID *pIds;

			if (nInBufferSize!=sizeof(DWORD)) return -1;
			nMax = *(DWORD *)lpInBuffer;
			if (nOutBufferSize < sizeof(DWORD)+sizeof(DWORD)*nMax) return -1;

			pIds = ((DEVICEID *)lpOutBuffer)+1;
			*(DWORD*)lpOutBuffer = hpt_get_logical_devices(pIds, nMax);
		}
		break;

	case HPT_IOCTL_GET_DEVICE_INFO:
		{
			DEVICEID id;
			PLOGICAL_DEVICE_INFO pInfo;

			if (nInBufferSize!=sizeof(DEVICEID)) return -1;
			if (nOutBufferSize!=sizeof(LOGICAL_DEVICE_INFO)) return -1;

			id = *(DWORD *)lpInBuffer;
			if (id == INVALID_DEVICEID)	return -1;

			pInfo = (PLOGICAL_DEVICE_INFO)lpOutBuffer;
			memset(pInfo, 0, sizeof(LOGICAL_DEVICE_INFO));

			if (hpt_get_device_info(id, pInfo)!=0)
				return -1;
		}
		break;

	case HPT_IOCTL_GET_DEVICE_INFO_V2:
		{
			DEVICEID id;
			PLOGICAL_DEVICE_INFO_V2 pInfo;

			if (nInBufferSize!=sizeof(DEVICEID)) return -1;
			if (nOutBufferSize!=sizeof(LOGICAL_DEVICE_INFO_V2)) return -1;

			id = *(DWORD *)lpInBuffer;
			if (id == INVALID_DEVICEID)	return -1;

			pInfo = (PLOGICAL_DEVICE_INFO_V2)lpOutBuffer;
			memset(pInfo, 0, sizeof(LOGICAL_DEVICE_INFO_V2));

			if (hpt_get_device_info_v2(id, pInfo)!=0)
				return -1;
		}
		break;

#ifdef SUPPORT_ARRAY
	case HPT_IOCTL_CREATE_ARRAY:
		{
			if (nInBufferSize!=sizeof(CREATE_ARRAY_PARAMS)) return -1;
			if (nOutBufferSize!=sizeof(DEVICEID)) return -1;

			*(DEVICEID *)lpOutBuffer = hpt_create_array(_VBUS_P (PCREATE_ARRAY_PARAMS)lpInBuffer);

			if(*(DEVICEID *)lpOutBuffer == INVALID_DEVICEID)
				return -1;
		}
		break;

	case HPT_IOCTL_CREATE_ARRAY_V2:
		{
			if (nInBufferSize!=sizeof(CREATE_ARRAY_PARAMS_V2)) return -1;
			if (nOutBufferSize!=sizeof(DEVICEID)) return -1;

			*(DEVICEID *)lpOutBuffer = hpt_create_array_v2(_VBUS_P (PCREATE_ARRAY_PARAMS_V2)lpInBuffer);

			if (*(DEVICEID *)lpOutBuffer == INVALID_DEVICEID)
				return -1;
		}
		break;

	case HPT_IOCTL_SET_ARRAY_INFO:
		{
			DEVICEID idArray;
			PALTERABLE_ARRAY_INFO pInfo;

			if (nInBufferSize!=sizeof(HPT_SET_ARRAY_INFO)) return -1;
			if (nOutBufferSize!=0) return -1;

			idArray = ((PHPT_SET_ARRAY_INFO)lpInBuffer)->idArray;
			pInfo = &((PHPT_SET_ARRAY_INFO)lpInBuffer)->Info;

			if(hpt_set_array_info(_VBUS_P idArray,  pInfo))
				return -1;
		}
		break;

	case HPT_IOCTL_SET_DEVICE_INFO:
		{
			DEVICEID idDisk;
			PALTERABLE_DEVICE_INFO pInfo;

			if (nInBufferSize!=sizeof(HPT_SET_DEVICE_INFO)) return -1;
			if (nOutBufferSize!=0) return -1;

			idDisk = ((PHPT_SET_DEVICE_INFO)lpInBuffer)->idDisk;
			pInfo = &((PHPT_SET_DEVICE_INFO)lpInBuffer)->Info;
			if(hpt_set_device_info(_VBUS_P idDisk, pInfo) != 0)
				return -1;
		}
		break;

	case HPT_IOCTL_SET_DEVICE_INFO_V2:
		{
			DEVICEID idDisk;
			PALTERABLE_DEVICE_INFO_V2 pInfo;

			if (nInBufferSize < sizeof(HPT_SET_DEVICE_INFO_V2)) return -1;
			if (nOutBufferSize!=0) return -1;

			idDisk = ((PHPT_SET_DEVICE_INFO_V2)lpInBuffer)->idDisk;
			pInfo = &((PHPT_SET_DEVICE_INFO_V2)lpInBuffer)->Info;
			if(hpt_set_device_info_v2(_VBUS_P idDisk, pInfo) != 0)
				return -1;
		}
		break;

	case HPT_IOCTL_SET_BOOT_MARK:
		{
			DEVICEID id;
			PVDevice pTop;
			int i;
			IAL_ADAPTER_T *pAdapter = gIal_Adapter;
			PVBus pVBus;

			if (nInBufferSize!=sizeof(DEVICEID)) return -1;
			id = *(DEVICEID *)lpInBuffer;
			while(pAdapter != 0)
			{
				pVBus = &pAdapter->VBus;
				for(i = 0; i < MAX_ARRAY_PER_VBUS; i++)
				{
					if(!(pTop = pVBus->pVDevice[i])) continue;
					if (pTop->pVBus!=_vbus_p) return -1;
					while (pTop->pParent) pTop = pTop->pParent;
					if (id==0 && pTop->vf_bootmark)
						pTop->vf_bootmark = 0;
					else if (pTop==ID_TO_VDEV(id) && !pTop->vf_bootmark)
						pTop->vf_bootmark = 1;
					else
						continue;
					SyncArrayInfo(pTop);
					break;
				}
				pAdapter = pAdapter->next;
			}
		}
		break;

	case HPT_IOCTL_ADD_SPARE_DISK:
		{
			DEVICEID id;

			if (nInBufferSize!=sizeof(DEVICEID)) return -1;
			if (nOutBufferSize!=0) return -1;

			id = *(DEVICEID *)lpInBuffer;

			if(hpt_add_spare_disk(_VBUS_P id))
				return -1;
		}
		break;

	case HPT_IOCTL_REMOVE_SPARE_DISK:
		{
			DEVICEID id;

			if (nInBufferSize!=sizeof(DEVICEID)) return -1;
			if (nOutBufferSize!=0) return -1;

			id = *(DEVICEID *)lpInBuffer;

			if(hpt_remove_spare_disk(_VBUS_P id))
				return -1;
		}
		break;

	case HPT_IOCTL_ADD_DISK_TO_ARRAY:
		{
			DEVICEID id1,id2;
			id1 = ((PHPT_ADD_DISK_TO_ARRAY)lpInBuffer)->idArray;
			id2 = ((PHPT_ADD_DISK_TO_ARRAY)lpInBuffer)->idDisk;

			if (nInBufferSize != sizeof(HPT_ADD_DISK_TO_ARRAY)) return -1;
			if (nOutBufferSize != 0) return -1;

			if(hpt_add_disk_to_array(_VBUS_P id1, id2))
				return -1;
		}
		break;
#endif
	case HPT_IOCTL_GET_DRIVER_CAPABILITIES:
		{
			PDRIVER_CAPABILITIES cap;
			if (nOutBufferSize<sizeof(DRIVER_CAPABILITIES)) return -1;
			cap = (PDRIVER_CAPABILITIES)lpOutBuffer;

			if(hpt_get_driver_capabilities(cap))
				return -1;
		}
		break;

#ifdef SUPPORT_HPT601	
	case HPT_IOCTL_GET_601_INFO:
		{
			DEVICEID id;
			PHPT601_INFO pInfo;

			if (nInBufferSize!=sizeof(DEVICEID)) return -1;
			if (nOutBufferSize!=sizeof(HPT601_INFO)) return -1;

			id = *(DWORD *)lpInBuffer;
			if (id == INVALID_DEVICEID)	return -1;

			pInfo = (PHPT601_INFO)lpOutBuffer;
			memset(pInfo, 0, sizeof(HPT601_INFO));

			if (hpt_get_601_info(id, pInfo)!=0)
				return -1;
		}
		break;

	case HPT_IOCTL_SET_601_INFO:
		{
			DEVICEID id;
			PHPT601_INFO pInfo;

			if (nInBufferSize!=sizeof(HPT_SET_601_INFO)) return -1;
			if (nOutBufferSize!=0) return -1;

			id = ((PHPT_SET_601_INFO)lpInBuffer)->idDisk;
			pInfo = &((PHPT_SET_601_INFO)lpInBuffer)->Info;
			if(hpt_set_601_info(id, pInfo) != 0)
				return -1;
		}
		break;
#endif
	default:
		return -1;
	}

	if (lpBytesReturned)
		*lpBytesReturned = nOutBufferSize;
	return 0;
}

static int hpt_delete_array(_VBUS_ARG DEVICEID id, DWORD options)
{
	PVDevice	pArray = ID_TO_VDEV(id);
	BOOLEAN	del_block0 = (options & DAF_KEEP_DATA_IF_POSSIBLE)?0:1;
	int i;
	PVDevice pa;

	if((id == 0) || check_VDevice_valid(pArray))
		return -1;

	if(!mIsArray(pArray)) return -1;
	
	if (pArray->u.array.rf_rebuilding || pArray->u.array.rf_verifying ||
		pArray->u.array.rf_initializing)
		return -1;

	for(i=0; i<pArray->u.array.bArnMember; i++) {
		pa = pArray->u.array.pMember[i];
		if (pa && mIsArray(pa)) {
			if (pa->u.array.rf_rebuilding || pa->u.array.rf_verifying || 
				pa->u.array.rf_initializing)
				return -1;
		}
	}

	if (pArray->pVBus!=_vbus_p) {
		HPT_ASSERT(0); 
		return -1;
	}
	fDeleteArray(_VBUS_P pArray, del_block0);
	return 0;
}

#define MAX_EVENTS 10
static HPT_EVENT hpt_event_queue[MAX_EVENTS];
static int event_queue_head=0, event_queue_tail=0;

static int event_queue_add(PHPT_EVENT pEvent)
{
	int p;

	p = (event_queue_tail + 1) % MAX_EVENTS;
	if (p==event_queue_head)
	{
		return -1;
	}
	hpt_event_queue[event_queue_tail] = *pEvent;
	event_queue_tail = p;
	return 0;
}

static int event_queue_remove(PHPT_EVENT pEvent)
{
	if (event_queue_head != event_queue_tail)
	{
		*pEvent = hpt_event_queue[event_queue_head];
		event_queue_head++;
		event_queue_head %= MAX_EVENTS;
		return 0;
	}
	return -1;
}

void HPTLIBAPI ioctl_ReportEvent(UCHAR event, PVOID param)
{
	HPT_EVENT e;
	ZeroMemory(&e, sizeof(e));
	e.EventType = event;
	switch(event)
	{
#ifdef SUPPORT_ARRAY
		case ET_INITIALIZE_ABORTED:
		case ET_INITIALIZE_FAILED:
			memcpy(e.Data, ((PVDevice)param)->u.array.ArrayName, MAX_ARRAY_NAME);
		case ET_INITIALIZE_STARTED:
		case ET_INITIALIZE_FINISHED:
		case ET_REBUILD_STARTED:
		case ET_REBUILD_ABORTED:
		case ET_REBUILD_FAILED:
		case ET_REBUILD_FINISHED:
		case ET_VERIFY_STARTED:
		case ET_VERIFY_ABORTED:
		case ET_VERIFY_FAILED:
		case ET_VERIFY_FINISHED:
		case ET_VERIFY_DATA_ERROR:
		case ET_SPARE_TOOK_OVER:
#endif
		case ET_DEVICE_REMOVED:
		case ET_DEVICE_PLUGGED:
		case ET_DEVICE_ERROR:
			e.DeviceID = VDEV_TO_ID((PVDevice)param);
			break;

		default:
			break;
	}
	event_queue_add(&e);
}

static int hpt_delete_array(_VBUS_ARG DEVICEID id, DWORD options);
static int hpt_get_event(PHPT_EVENT pEvent);
static int hpt_set_array_state(DEVICEID idArray, DWORD state);

/* just to prevent driver from sending more commands */
static void HPTLIBAPI nothing(_VBUS_ARG void *notused){}

static int is_adapter_idle(IAL_ADAPTER_T *pAdapter)
{
	MV_U8 channelIndex;
	
	if (pAdapter->outstandingCommands)
		return 0;

	for (channelIndex=0; channelIndex<MV_SATA_CHANNELS_NUM; channelIndex++) {
		if (pAdapter->mvSataAdapter.sataChannel[channelIndex])
			if (pAdapter->mvSataAdapter.sataChannel[channelIndex]->EdmaQueuedCommands)
				return 0;
	}
	return 1;
}

unsigned long lock_driver_idle(IAL_ADAPTER_T *pAdapter)
{
	unsigned long flags;
	_VBUS_INST(&pAdapter->VBus)
	spin_lock_irqsave(&pAdapter->driver_lock, flags);
	if (!is_adapter_idle(pAdapter)) {
		MV_PRINTD("outstandingCommands is %d, dpc %d, wait..\n", 
			pAdapter->outstandingCommands, atomic_read(&pAdapter->DPC_Request));
		if (!mWaitingForIdle(_VBUS_P0)) CallWhenIdle(_VBUS_P nothing, 0);
		atomic_inc(&pAdapter->DPC_Request);
		do {
			spin_unlock_irqrestore(&pAdapter->driver_lock, flags);
			schedule();
			spin_lock_irqsave(&pAdapter->driver_lock, flags);
		}
		while (!is_adapter_idle(pAdapter));
	}
	else
		atomic_inc(&pAdapter->DPC_Request);
	CheckIdleCall(_VBUS_P0);
	return flags;
}

void unlock_driver_idle(IAL_ADAPTER_T *pAdapter, unsigned long flags)
{
	_VBUS_INST(&pAdapter->VBus)
	atomic_dec(&pAdapter->DPC_Request);
	Check_Idle_Call(pAdapter);
	CheckPendingCall(_VBUS_P0);
	spin_unlock_irqrestore(&pAdapter->driver_lock, flags);
}

static void HPTLIBAPI pass_through_cmd_done(_VBUS_ARG PCommand pCmd)
{
	up(pCmd->pOrgCommand);
}

#ifdef SUPPORT_ARRAY
int verify_vd(PVDevice pVDev, int source_disk_of_RAID1)
{

	PVDevice pMem = pVDev;
	int flag = source_disk_of_RAID1, i;
	PVBus	_vbus_p;

	
	if (check_VDevice_valid(pMem)) return INVALID_DEVICEID;
	if (mIsArray(pMem)) return INVALID_DEVICEID;
	if (!pMem->vf_online) return INVALID_DEVICEID;
	if (!(_vbus_p = pMem->u.disk.pVBus)) return INVALID_DEVICEID;

	/*Not allowed to create array(except create and duplicate RAID1) with the mounted disks*/
	for(i = 0; i < MAX_VDEVICE_PER_VBUS; i++) {
		if(pMem == _vbus_p->pVDevice[i] && (!flag)) {
			IAL_ADAPTER_T *pAdapter = (IAL_ADAPTER_T *)(_vbus_p->OsExt);
			if (sd_inuse(pAdapter->host, i)) {
				hpt_printk("hptmv: Can not Unregister disk, which is used by OS\n");
				return INVALID_DEVICEID;
			}
			break;
		}
	}
	return 0;
}

#endif
int Kernel_DeviceIoControl(_VBUS_ARG
							DWORD dwIoControlCode,       	/* operation control code */
							PVOID lpInBuffer,            	/* input data buffer */
							DWORD nInBufferSize,         	/* size of input data buffer */
							PVOID lpOutBuffer,           	/* output data buffer */
							DWORD nOutBufferSize,        	/* size of output data buffer */
							PDWORD lpBytesReturned      	/* byte count */
						)
{
	IAL_ADAPTER_T *pAdapter;
	unsigned long flags;
	
	switch(dwIoControlCode)
	{
#ifdef SUPPORT_ARRAY
		case HPT_IOCTL_DELETE_ARRAY:
		{
			DEVICEID idArray;
			int iSuccess;
			int i;
			PVDevice pArray;
			PVBus _vbus_p;

			if (nInBufferSize!=sizeof(DEVICEID)+sizeof(DWORD)) return -1;
			if (nOutBufferSize!=sizeof(int)) return -1;
			idArray = *(DEVICEID *)lpInBuffer;

			pArray = ID_TO_VDEV(idArray);

			if((idArray == 0) || check_VDevice_valid(pArray))	
				return -1;
		
			if(!mIsArray(pArray))
				return -1;

			_vbus_p=pArray->pVBus;

			for(i = 0; i < MAX_VDEVICE_PER_VBUS; i++)
				if(pArray == _vbus_p->pVDevice[i]) {
					if (sd_inuse(((IAL_ADAPTER_T *)_vbus_p->OsExt)->host, i)) {
						hpt_printk("Can not delete a mounted device.\n");
						return -1;
					}
					break;
				} 

			pAdapter = (IAL_ADAPTER_T *)_vbus_p->OsExt;
			flags = lock_driver_idle(pAdapter);
			iSuccess = hpt_delete_array(_VBUS_P idArray, *(DWORD*)((DEVICEID *)lpInBuffer+1));
			unlock_driver_idle(pAdapter, flags);

			*(int*)lpOutBuffer = iSuccess;

			if(iSuccess != 0)
				return -1;
			break;
		}
#endif

		case HPT_IOCTL_IDE_PASS_THROUGH:
		{
			PCommand	pCmd;
			unsigned long	lock_flags;
			struct semaphore	sem;			

			PIDE_PASS_THROUGH_HEADER in = (PIDE_PASS_THROUGH_HEADER)lpInBuffer;
			PIDE_PASS_THROUGH_HEADER out = (PIDE_PASS_THROUGH_HEADER)lpOutBuffer;					

			PVDevice		pVDev = ID_TO_VDEV(in->idDisk);

			pAdapter = (IAL_ADAPTER_T *)pVDev->pVBus->OsExt;

		
			lock_flags = lock_driver_idle(pAdapter);

			pCmd = AllocateCommand(_VBUS_P0);
			if (!pCmd) {
				unlock_driver_idle(pAdapter, lock_flags);
				return -1;
			}
			
			pCmd->cf_ide_passthrough = 1;
			sema_init(&sem, 0);
			pCmd->pOrgCommand = &sem;
			pCmd->pVDevice = pVDev;
			pCmd->pfnCompletion = pass_through_cmd_done;

			if (in->protocol==IO_COMMAND_READ) {
				pCmd->cf_data_in = 1;
				pCmd->uCmd.Passthrough.pDataBuffer = IDE_PASS_THROUGH_buffer(out);
			}
			else if (in->protocol==IO_COMMAND_WRITE) {
				pCmd->cf_data_out = 1;
				pCmd->uCmd.Passthrough.pDataBuffer = IDE_PASS_THROUGH_buffer(in);
			}

			pCmd->uCmd.Passthrough.bCommandReg = in->bCommandReg;
			pCmd->uCmd.Passthrough.bLbaLowReg = in->bLbaLowReg;
			pCmd->uCmd.Passthrough.bLbaMidReg = in->bLbaMidReg;
			pCmd->uCmd.Passthrough.bLbaHighReg = in->bLbaHighReg;
			pCmd->uCmd.Passthrough.bDriveHeadReg = in->bDriveHeadReg;
			pCmd->uCmd.Passthrough.bFeaturesReg = in->bFeaturesReg;
			pCmd->uCmd.Passthrough.bSectorCountReg = in->bSectorCountReg;
			pCmd->uCmd.Passthrough.nSectors = in->nSectors;

			pVDev->pfnSendCommand(_VBUS_P pCmd);
			CheckPendingCall(_VBUS_P0);
			
			unlock_driver_idle(pAdapter, lock_flags);

			timer_start_for_cmd(pCmd, 10);
			

			spin_lock_irqsave(&pAdapter->driver_lock, lock_flags);
			if (pCmd->Result != RETURN_SUCCESS) {
				fMvResetBus(_VBUS_P0);
				FreeCommand(_VBUS_P pCmd);
				spin_unlock_irqrestore(&pAdapter->driver_lock, lock_flags);				
				return -1;
			}
			
			out->bCommandReg = pCmd->uCmd.Passthrough.bCommandReg;
			out->bLbaLowReg = pCmd->uCmd.Passthrough.bLbaLowReg;
			out->bLbaMidReg = pCmd->uCmd.Passthrough.bLbaMidReg;
			out->bLbaHighReg = pCmd->uCmd.Passthrough.bLbaHighReg;
			out->bDriveHeadReg = pCmd->uCmd.Passthrough.bDriveHeadReg;
			out->bFeaturesReg = pCmd->uCmd.Passthrough.bFeaturesReg;
			out->bSectorCountReg = pCmd->uCmd.Passthrough.bSectorCountReg;
			out->nSectors = pCmd->uCmd.Passthrough.nSectors;

			FreeCommand(_VBUS_P pCmd);
			spin_unlock_irqrestore(&pAdapter->driver_lock, lock_flags);
		}
		break;
		case HPT_IOCTL_GET_EVENT:
		{
			PHPT_EVENT pInfo;

			if (nInBufferSize!=0) return -1;
			if (nOutBufferSize!=sizeof(HPT_EVENT)) return -1;

			pInfo = (PHPT_EVENT)lpOutBuffer;

			if (hpt_get_event(pInfo)!=0)
				return -1;
		}
		break;
#ifdef SUPPORT_ARRAY
		case HPT_IOCTL_REBUILD_DATA_BLOCK:
			return -1;

		case HPT_IOCTL_SET_ARRAY_STATE:
		{
			DEVICEID idArray;
			DWORD state;

			if (nInBufferSize!=sizeof(HPT_SET_STATE_PARAM)) return -1;
			if (nOutBufferSize!=0) return -1;

			idArray = ((PHPT_SET_STATE_PARAM)lpInBuffer)->idArray;
			state = ((PHPT_SET_STATE_PARAM)lpInBuffer)->state;

			if(hpt_set_array_state(idArray, state)!=0)
				return -1;
		}
		break;
#endif
		case HPT_IOCTL_RESCAN_DEVICES:
		{
			if (nInBufferSize!=0) return -1;
			if (nOutBufferSize!=0) return -1;

			/* stop buzzer if user perform rescan */
			for (pAdapter=gIal_Adapter; pAdapter; pAdapter=pAdapter->next) {
				if (pAdapter->beeping) {
					spin_lock_irqsave(&pAdapter->driver_lock, flags);
					pAdapter->beeping = 0;
					BeepOff(pAdapter->mvSataAdapter.adapterIoBaseAddress);
					spin_unlock_irqrestore(&pAdapter->driver_lock, flags);
				}
			}
		}
		break;

		case HPT_IOCTL_LOCK_DEVICE:
			return -1;
	
		case HPT_IOCTL_UNLOCK_DEVICE:
			return -1;

		default:
		{
			PVDevice pVDev;
			switch(dwIoControlCode) {
			/* read-only ioctl functions can be called directly. */
			case HPT_IOCTL_GET_VERSION:
			case HPT_IOCTL_GET_CONTROLLER_IDS:
			case HPT_IOCTL_GET_CONTROLLER_COUNT:
			case HPT_IOCTL_GET_CONTROLLER_INFO:
			case HPT_IOCTL_GET_CHANNEL_INFO:
			case HPT_IOCTL_GET_LOGICAL_DEVICES:
			case HPT_IOCTL_GET_DEVICE_INFO:
			case HPT_IOCTL_GET_DEVICE_INFO_V2:
			case HPT_IOCTL_GET_EVENT:
			case HPT_IOCTL_GET_DRIVER_CAPABILITIES:
				if(hpt_default_ioctl(_VBUS_P dwIoControlCode, lpInBuffer, nInBufferSize, 
					lpOutBuffer, nOutBufferSize, lpBytesReturned) == -1) return -1;
				break;

			default:

				/*Not allowed to create array/add disk(except RAID1 create and duplicate) to array with OS-USING disk
				  * source_disk_of_RAID1 = 1 : means create RAID1 with create and duplicate option, and the disk is source
				 */
#ifdef SUPPORT_ARRAY
				switch(dwIoControlCode) {
					case HPT_IOCTL_CREATE_ARRAY: 
					{
						 PCREATE_ARRAY_PARAMS param = (PCREATE_ARRAY_PARAMS)lpInBuffer;
						 int i;
						for (i=0; i<param->nDisk; i++) {
							PVDevice pDisk = ID_TO_VDEV(param->Members[i]);
							int	source_disk_of_RAID1=0;							
							if (param->ArrayType == AT_RAID1 && param->CreateFlags == CAF_CREATE_AND_DUPLICATE && i == 0 && param->nDisk == 2)
								source_disk_of_RAID1 = 1;
							else
								source_disk_of_RAID1 = 0;
							if (verify_vd(pDisk, source_disk_of_RAID1))
								return INVALID_DEVICEID; 
						}
						break;
					}
					case HPT_IOCTL_CREATE_ARRAY_V2:
					{
						PCREATE_ARRAY_PARAMS_V2 param = (PCREATE_ARRAY_PARAMS_V2)lpInBuffer;
						int i;
						for (i=0; i<param->nDisk; i++) {
							PVDevice	 pDisk = ID_TO_VDEV(param->Members[i]);
							int	source_disk_of_RAID1=0;		
							if (param->ArrayType == AT_RAID1 && param->CreateFlags == CAF_CREATE_AND_DUPLICATE && i == 0 && param->nDisk == 2)
								source_disk_of_RAID1 = 1;
							else
								source_disk_of_RAID1 = 0;
							if (verify_vd(pDisk, source_disk_of_RAID1))
								return INVALID_DEVICEID; 
						}
						break;
					}
					case HPT_IOCTL_ADD_SPARE_DISK:
					{
						PVDevice pDisk = ID_TO_VDEV(*(DEVICEID *)lpInBuffer);
						if (verify_vd(pDisk, 0))
							return INVALID_DEVICEID; 
						break;
					}
					case HPT_IOCTL_ADD_DISK_TO_ARRAY:
					{
						PVDevice pDisk = ID_TO_VDEV(((PHPT_ADD_DISK_TO_ARRAY)lpInBuffer)->idDisk);
						if (verify_vd(pDisk, 0))
							return INVALID_DEVICEID; 
						break;
					}
				}
#endif
				/* 
				 * GUI always use /proc/scsi/hptmv/0, so the _vbus_p param will be 
				 * wrong for second controller. 
				 */
				switch(dwIoControlCode) {
				case HPT_IOCTL_CREATE_ARRAY:
					pVDev = ID_TO_VDEV(((PCREATE_ARRAY_PARAMS)lpInBuffer)->Members[0]); break;
				case HPT_IOCTL_CREATE_ARRAY_V2:
					pVDev = ID_TO_VDEV(((PCREATE_ARRAY_PARAMS_V2)lpInBuffer)->Members[0]); break;
				case HPT_IOCTL_SET_ARRAY_INFO:
					pVDev = ID_TO_VDEV(((PHPT_SET_ARRAY_INFO)lpInBuffer)->idArray); break;
				case HPT_IOCTL_SET_DEVICE_INFO:
					pVDev = ID_TO_VDEV(((PHPT_SET_DEVICE_INFO)lpInBuffer)->idDisk); break;
				case HPT_IOCTL_SET_DEVICE_INFO_V2:
					pVDev = ID_TO_VDEV(((PHPT_SET_DEVICE_INFO_V2)lpInBuffer)->idDisk); break;
				case HPT_IOCTL_SET_BOOT_MARK:
				case HPT_IOCTL_ADD_SPARE_DISK:
				case HPT_IOCTL_REMOVE_SPARE_DISK:
					pVDev = ID_TO_VDEV(*(DEVICEID *)lpInBuffer); break;
				case HPT_IOCTL_ADD_DISK_TO_ARRAY:
					pVDev = ID_TO_VDEV(((PHPT_ADD_DISK_TO_ARRAY)lpInBuffer)->idArray); break;
				default:
					pVDev = 0;
				}

				if (pVDev && !check_VDevice_valid(pVDev)) {
					_vbus_p = pVDev->pVBus;

					pAdapter = (IAL_ADAPTER_T *)_vbus_p->OsExt;
					/* 
					 * create_array, and other functions can't be executed while channel is 
					 * perform I/O commands. Wait until driver is idle.
					 */
					flags = lock_driver_idle(pAdapter);
					if (hpt_default_ioctl(_VBUS_P dwIoControlCode, lpInBuffer, nInBufferSize, 
						lpOutBuffer, nOutBufferSize, lpBytesReturned) == -1) {
						unlock_driver_idle(pAdapter, flags);
						return -1;
					}
					unlock_driver_idle(pAdapter, flags);
					
					switch(dwIoControlCode)	{
						case HPT_IOCTL_CREATE_ARRAY:
						{
							pAdapter=(IAL_ADAPTER_T *)(ID_TO_VDEV(*(DEVICEID *)lpOutBuffer))->pVBus->OsExt;
							spin_lock_irqsave(&pAdapter->driver_lock, flags);
							if(((PCREATE_ARRAY_PARAMS)lpInBuffer)->CreateFlags & CAF_CREATE_AND_DUPLICATE)
							{
								  (ID_TO_VDEV(*(DEVICEID *)lpOutBuffer))->u.array.rf_auto_rebuild = 0;
								  hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pAdapter, 
											ID_TO_VDEV(*(DEVICEID *)lpOutBuffer), DUPLICATE);
							}
							else if(((PCREATE_ARRAY_PARAMS)lpInBuffer)->CreateFlags & CAF_CREATE_R5_ZERO_INIT)
							{
								  hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pAdapter, 
											ID_TO_VDEV(*(DEVICEID *)lpOutBuffer), INITIALIZE);
							}
							else if(((PCREATE_ARRAY_PARAMS)lpInBuffer)->CreateFlags & CAF_CREATE_R5_BUILD_PARITY)
							{
								  hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pAdapter, 
											ID_TO_VDEV(*(DEVICEID *)lpOutBuffer), REBUILD_PARITY);
							}
							spin_unlock_irqrestore(&pAdapter->driver_lock, flags);
							break;
						}
		
						case HPT_IOCTL_CREATE_ARRAY_V2:
						{
							pAdapter=(IAL_ADAPTER_T *)(ID_TO_VDEV(*(DEVICEID *)lpOutBuffer))->pVBus->OsExt;
							spin_lock_irqsave(&pAdapter->driver_lock, flags);
							if(((PCREATE_ARRAY_PARAMS_V2)lpInBuffer)->CreateFlags & CAF_CREATE_AND_DUPLICATE)
							{
								  (ID_TO_VDEV(*(DEVICEID *)lpOutBuffer))->u.array.rf_auto_rebuild = 0;
								  hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pAdapter, 
											ID_TO_VDEV(*(DEVICEID *)lpOutBuffer), DUPLICATE);
							}
							else if(((PCREATE_ARRAY_PARAMS_V2)lpInBuffer)->CreateFlags & CAF_CREATE_R5_ZERO_INIT)
							{
								  hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pAdapter, 
											ID_TO_VDEV(*(DEVICEID *)lpOutBuffer), INITIALIZE);
							}
							else if(((PCREATE_ARRAY_PARAMS_V2)lpInBuffer)->CreateFlags & CAF_CREATE_R5_BUILD_PARITY)
							{
								  hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pAdapter, 
											ID_TO_VDEV(*(DEVICEID *)lpOutBuffer), REBUILD_PARITY);
							}
							spin_unlock_irqrestore(&pAdapter->driver_lock, flags);
							break;
						}
		
						case HPT_IOCTL_ADD_DISK_TO_ARRAY:
						{
							PVDevice pArray = ID_TO_VDEV(((PHPT_ADD_DISK_TO_ARRAY)lpInBuffer)->idArray);
							pAdapter=(IAL_ADAPTER_T *)pArray->pVBus->OsExt;
							if(pArray->u.array.rf_rebuilding == 0)
							{		
								unsigned long timeout;
								timeout = jiffies + 3 * HZ;	
								spin_lock_irqsave(&pAdapter->driver_lock, flags);
								pArray->u.array.rf_auto_rebuild = 0;
								pArray->u.array.rf_abort_rebuild = 0;
								hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pAdapter, pArray, DUPLICATE);
								spin_unlock_irqrestore(&pAdapter->driver_lock, flags);
								while (!pArray->u.array.rf_rebuilding && time_before(jiffies, timeout)) schedule();
							}
							break;
						}
					}
				}
				else
					return -1;
				break;
			}
		}
	}

	if (lpBytesReturned)
		*lpBytesReturned = nOutBufferSize;
	return 0;
}

int hpt_get_event(PHPT_EVENT pEvent)
{
	return event_queue_remove(pEvent);
}

#ifdef SUPPORT_ARRAY
int hpt_set_array_state(DEVICEID idArray, DWORD state)
{
	IAL_ADAPTER_T *pAdapter;
	PVDevice pVDevice = ID_TO_VDEV(idArray);
	int	i;
	unsigned long timeout;
	unsigned long flags;

	if(idArray == 0 || check_VDevice_valid(pVDevice))	return -1;
	if(!mIsArray(pVDevice)) return -1;
	if(!pVDevice->vf_online || pVDevice->u.array.rf_broken) return -1;

	pAdapter=(IAL_ADAPTER_T *)pVDevice->pVBus->OsExt;

	timeout = jiffies + 80 * HZ; /* should be long than hard-reset time */
	switch(state)
	{
		case MIRROR_REBUILD_START:
		{
			if (pVDevice->u.array.rf_rebuilding ||
				pVDevice->u.array.rf_verifying ||
				pVDevice->u.array.rf_initializing)
				return -1;
			
			spin_lock_irqsave(&pAdapter->driver_lock, flags);

			pVDevice->u.array.rf_auto_rebuild = 0;
			pVDevice->u.array.rf_abort_rebuild = 0;
			hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pAdapter, pVDevice, 
				(UCHAR)((pVDevice->u.array.CriticalMembers || pVDevice->VDeviceType == VD_RAID_1)? DUPLICATE : REBUILD_PARITY));

			spin_unlock_irqrestore(&pAdapter->driver_lock, flags);

			while (!pVDevice->u.array.rf_rebuilding && time_before(jiffies, timeout)) schedule();
		}

		break;

		case AS_VERIFY_START:
		{
			if (pVDevice->u.array.rf_rebuilding ||
				pVDevice->u.array.rf_verifying ||
				pVDevice->u.array.rf_initializing)
				return -1;
	
			spin_lock_irqsave(&pAdapter->driver_lock, flags);
			
			pVDevice->u.array.RebuildSectors = 0;
			hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pAdapter, pVDevice, VERIFY_R);

			spin_unlock_irqrestore(&pAdapter->driver_lock, flags);
			
			while (!(pVDevice->u.array.rf_verifying || pVDevice->u.array.rf_rebuilding)
					&& time_before(jiffies, timeout)) schedule();
		}
		break;

		case AS_INITIALIZE_START:
		{
			if (pVDevice->u.array.rf_rebuilding ||
				pVDevice->u.array.rf_verifying ||
				pVDevice->u.array.rf_initializing)
				return -1;

			spin_lock_irqsave(&pAdapter->driver_lock, flags);
			hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pAdapter, pVDevice, VERIFY_R);
			spin_unlock_irqrestore(&pAdapter->driver_lock, flags);
			
			while (!pVDevice->u.array.rf_initializing && time_before(jiffies, timeout)) schedule();
		}
		break;

		case MIRROR_REBUILD_ABORT:
		{
			for(i = 0; i < pVDevice->u.array.bArnMember; i++) {
				if (pVDevice->u.array.pMember[i] && 
					mIsArray(pVDevice->u.array.pMember[i]) &&
					pVDevice->u.array.pMember[i]->u.array.rf_rebuilding)
				{
					/* take care if you use spinlock in recursive call */
					hpt_set_array_state(VDEV_TO_ID(pVDevice->u.array.pMember[i]), state);
				}
			}
			
			if(pVDevice->u.array.rf_rebuilding != 1)
				return -1;
			
			spin_lock_irqsave(&pAdapter->driver_lock, flags);
			pVDevice->u.array.rf_abort_rebuild = 1;
			spin_unlock_irqrestore(&pAdapter->driver_lock, flags);
			
			while (pVDevice->u.array.rf_rebuilding && time_before(jiffies, timeout)) schedule();
		}
		break;

		case AS_VERIFY_ABORT:
		{
			if(pVDevice->u.array.rf_verifying != 1)
				return -1;

			spin_lock_irqsave(&pAdapter->driver_lock, flags);
			pVDevice->u.array.rf_abort_rebuild = 1;
			spin_unlock_irqrestore(&pAdapter->driver_lock, flags);
			
			while (pVDevice->u.array.rf_verifying && time_before(jiffies, timeout)) schedule();
		}
		break;

		case AS_INITIALIZE_ABORT:
		{
			if(pVDevice->u.array.rf_initializing != 1)
				return -1;

			spin_lock_irqsave(&pAdapter->driver_lock, flags);
			pVDevice->u.array.rf_abort_rebuild = 1;
			spin_unlock_irqrestore(&pAdapter->driver_lock, flags);
			
			while (pVDevice->u.array.rf_initializing && time_before(jiffies, timeout)) schedule();
		}
		break;

		default:
			return -1;
	}

	return 0;
}
#endif

/***************************************************************************
 * Function:     hpt_rebuild_data_block
 * Description:  rebuild a mirror block
 * Returns:      None
 ***************************************************************************/
#ifdef SUPPORT_ARRAY
int HPTLIBAPI R1ControlSgl(_VBUS_ARG PCommand pCmd, FPSCAT_GATH pSgTable, int logical)
{
	ULONG bufferSize = SECTOR_TO_BYTE(pCmd->uCmd.R1Control.nSectors);
	if (pCmd->uCmd.R1Control.Command==CTRL_CMD_VERIFY)
		bufferSize<<=1;
	if (logical) {
		pSgTable->dSgAddress = (ULONG_PTR)pCmd->uCmd.R1Control.Buffer;
		pSgTable->wSgSize = (USHORT)bufferSize;
		pSgTable->wSgFlag = SG_FLAG_EOT;
	}
	else {
		/* build physical SG table for pCmd->uCmd.R1Control.Buffer */
		pSgTable->dSgAddress = fOsPhysicalAddress(pCmd->uCmd.R1Control.Buffer);
		if(pSgTable->dSgAddress & 1) return 0;
		pSgTable->wSgSize = (USHORT)bufferSize;
		pSgTable->wSgFlag = SG_FLAG_EOT;
	}
	return 1;
}

static int End_Job = 0;
static void HPTLIBAPI thread_io_done(_VBUS_ARG PCommand pCmd)
{
	End_Job = 1;
	up((struct semaphore *)pCmd->pOrgCommand);
}

void hpt_rebuild_data_block(IAL_ADAPTER_T *pAdapter, PVDevice pArray, UCHAR flags)
{
	LBA_T capacity;
	PCommand pCmd;
	UINT result;
	int needsync=0, retry=0, needdelete=0;
	unsigned long Lock_Flag;
	void *buffer = 0;
	struct semaphore sem;

	_VBUS_INST(&pAdapter->VBus)
	
	capacity = pArray->VDeviceCapacity;
	do_div(capacity, (pArray->u.array.bArnMember-1));
	
	if (pArray->u.array.RebuildSectors>=capacity)
		return;

	spin_lock_irqsave(&pAdapter->driver_lock, Lock_Flag);
	
	if (pArray->u.array.rf_broken) goto cmd_fail;

	switch(flags) {
	case DUPLICATE:
	case REBUILD_PARITY:
		if(pArray->u.array.rf_rebuilding == 0){
			pArray->u.array.rf_rebuilding = 1;
			hpt_printk("Rebuilding started.\n");
			ioctl_ReportEvent(ET_REBUILD_STARTED, pArray);
		}
		break;

	case INITIALIZE:
		if(pArray->u.array.rf_initializing == 0){
			pArray->u.array.rf_initializing = 1;
			hpt_printk("Initializing started.\n");
			ioctl_ReportEvent(ET_INITIALIZE_STARTED, pArray);
		}
		break;
	case VERIFY_R:
		if(pArray->u.array.rf_verifying == 0){
			pArray->u.array.rf_verifying = 1;
			hpt_printk("Verifying started.\n");
			ioctl_ReportEvent(ET_VERIFY_STARTED, pArray);
		}
		break;
	}

	MV_PRINTD("rebuild_begin\n");

retry_cmd:
	pCmd = AllocateCommand(_VBUS_P0);
	pCmd->cf_control = 1;
	End_Job = 0;
	sema_init(&sem, 0);
	pCmd->pOrgCommand = &sem;

	if (pArray->VDeviceType==VD_RAID_1) {
		/* 32K rebuild. */
		#define MAX_REBUILD_SECTORS 0x40
		if(!(buffer = kmalloc(SECTOR_TO_BYTE(MAX_REBUILD_SECTORS), GFP_ATOMIC))) {
			FreeCommand(_VBUS_P pCmd);
			spin_unlock_irqrestore(&pAdapter->driver_lock, Lock_Flag);
			return;
		}

		switch(flags) {
		case DUPLICATE:
			pCmd->uCmd.R1Control.Command = CTRL_CMD_REBUILD;
			pCmd->uCmd.R1Control.nSectors = MAX_REBUILD_SECTORS;
			break;
		case VERIFY_R:
			pCmd->uCmd.R1Control.Command = CTRL_CMD_VERIFY;
			pCmd->uCmd.R1Control.nSectors = MAX_REBUILD_SECTORS/2;
			break;
		case INITIALIZE:
			pCmd->uCmd.R1Control.Command = CTRL_CMD_INIT;
			pCmd->uCmd.R1Control.nSectors = MAX_REBUILD_SECTORS;
			break;
		}

		pCmd->uCmd.R1Control.Lba = pArray->u.array.RebuildSectors;
		if (capacity - pArray->u.array.RebuildSectors < pCmd->uCmd.R1Control.nSectors)
			pCmd->uCmd.R1Control.nSectors = capacity - pArray->u.array.RebuildSectors;

		pCmd->uCmd.R1Control.Buffer = buffer;
		pCmd->pfnBuildSgl = R1ControlSgl;
	}
	else if (pArray->VDeviceType==VD_RAID_5) {
		switch(flags) {
		case DUPLICATE:
		case REBUILD_PARITY:
			pCmd->uCmd.R5Control.Command = CTRL_CMD_REBUILD; break;
		case VERIFY_R:
			pCmd->uCmd.R5Control.Command = CTRL_CMD_VERIFY; break;
		case INITIALIZE:
			pCmd->uCmd.R5Control.Command = CTRL_CMD_INIT; break;
		}
		pCmd->uCmd.R5Control.StripeLine=pArray->u.array.RebuildSectors>>pArray->u.array.bArBlockSizeShift;
	}
	else
		HPT_ASSERT(0);

	pCmd->pVDevice = pArray;
	pCmd->pfnCompletion = thread_io_done;
	pArray->pfnSendCommand(_VBUS_P pCmd);
	CheckPendingCall(_VBUS_P0);

	spin_unlock_irqrestore(&pAdapter->driver_lock, Lock_Flag);
	
	if (!End_Job) {
		timer_start_for_cmd(pCmd, 60);
	}
	
	spin_lock_irqsave(&pAdapter->driver_lock, Lock_Flag);

	if (!End_Job) {
		hpt_printk("timeout, reset\n");
		fMvResetBus(_VBUS_P0);
	}
	result = pCmd->Result;
	FreeCommand(_VBUS_P pCmd);
	if (buffer) kfree(buffer);
	MV_PRINTD("cmd finished %d", result);

	switch(result) {
	case RETURN_SUCCESS:
		if (!pArray->u.array.rf_abort_rebuild) {
			if(pArray->u.array.RebuildSectors < capacity) {
				hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pAdapter, pArray, flags);
			}
			else {
				switch (flags) {
				case DUPLICATE:
				case REBUILD_PARITY:
					needsync = 1;
					pArray->u.array.rf_rebuilding = 0;
					pArray->u.array.rf_need_rebuild = 0;
					pArray->u.array.CriticalMembers = 0;
					pArray->u.array.RebuildSectors = MAX_LBA_T;
					pArray->u.array.rf_duplicate_and_create = 0;
					hpt_printk("Rebuilding finished.\n");
					ioctl_ReportEvent(ET_REBUILD_FINISHED, pArray);
					break;
				case INITIALIZE:
					needsync = 1;
					pArray->u.array.rf_initializing = 0;
					pArray->u.array.rf_need_rebuild = 0;
					pArray->u.array.RebuildSectors = MAX_LBA_T;
					hpt_printk("Initializing finished.\n");
					ioctl_ReportEvent(ET_INITIALIZE_FINISHED, pArray);
					break;
				case VERIFY_R:
					pArray->u.array.rf_verifying = 0;
					hpt_printk("Verifying finished.\n");
					ioctl_ReportEvent(ET_VERIFY_FINISHED, pArray);
					break;
				}
			}
		}
		else {
			pArray->u.array.rf_abort_rebuild = 0;
			if (pArray->u.array.rf_rebuilding) {
				hpt_printk("Abort rebuilding.\n");
				pArray->u.array.rf_rebuilding = 0;
				pArray->u.array.rf_duplicate_and_create = 0;
				ioctl_ReportEvent(ET_REBUILD_ABORTED, pArray);
			}
			else if (pArray->u.array.rf_verifying) {
				hpt_printk("Abort verifying.\n");
				pArray->u.array.rf_verifying = 0;
				ioctl_ReportEvent(ET_VERIFY_ABORTED, pArray);
			}
			else if (pArray->u.array.rf_initializing) {
				hpt_printk("Abort initializing.\n");
				pArray->u.array.rf_initializing = 0;
				ioctl_ReportEvent(ET_INITIALIZE_ABORTED, pArray);
			}
			needdelete=1;
		}
		break;
	
	case RETURN_DATA_ERROR:
		if (flags==VERIFY_R) {
			/*
			 * verify may fail due to OS update I/O buffer asynchronously. 
			 * try to recover this case.
			 */
			if (pArray->VDeviceType==VD_RAID_1) {
				if (++retry<3) {
					spin_unlock_irqrestore(&pAdapter->driver_lock, Lock_Flag);
					set_current_state(TASK_INTERRUPTIBLE);
					schedule_timeout(10*HZ);
					spin_lock_irqsave(&pAdapter->driver_lock, Lock_Flag);
					goto retry_cmd;
				}
			}
			needsync = 1;
			pArray->u.array.rf_verifying = 0;
			pArray->u.array.rf_need_rebuild = 1;
			hpt_printk("Verifying failed: found inconsistency\n");
			ioctl_ReportEvent(ET_VERIFY_DATA_ERROR, pArray);
			ioctl_ReportEvent(ET_VERIFY_FAILED, pArray);

			if (!pArray->vf_online || pArray->u.array.rf_broken) break;

			pArray->u.array.rf_auto_rebuild = 0;
			pArray->u.array.rf_abort_rebuild = 0;
			hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pAdapter, pArray,
				(pArray->VDeviceType == VD_RAID_1) ? DUPLICATE : REBUILD_PARITY);
		}
		break;
		
	default:
		hpt_printk("command failed with error %d\n", result);
		if (++retry<3 && result != RETURN_BAD_DEVICE) {
			hpt_printk("retry (%d)\n", retry);
			goto retry_cmd;
		}
cmd_fail:
		pArray->u.array.rf_abort_rebuild = 0;
		switch (flags) {
		case DUPLICATE:
		case REBUILD_PARITY:
			needsync = 1;
			pArray->u.array.rf_rebuilding = 0;
			pArray->u.array.rf_duplicate_and_create = 0;
			hpt_printk((flags==DUPLICATE)? "Duplicating failed.\n":"Rebuilding failed.\n");
			ioctl_ReportEvent(ET_REBUILD_FAILED, pArray);
			break;
		case INITIALIZE:
			needsync = 1;
			pArray->u.array.rf_initializing = 0;
			hpt_printk("Initializing failed.\n");
			ioctl_ReportEvent(ET_INITIALIZE_FAILED, pArray);
			break;
		case VERIFY_R:
			needsync = 1;
			pArray->u.array.rf_verifying = 0;
			hpt_printk("Verifying failed.\n");
			ioctl_ReportEvent(ET_VERIFY_FAILED, pArray);
			break;
		}
		needdelete=1;
	}			

	/* wait until no other active cmds. */
	while (pAdapter->outstandingCommands) {
		MV_PRINTD("outstandingCommands is %d, wait..\n", pAdapter->outstandingCommands);
		/* put this to have driver stop processing system commands quickly */
		if (!mWaitingForIdle(_VBUS_P0)) CallWhenIdle(_VBUS_P nothing, 0);
		spin_unlock_irqrestore(&pAdapter->driver_lock, Lock_Flag);
		schedule();
		spin_lock_irqsave(&pAdapter->driver_lock, Lock_Flag);
	}

	if (needsync) SyncArrayInfo(pArray);
	if(needdelete && (pArray->u.array.rf_duplicate_must_done || (flags == INITIALIZE))) {
		CheckIdleCall(_VBUS_P0);
		fDeleteArray(_VBUS_P pArray, TRUE);
	}

	if (atomic_read(&pAdapter->DPC_Request)<1) /* see also lock_driver_idle */
	{
		Check_Idle_Call(pAdapter);
		CheckPendingCall(_VBUS_P0);
	}

	spin_unlock_irqrestore(&pAdapter->driver_lock, Lock_Flag);
}
#endif

typedef struct _hpt_GET_INFO {
	char *buffer;
	int buflength;
	int bufoffset;
	int buffillen;
	int filpos;	
} HPT_GET_INFO;

#define shortswap(w) ((WORD)((w)>>8 | ((w) & 0xFF)<<8))

static void get_disk_name(char *name, PDevice pDev)
{
	int i;
	MV_SATA_CHANNEL *pMvSataChannel = pDev->mv;
	IDENTIFY_DATA2 *pIdentifyData = (IDENTIFY_DATA2 *)pMvSataChannel->identifyDevice;
	
	for (i = 0; i < 10; i++) 
		((WORD*)name)[i] = shortswap(pIdentifyData->ModelNumber[i]);
	name[20] = '\0';
}

static void hpt_copy_mem_info(HPT_GET_INFO *pInfo, char *data, int datalen)
{
	if (pInfo->filpos < pInfo->bufoffset) {	/* Current offset before buffer offset */
		if (pInfo->filpos + datalen <= pInfo->bufoffset) {
			pInfo->filpos += datalen; 		/* Discard if completely before buffer */
			return;
		} else {				/* Partial copy, set to begin */
			data += (pInfo->bufoffset - pInfo->filpos);
			datalen  -= (pInfo->bufoffset - pInfo->filpos);
			pInfo->filpos = pInfo->bufoffset;
		}
	}
	
	pInfo->filpos += datalen;		/* Update current offset */
	if (pInfo->buffillen == pInfo->buflength) /* Buffer full, discard */
		return;
	
	if (pInfo->buflength - pInfo->buffillen < datalen)  /* Overflows buffer ? */
	datalen = pInfo->buflength - pInfo->buffillen;
	
	memcpy(pInfo->buffer + pInfo->buffillen, data, datalen);
	pInfo->buffillen += datalen;
}

static int hpt_copy_info(HPT_GET_INFO *pinfo, char *fmt, ...) 
{
	va_list args;
	char buf[128];
	int len;
  
	va_start(args, fmt);
	len = vsprintf(buf, fmt, args);
	va_end(args);
	hpt_copy_mem_info(pinfo, buf, len);
	return len;
}

static void hpt_copy_disk_info(HPT_GET_INFO *pinfo, PVDevice pVDev, int hostno, MV_U8 channel)
{
	char name[32], arrayname[16], *status;

	get_disk_name(name, &pVDev->u.disk);
	
	if (!pVDev->u.disk.df_on_line)
		status = "Disabled";
	else if (pVDev->VDeviceType==VD_SPARE)
		status = "Spare   ";
	else
		status = "Normal  ";

#ifdef SUPPORT_ARRAY
	if(pVDev->pParent) {
		memcpy(arrayname, pVDev->pParent->u.array.ArrayName, MAX_ARRAY_NAME);
		if (pVDev->pParent->u.array.CriticalMembers & (1<<pVDev->bSerialNumber))
			status = "Degraded";
	}
	else
#endif
		arrayname[0]=0;
	
	hpt_copy_info(pinfo, "Controller %d Channel %d  %s  %5dMB  %s %s\n",
	  hostno, channel, name, (ULONG)(pVDev->VDeviceCapacity>>11), status, arrayname);
}

#ifdef SUPPORT_ARRAY
static void hpt_copy_array_info(HPT_GET_INFO *pinfo, int nld, PVDevice pArray, char *sd)
{
	int i;
	char *sType=0, *sStatus=0;
	char buf[32];
	PVDevice pTmpArray;

	switch (pArray->VDeviceType) {
	case VD_RAID_0:
		for (i = 0; (UCHAR)i < pArray->u.array.bArnMember; i++) 
		  if(pArray->u.array.pMember[i])
		  {
			  if(mIsArray(pArray->u.array.pMember[i]))
				 sType = "RAID 1/0   ";
			  else
				 sType = "RAID 0     ";
			  break;
		  }
		break;
	case VD_RAID_1:
		sType = "RAID 1     ";
		break;
	case VD_JBOD:
		sType = "JBOD       ";
		break;
	case VD_RAID_5:
		sType = "RAID 5     ";
		break;
	default:
		sType = "N/A        ";
		break;
	}
	
	if (pArray->vf_online == 0)
		sStatus = "Disabled";
	else if (pArray->u.array.rf_broken)
		sStatus = "Critical";
	for (i = 0; (UCHAR)i < pArray->u.array.bArnMember; i++)
	{
		if (!sStatus) 
		{
			if(mIsArray(pArray->u.array.pMember[i]))
				pTmpArray = pArray->u.array.pMember[i];
			else
			   pTmpArray = pArray;
			if (pTmpArray->u.array.rf_rebuilding) {
				sprintf(buf, "Rebuilding %d%%", 
					(ULONG)(pTmpArray->u.array.RebuildSectors>>11)*100 /
					((ULONG)(pTmpArray->VDeviceCapacity>>11)/(pTmpArray->u.array.bArnMember-1)));
				sStatus = buf;
			}
			else if (pTmpArray->u.array.rf_verifying) {
				sprintf(buf, "Verifying %d%%", 
					(ULONG)(pTmpArray->u.array.RebuildSectors>>11)*100 /
					((ULONG)(pTmpArray->VDeviceCapacity>>11)/(pTmpArray->u.array.bArnMember-1)));
				sStatus = buf;
			}
			else if (pTmpArray->u.array.rf_need_rebuild)
				sStatus = "Critical";
			else if (pTmpArray->u.array.rf_broken)
				sStatus = "Critical";
			if(pTmpArray == pArray) goto out;
		}
		else
			goto out;
	}
out:	
	if (!sStatus) sStatus = "Normal";
	hpt_copy_info(pinfo, "%2d  %8s   %11s  %-20s  %5dMB  %s\n", 
		nld, sd, sType, pArray->u.array.ArrayName, (ULONG)(pArray->VDeviceCapacity>>11), sStatus);
}
#endif

static void get_sd_name(char *name, struct Scsi_Host *host, int id)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	Scsi_Device *SDptr;
	int first_devno = 0;
	name[0] = 0;
	SDptr = host->host_queue;
	if (SDptr && SDptr->device_queue) 
		first_devno = MINOR(SDptr->device_queue->request.rq_dev)/16;
	while (SDptr) {
	   if (SDptr->id == id) {
		   if(SDptr->device_queue) {
			   int devno = (SDptr->device_queue->request.rq_dev)?(MINOR(SDptr->device_queue->request.rq_dev)/16):(first_devno+id);
			   sprintf(name, "/dev/sd%c", 'a' + devno);
		   }
		   break;
	   }
	   SDptr = SDptr->next;
	}
#else 
	struct block_device *bdev = get_bdev(host, id);
	name[0] = 0;
	if (bdev) {
		sprintf(name, "/dev/%s", bdev->bd_disk->disk_name);
		blkdev_put(bdev __BDEV_RAW);
	}
#endif
}

int hpt_get_info(int hostno,char *buffer,char **start,off_t offset,int length)
{
	IAL_ADAPTER_T *pAdapter;
	struct Scsi_Host *host;
	PVBus _vbus_p;

	HPT_GET_INFO info;
	MV_U8 channel;
	UINT j,i;
	PVDevice pVDev;
	char sd[32];

	pAdapter = gIal_Adapter;

	do{
		host = pAdapter->host;
		if(host->host_no == hostno) break;
		pAdapter = pAdapter->next;
	}while(pAdapter != 0);

	HPT_ASSERT(pAdapter);
	
	if (pAdapter->beeping) {
		unsigned long flags;
		spin_lock_irqsave(&pAdapter->driver_lock, flags);
		pAdapter->beeping = 0;
		BeepOff(pAdapter->mvSataAdapter.adapterIoBaseAddress);
		spin_unlock_irqrestore(&pAdapter->driver_lock, flags);
	}
	
	_vbus_p = &pAdapter->VBus;
 
	info.buffer     = buffer;
	info.buflength  = length;
	info.bufoffset  = offset;
	info.filpos     = 0;
	info.buffillen  = 0;

	if (start) *start = buffer;

	hpt_copy_info(&info, "%s\nVersion %s(%s/%s)\n\n", 
		DRIVER_NAME, DRIVER_VERSION, __DATE__, __TIME__);

	hpt_copy_info(&info, "Physical device list\n");
	hpt_copy_info(&info, "Controller   Channel    Model                Capacity  Status   Array\n");
	hpt_copy_info(&info, "------------------------------------------------------------------------\n");

	for (channel = 0; channel < MV_SATA_CHANNELS_NUM; channel++)
	{
		pVDev = &(pAdapter->VDevices[channel]);
		if(pVDev->u.disk.df_on_line)
			 hpt_copy_disk_info(&info, pVDev, pAdapter->mvSataAdapter.adapterId+1, channel+1);
	}

	hpt_copy_info(&info, "\nLogical device list\n");
	hpt_copy_info(&info, "No. sd         Type         Name                 Capacity  Status\n");
	hpt_copy_info(&info, "------------------------------------------------------------------------\n");
	
	j=1;
	for(i = 0; i < MAX_VDEVICE_PER_VBUS; i++){
		pVDev = _vbus_p->pVDevice[i];
		if(pVDev){
			j=i+1;
			get_sd_name(sd, host, i);
#ifdef SUPPORT_ARRAY
			if(mIsArray(pVDev)){
	is_array:
				hpt_copy_array_info(&info, j, pVDev, sd);
			}
			else
#endif
			{
				char name[32];
#ifdef SUPPORT_ARRAY
				/* it may be add to an array after driver loaded, check it */
				if (pVDev->pParent)
					/* in this case, pVDev can only be a RAID 1 source disk. */
					if (pVDev->pParent->VDeviceType==VD_RAID_1 && pVDev==pVDev->pParent->u.array.pMember[0]) 
						goto is_array;
#endif
				if (pVDev->vf_online)
				{
					get_disk_name(name, &pVDev->u.disk);
					hpt_copy_info(&info, 
						"%2d  %3s   %s  %s  %5dMB  %s\n",
						j, sd, "Single disk", name, (ULONG)(pVDev->VDeviceCapacity>>11),
						/* gmm 2001-6-19: Check if pVDev has been added to an array. */
						((pVDev->pParent) ? "Unavailable" : "Normal"));
				}
			 }
		 }
	 }
	
	
	return info.buffillen;
}

/*************************************************************************
 * hpt_set_asc_info
 * Usage:
 *   echo "hpt rebuild start" > /proc/scsi/hptmv/x
 *   echo "hpt rebuild stop" > /proc/scsi/hptmv/x
 *   echo "hpt verify start x" > /proc/scsi/hptmv/x
 *   echo "hpt verify stop x" > /proc/scsi/hptmv/x
 *************************************************************************/
int hpt_add_disk_to_array(_VBUS_ARG DEVICEID idArray, DEVICEID idDisk);

int hpt_set_asc_info(IAL_ADAPTER_T *pAdapter, char *buffer,int length)
{
	int orig_length = length+4;

	PVBus _vbus_p = &pAdapter->VBus;
	unsigned long flags;
	
#ifdef SUPPORT_ARRAY
	PVDevice pArray;
	PVDevice pSubArray, pVDev;
	UINT iarray, ichan, i;
#endif

#ifdef SUPPORT_ARRAY
	if (length>=8 && strncmp(buffer, "rebuild ", 8)==0) {
		buffer+=8;
		length-=8;
		if (length>=5 && strncmp(buffer, "start", 5)==0) {
			/* start rebuild */
			spin_lock_irqsave(&pAdapter->driver_lock, flags);
			for(i = 0; i < MAX_ARRAY_PER_VBUS; i++)
				if ((pArray=ArrayTables(i))->u.array.dArStamp==0)
					continue; 
				else{
					if (pArray->u.array.rf_need_rebuild && !pArray->u.array.rf_rebuilding)
						hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pAdapter, pArray, 
							(UCHAR)((pArray->u.array.CriticalMembers || pArray->VDeviceType == VD_RAID_1)? DUPLICATE : REBUILD_PARITY));
				}
			spin_unlock_irqrestore(&pAdapter->driver_lock, flags);
			return orig_length;
		}
		else if (length>=4 && strncmp(buffer, "stop", 4)==0) {
			/* stop rebuild */
			spin_lock_irqsave(&pAdapter->driver_lock, flags);
			for(i = 0; i < MAX_ARRAY_PER_VBUS; i++)
				if ((pArray=ArrayTables(i))->u.array.dArStamp==0)
					continue; 
				else{
					if (pArray->u.array.rf_rebuilding)
						pArray->u.array.rf_abort_rebuild = 1;
				}
			spin_unlock_irqrestore(&pAdapter->driver_lock, flags);
			return orig_length;
		}
		else if (length>=3 && buffer[1]==','&& buffer[0]>='1'&& buffer[2]>='1')
		{
			/* hpt rebuild x,x */
			iarray = buffer[0]-'1';
			ichan = buffer[2]-'1';

			if(iarray >= MAX_VDEVICE_PER_VBUS || ichan >= MV_SATA_CHANNELS_NUM) return -EINVAL;

			pArray = _vbus_p->pVDevice[iarray];
			if (!pArray || (pArray->vf_online == 0)) return -EINVAL;

			pVDev = &pAdapter->VDevices[ichan];
			if(!pVDev->u.disk.df_on_line || pVDev->pParent) return -EINVAL;

			/* Not allow to use a mounted disk ??? test*/
			for(i = 0; i < MAX_VDEVICE_PER_VBUS; i++)
				if(pVDev == _vbus_p->pVDevice[i])
				{
					if (sd_inuse(pAdapter->host, i)) {
						hpt_printk("Can not use disk used by OS\n");
						return -EINVAL;
					}
					break;
				}

			switch(pArray->VDeviceType) {
			case VD_RAID_1:
			case VD_RAID_5:
			{
				pSubArray = pArray;
loop:
				flags = lock_driver_idle(pAdapter);
				if (hpt_add_disk_to_array(_VBUS_P VDEV_TO_ID(pSubArray), VDEV_TO_ID(pVDev))) {
					unlock_driver_idle(pAdapter, flags);
					return -EINVAL;
				}
				pSubArray->u.array.rf_auto_rebuild = 0;
				pSubArray->u.array.rf_abort_rebuild = 0;
				hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pAdapter, pSubArray, DUPLICATE);
				unlock_driver_idle(pAdapter, flags);
				break;
			}
			case VD_RAID_0:
				for (i = 0; (UCHAR)i < pArray->u.array.bArnMember; i++) 
					if(pArray->u.array.pMember[i] && mIsArray(pArray->u.array.pMember[i]) &&
					   (pArray->u.array.pMember[i]->u.array.rf_broken == 1))
					{
						  pSubArray = pArray->u.array.pMember[i];
						  goto loop;
					}
			default:
				return -EINVAL;
			}
			return orig_length;
		}
	}
	else if (length>=7 && strncmp(buffer, "verify ", 7)==0) {
		buffer+=7;
		length-=7;
		if (length>=6 && strncmp(buffer, "start ", 6)==0) {
			buffer+=6;
			length-=6;
			if (length>=1 && *buffer>='0') {
				iarray = *buffer-'1';
				if(iarray >= MAX_VDEVICE_PER_VBUS) return -EINVAL;

				pArray = _vbus_p->pVDevice[iarray];
				if (!pArray || (pArray->vf_online == 0)) return -EINVAL;
				
				if(pArray->VDeviceType != VD_RAID_1 && pArray->VDeviceType != VD_RAID_5)
					return -EINVAL;

				if (!(pArray->u.array.rf_need_rebuild ||
					pArray->u.array.rf_rebuilding ||
					pArray->u.array.rf_verifying ||
					pArray->u.array.rf_initializing))
				{
					spin_lock_irqsave(&pAdapter->driver_lock, flags);
					pArray->u.array.RebuildSectors = 0;
					hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pAdapter, pArray, VERIFY_R);
					spin_unlock_irqrestore(&pAdapter->driver_lock, flags);
				}
				return orig_length;
			}
		}
		else if (length>=5 && strncmp(buffer, "stop ", 5)==0) {
			buffer+=5;
			length-=5;
			if (length>=1 && *buffer>='0') {
				iarray = *buffer-'1';
				if(iarray >= MAX_VDEVICE_PER_VBUS) return -EINVAL;

				pArray = _vbus_p->pVDevice[iarray];
				if (!pArray || (pArray->vf_online == 0)) return -EINVAL;
				if(pArray->u.array.rf_verifying) {
					spin_lock_irqsave(&pAdapter->driver_lock, flags);
					pArray->u.array.rf_abort_rebuild = 1;
					spin_unlock_irqrestore(&pAdapter->driver_lock, flags);
				}
				return orig_length;
			}
		}
	}
#endif
	return -EINVAL;
}

int Kernel_DeviceIoControl(_VBUS_ARG
							DWORD dwIoControlCode,       	/* operation control code */
							PVOID lpInBuffer,            	/* input data buffer */
							DWORD nInBufferSize,         	/* size of input data buffer */
							PVOID lpOutBuffer,           	/* output data buffer */
							DWORD nOutBufferSize,        	/* size of output data buffer */
							PDWORD lpBytesReturned      	/* byte count */
						);

int hpt_set_info(int hostno,char *buffer,int length)
{
	int retval;
#ifdef SUPPORT_IOCTL
	PUCHAR ke_area;
	int err;
	DWORD dwRet;
	PHPT_IOCTL_PARAM32 piop=(PHPT_IOCTL_PARAM32)buffer;
#endif
	
	IAL_ADAPTER_T *pAdapter = gIal_Adapter;
	PVBus _vbus_p = &pAdapter->VBus;

	if (length >= 4) {
		if (strncmp(buffer,"hpt ",4) == 0) {
			do{
				if (pAdapter->host->host_no == hostno) break;
				pAdapter = pAdapter->next;
			}
			while(pAdapter != 0);
			HPT_ASSERT(pAdapter);
			return (retval = hpt_set_asc_info(pAdapter, buffer+4, length-4)) >= 0? retval : -EINVAL;
		}

#ifdef SUPPORT_IOCTL
		if (piop->Magic == HPT_IOCTL_MAGIC) {
			MV_PRINTD("ioctl=%d in=%x len=%d out=%x len=%d\n", 
				piop->dwIoControlCode,
				piop->lpInBuffer,
				piop->nInBufferSize,
				piop->lpOutBuffer,
				piop->nOutBufferSize);
			/* 
			 * verify user buffer.
			 */
			if (piop->nInBufferSize) {
				if (hpt_verify_area(VERIFY_READ, (void*)(ULONG_PTR)piop->lpInBuffer, piop->nInBufferSize)) {
					MV_PRINTD("Got bad user address.\n");
					return -EINVAL;
				}
			}
			if (piop->nOutBufferSize) {
				if (hpt_verify_area(VERIFY_WRITE, (void*)(ULONG_PTR)piop->lpOutBuffer, piop->nOutBufferSize)) {
					MV_PRINTD("Got bad user address.\n");
					return -EINVAL;
				}
			}
			if (piop->lpBytesReturned) {
				if (hpt_verify_area(VERIFY_WRITE, (void*)(ULONG_PTR)piop->lpBytesReturned, sizeof(DWORD))) {
					MV_PRINTD("Got bad user address.\n");
					return -EINVAL;
				}
			}
			/*
			 * map buffer to kernel.
			 */
			if (piop->nInBufferSize+piop->nOutBufferSize > 0x10000) {
				MV_PRINTD("User buffer too large\n");
				return -EINVAL;
			}
			ke_area = kmalloc(piop->nInBufferSize+piop->nOutBufferSize, GFP_ATOMIC);
			if (ke_area == NULL) {
				hpt_printk("Couldn't allocate kernel mem.\n");
				return -EINVAL;
			}
			if (piop->nInBufferSize)
				if (copy_from_user(ke_area, (void*)(ULONG_PTR)piop->lpInBuffer, piop->nInBufferSize)) {
					kfree(ke_area);
					hpt_printk("copy_from_user error\n");
					return -EINVAL;
				}
			/*
			 * call kernel handler.
			 */
			err = Kernel_DeviceIoControl(_VBUS_P piop->dwIoControlCode, ke_area, piop->nInBufferSize,
					ke_area + piop->nInBufferSize, piop->nOutBufferSize, &dwRet);    
			if (err==0) {
				if (piop->nOutBufferSize)
					if(copy_to_user((void*)(ULONG_PTR)piop->lpOutBuffer, 
							ke_area + piop->nInBufferSize, 
							piop->nOutBufferSize)){
						kfree(ke_area);
						hpt_printk("copy_to_user 1 error\n");
						return -EINVAL;
						}
				if (piop->lpBytesReturned)
					if (copy_to_user((void*)(ULONG_PTR)piop->lpBytesReturned, &dwRet, sizeof(DWORD))) {
						kfree(ke_area);
						hpt_printk("copy_to_user 2 error\n");
						return -EINVAL;
					}
				kfree(ke_area);
				return length;
			}
			else {
				/*hpt_printk("Kernel_ioctl(): return %d\n", err); */
			}
			
			kfree(ke_area);
			return -EINVAL;
			
		} else {
			hpt_printk("Wrong signature: %x\n", piop->Magic);
			return -EINVAL;
		}
#endif
	}
	return -EINVAL;
}

/*******************************************************************************
 * hptmv_proc_info
 * Parameters:
 *  *buffer: I/O buffer
 *  **start: if inout == FALSE pointer into buffer where user read should start
 *  offset: current offset into a /proc/scsi/hptmv/[0...] file
 *  length: length of buffer
 *  hostno: Scsi_Host host_no
 *  inout: TRUE - user is writing; FALSE - user is reading
 * Returns:
 *   Length of data
 *******************************************************************************/
int hptmv_proc_info(char *buffer,char **start,off_t offset,int length,   
				   int hostno,int inout)
{
	KdPrintI(("hpt3xx_ProcInfo() length %d ha %d offs %d inout %d\n",
		length,hostno,(int)offset,inout));

	if (inout)
		return(hpt_set_info(hostno,buffer,length));
	else
		return(hpt_get_info(hostno,buffer,start,offset,length));
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
int hptmv_proc_info26(struct Scsi_Host *host, char *buffer, char **start, 
			off_t offset, int length, int inout)
{
	return hptmv_proc_info(buffer, start, offset, length, host->host_no, inout);
}
#endif

int hptmv_scsi_ioctl_handle(Scsi_Device *dev, int cmd, void *arg)  
{
	unsigned char		*buff=NULL;
	unsigned short	copydata=0, offset=0;
	int err;
	DWORD			dwRet;
	IDE_PASS_THROUGH_HEADER ide_passthrough_header;
	HPT_IOCTL_PARAM	ioctl_param;
	IAL_ADAPTER_T *  pAdapter = HOSTDATA(dev->host);
	PVBus		pVBus = &pAdapter->VBus;
	PVDevice		pVDev = pVBus->pVDevice[dev->id];

	_VBUS_INST(pVBus)

			
	if (!pVDev || !pVDev->vf_online) {
		KdPrint(("RR182x: Device %d is offline. \n", dev->id));
		return -EINVAL;
	}

	/*We only support S.M.A.R.T fucntion on single disk, not array*/
	if (pVDev->VDeviceType != VD_SINGLE_DISK)  {
		KdPrint(("RR182X: Device %d is a RAID Array, Not supported. \n", dev->id));
		return -EINVAL;
	}

	/* Send the ioctl of the ide_pass_through_command */
	memset(&ioctl_param, 0, sizeof(ioctl_param));
	
	ide_passthrough_header.idDisk=VDEV_TO_ID(pVDev);
	if ( copy_from_user(&ide_passthrough_header.bFeaturesReg, arg+2, 1)
		||copy_from_user(&ide_passthrough_header.bSectorCountReg, arg+3, 1)
		||copy_from_user(&ide_passthrough_header.bLbaLowReg, arg+1, 1)
		||copy_from_user(&ide_passthrough_header.bCommandReg, arg, 1) )
		goto invalid;
	ide_passthrough_header.bLbaMidReg=0x4f;
	ide_passthrough_header.bLbaHighReg=0xc2;
	ide_passthrough_header.nSectors=1;
	ide_passthrough_header.protocol=0;

	switch ( cmd ) {
	/* CMD: HDIO_GET_IDENTITY */
	case 0x030d:
		offset=0;
		ide_passthrough_header.bCommandReg = 0xec;
		ide_passthrough_header.protocol=IO_COMMAND_READ;
		ide_passthrough_header.bSectorCountReg=1;
		copydata=512;
		ioctl_param.nInBufferSize=sizeof(IDE_PASS_THROUGH_HEADER);
		ioctl_param.nOutBufferSize=copydata+sizeof(IDE_PASS_THROUGH_HEADER);
	break;
	/* CMD: HDIO_DRIVER_CMD */
	case 0x031f:
		offset=4;
		switch ( ide_passthrough_header.bCommandReg ) {
		/* CHECK_POWER_MODE */
		case 0xe5 :
			ide_passthrough_header.protocol=IO_COMMAND_READ;
			/* Though smart monitor only copies one byte, error appears here if set to 1 */
			copydata=512;
		break;
		/* IDENTITY */
		case 0xec :
			ide_passthrough_header.protocol=IO_COMMAND_READ;
			copydata=512;
		break;
		/* PIDENTITY */
		case 0xa1 :
			ide_passthrough_header.protocol=IO_COMMAND_READ;
			copydata=512;
		break;
		/* ATA_SMART_COMMAND */
		case 0xb0:
			switch ( ide_passthrough_header.bFeaturesReg ) {
			/* READ_VALUES */
			case 0xd0 :
				ide_passthrough_header.protocol=IO_COMMAND_READ;
				copydata=512;
			break;
			/* READ_THRESHOLDS */
			case 0xd1 :
				ide_passthrough_header.protocol=IO_COMMAND_READ;
				copydata=512;
			break;
			/* READ_LOG */
			case 0xd5 :
				ide_passthrough_header.protocol=IO_COMMAND_READ;
				copydata=512;
			break;
			/* STATUS, ENABLE, DISABLE, AUTO_OFFLINE, AUTO_SAVE,
			IMMEDIATE_OFFLINE */
			}
		break;
		}
		ioctl_param.nInBufferSize=sizeof(IDE_PASS_THROUGH_HEADER);
		ioctl_param.nOutBufferSize=copydata+sizeof(IDE_PASS_THROUGH_HEADER);
	break;
	/* HDIO_DRIVER_TASK */
	case 0x031e:
		offset=7;
		if ( copy_from_user(&ide_passthrough_header.bFeaturesReg, arg+1, 1) )
			goto invalid;
		switch ( ide_passthrough_header.bFeaturesReg ){
		/* STATUS_CHECK */
		case 0xda:
			ioctl_param.nInBufferSize=sizeof(IDE_PASS_THROUGH_HEADER);
			ioctl_param.nOutBufferSize=sizeof(IDE_PASS_THROUGH_HEADER);
		break;
		}
	break;
	/* CMD: HDIO_DRIVER_TASKFILE */
	case 0x031d:
		if ( copy_from_user(&ide_passthrough_header.bCommandReg, arg+7, 1)
			||copy_from_user(&ide_passthrough_header.bSectorCountReg, arg+2, 1)
			||copy_from_user(&ide_passthrough_header.bLbaLowReg, arg+3, 1)
			||copy_from_user(&ide_passthrough_header.bDriveHeadReg, arg+6, 1)
			||copy_from_user(&ide_passthrough_header.bFeaturesReg, arg+1, 1) )
			goto invalid;
		ide_passthrough_header.protocol=IO_COMMAND_WRITE;
		switch ( ide_passthrough_header.bCommandReg ){
			/* SMART_ATA_COMMAND */
			case 0xb0 :
				switch ( ide_passthrough_header.bFeaturesReg ){
				/* WRITELOG */
				case 0xd6:
					copydata=512;
					ioctl_param.nInBufferSize=copydata+sizeof(IDE_PASS_THROUGH_HEADER);
					ioctl_param.nOutBufferSize=sizeof(IDE_PASS_THROUGH_HEADER);
				break;
				}
			break;
		}
	break;
	default:
		KdPrint(("RR182x: Unknown command %x!\n",cmd));
		return -EINVAL;
	}

	buff=kmalloc(copydata+2*sizeof(IDE_PASS_THROUGH_HEADER),GFP_ATOMIC);
	if ( !buff ) {
		hpt_printk("RR182x: No sufficient memory! \n");
		return -EINVAL;
	}
	memset(buff, 0, copydata+2*sizeof(IDE_PASS_THROUGH_HEADER));

	ioctl_param.dwIoControlCode=HPT_IOCTL_IDE_PASS_THROUGH;
	ioctl_param.lpInBuffer= buff;
	ioctl_param.lpOutBuffer=buff+sizeof(IDE_PASS_THROUGH_HEADER);
	
	if ( ide_passthrough_header.protocol==IO_COMMAND_WRITE && copydata )
		if ( (hpt_verify_area(VERIFY_READ, (void *)((ULONG_PTR)arg+40), copydata)) || 
				(copy_from_user(buff+sizeof(IDE_PASS_THROUGH_HEADER), arg+40, copydata))) {
			KdPrint(("Got bad user address, read user address fail. \n"));	
			goto invalid;
		}
	memcpy(buff, &ide_passthrough_header, sizeof(IDE_PASS_THROUGH_HEADER));

	err = Kernel_DeviceIoControl(_VBUS_P ioctl_param.dwIoControlCode, ioctl_param.lpInBuffer, 
		ioctl_param.nInBufferSize, ioctl_param.lpOutBuffer, ioctl_param.nOutBufferSize, &dwRet);    

	if (err) {
		KdPrint(("RR182x: hpt_scsi_ioctl return failed. \n"));
		goto invalid;
	}

	switch (cmd){
	case 0x031d:
		if( copy_to_user(arg, buff+sizeof(IDE_PASS_THROUGH_HEADER)+7, 1)
			||copy_to_user(arg+1, buff+sizeof(IDE_PASS_THROUGH_HEADER)+1, 6) )
			goto invalid;
	break;
	case 0x031f:
		if( copy_to_user(arg+1, buff+sizeof(IDE_PASS_THROUGH_HEADER)+1, 7) )
			goto invalid;
	break;
	default:
		if( copy_to_user(arg, buff+sizeof(IDE_PASS_THROUGH_HEADER)+7, 1)
			||copy_to_user(arg+1, buff+sizeof(IDE_PASS_THROUGH_HEADER)+3, 1)
			||copy_to_user(arg+2, buff+sizeof(IDE_PASS_THROUGH_HEADER)+2, 1) )
			goto invalid;
		/* according to Documentation/ioctl/hdio.txt (doc/hdio.txt) arg[3] undefined while ouput*/
	break;
	} 
	if( ide_passthrough_header.protocol==IO_COMMAND_READ && copydata )
		if( (hpt_verify_area(VERIFY_WRITE, (void * )(ULONG_PTR)(arg+offset), copydata)) ||
			(copy_to_user(arg+offset,buff+2*sizeof(IDE_PASS_THROUGH_HEADER), copydata))) {
			KdPrint(("Got bad user address, write user address  fail. \n"));
			goto invalid;
		}
	kfree(buff);
	return err;
invalid:
	kfree(buff);
	return -EINVAL;

}

