/*
 * RocketRAID 182x Controller Device Driver for Linux
 * Copyright (C) 2003-2004 HighPoint Technologies, Inc. All Rights Reserved.
 */
 
#include <linux/config.h>
#include <linux/version.h>

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)) && defined(MODVERSIONS)
#include <linux/modversions.h>
#endif

#include <linux/sched.h>
#include <linux/module.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#include <linux/blk.h>
#else
#include <linux/moduleparam.h>
#endif

MODULE_AUTHOR ("HighPoint Technologies, Inc.");
MODULE_DESCRIPTION ("RocketRAID 182x SATA Controller driver");

#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/major.h>
#include <asm/types.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/pci.h>
#include <linux/wait.h>
#include <linux/errno.h>
#include <linux/genhd.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>
#include <linux/kdev_t.h>
#include <linux/timer.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) /* 0x20300 */
	#include <asm/spinlock.h>
#else
	#include <linux/spinlock.h>
#endif
#if LINUX_VERSION_CODE == KERNEL_VERSION(2,2,16) 
#include <linux/kcomp.h>
#endif
#include <linux/kernel.h>
#include <linux/smp_lock.h>
#include <asm/dma.h>
#include <asm/system.h>
#include <asm/byteorder.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
#include <linux/reboot.h>
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
#include <linux/nmi.h>
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
#include <asm/div64.h>
#else
#define do_div(n, b) ({ int _res; _res = (n) % (b); (n) /= (b); _res; })
#endif

#if LINUX_VERSION_CODE == KERNEL_VERSION(2,2,14)
#define set_mb(var, value) do { xchg(&var, value); } while (0)

#define __set_current_state(state_value)			\
	do { current->state = (state_value); } while (0)
#ifdef CONFIG_SMP
#define set_current_state(state_value)		\
	set_mb(current->state, (state_value))
#else
#define set_current_state(state_value)		\
	__set_current_state(state_value)
#endif
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
#define pci_map_sg(pdev, sg_list, nseg, direction) (nseg)
#define pci_unmap_sg(pdev, sg_list, nseg, direction)
#define pci_map_single(pdev,buf,size,direction) (virt_to_bus(buf))
#define pci_unmap_single(pdev,bus_addr,size,direction)
#define sg_dma_address(sg) (virt_to_bus((sg)->address))
#define sg_dma_len(sg) ((sg)->length)
#define del_timer_sync(t) del_timer(t)
#endif

/*Deprecated since 2.6.11, defined it here*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
#define scsi_to_pci_dma_dir(scsi_dir) ((int)(scsi_dir))
#endif

/*verify_area will be deprecated since 2.6.11*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
#define hpt_verify_area(type, addr, size) (!access_ok((type), (addr), (size)))
#else
#define hpt_verify_area(type, addr, size) (verify_area((type), (addr), (size)))
#endif

#if defined(__MODULE_KERNEL_i686) && defined(__MODULE_KERNEL_athlon) && (__MODULE_KERNEL_i686==__MODULE_KERNEL_athlon)
#error "__MODULE_KERNEL_i686 and __MODULE_KERNEL_athlon is mistakely defined"
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#include "scsi.h"
#include <scsi/scsi_ioctl.h>
#else
#include <linux/blkdev.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi.h>
typedef struct scsi_host_template Scsi_Host_Template;
typedef struct scsi_device Scsi_Device;
typedef struct scsi_cmnd Scsi_Cmnd;
typedef struct scsi_request Scsi_Request;
typedef struct scsi_pointer Scsi_Pointer;
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,4)
#include "hosts.h"
#else
#include <scsi/scsi_host.h>
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#include "sd.h"
#endif

#include "mvOs.h"
#include "global.h"
#include "access601.h"

#define DRIVER_NAME "RocketRAID 182x SATA Controller driver"
#define PROC_DIR_NAME "hptmv"

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,3,18)

typedef ULONG_PTR dma_addr_t;

static inline void *pci_alloc_consistent(struct pci_dev *dev,int size,
										 dma_addr_t *dmahandle) {
   void * ptr = kmalloc(size, GFP_ATOMIC); 
   if(ptr){     
	  *dmahandle = virt_to_bus(ptr);
   }
   return ptr;
}

#define pci_free_consistent(a,size,address,dmahandle) kfree(address)

static __inline__ void list_del_init(struct list_head *entry)
{
	__list_del(entry->prev, entry->next);
	INIT_LIST_HEAD(entry); 
}
#endif

struct hpt_scsi_pointer {
	char * ptr;
	int mapped;
	int sgcnt;
	dma_addr_t dma_handle;
};

#define HPT_SCP(SCpnt) ((struct hpt_scsi_pointer *)&(SCpnt)->SCp)

#define hpt_printk printk

#if DBG
extern int hpt_dbg_level;
#define PRINT_CMD   print_command
#define MV_PRINTD(x...) do{ if (hpt_dbg_level>2) printk(x); }while(0)
#else
#define PRINT_CMD(x...)
#define MV_PRINTD(x...)
#endif

#define REQUESTS_ARRAY_SIZE			(9 * MV_EDMA_REQUEST_QUEUE_SIZE) /* 9 K bytes */
#define RESPONSES_ARRAY_SIZE		(12 * MV_EDMA_RESPONSE_QUEUE_SIZE) /* 3 K bytes */

#define PRD_ENTRIES_PER_CMD         (MAX_SG_DESCRIPTORS+1)
#define PRD_ENTRIES_SIZE            (MV_EDMA_PRD_ENTRY_SIZE*PRD_ENTRIES_PER_CMD)
#define PRD_TABLES_FOR_VBUS         (MV_SATA_CHANNELS_NUM*MV_EDMA_QUEUE_LENGTH)

typedef enum _SataEvent
{
	SATA_EVENT_NO_CHANGE = 0,
	SATA_EVENT_CHANNEL_CONNECTED,
	SATA_EVENT_CHANNEL_DISCONNECTED
} SATA_EVENT;

typedef struct _MV_CHANNEL
{
	unsigned int		maxUltraDmaModeSupported;
	unsigned int		maxDmaModeSupported;
	unsigned int		maxPioModeSupported;
	MV_BOOLEAN			online;
	MV_BOOLEAN			writeCacheSupported;
	MV_BOOLEAN			writeCacheEnabled;
	MV_BOOLEAN			readAheadSupported;
	MV_BOOLEAN			readAheadEnabled;
	MV_U8				queueDepth;
} MV_CHANNEL;

typedef struct IALAdapter 
{
	MV_SATA_ADAPTER     mvSataAdapter;
	struct Scsi_Host    *host;
	MV_CHANNEL          mvChannel[MV_SATA_CHANNELS_NUM];
	u8					*requestsArrayBaseAddr;
	u8					*requestsArrayBaseAlignedAddr;
	dma_addr_t			requestsArrayBaseDmaAddr;
	dma_addr_t			requestsArrayBaseDmaAlignedAddr;
	u8					*responsesArrayBaseAddr;
	u8					*responsesArrayBaseAlignedAddr;
	dma_addr_t			responsesArrayBaseDmaAddr;
	dma_addr_t			responsesArrayBaseDmaAlignedAddr;
	SATA_EVENT			sataEvents[MV_SATA_CHANNELS_NUM];
	struct timer_list   event_timer_connect;
	struct timer_list   event_timer_disconnect;

	struct _VBus        VBus;
	struct _VDevice     VDevices[MV_SATA_CHANNELS_NUM];
	PCommand            pCommandBlocks;
	void *				pFreePRDLink;
	Scsi_Cmnd           *pending_Q;	/* pending commands to issue */
	struct IALAdapter   *next;

	u8   			    outstandingCommands;
	UCHAR               status;
	UCHAR               ver_601;
	UCHAR               beeping;

	spinlock_t          driver_lock;
	atomic_t            DPC_Request;
	struct pci_dev *pci_dev;
	
	unsigned long last_active_jiffy;

}IAL_ADAPTER_T;

#define HOSTDATA(host)  ((IAL_ADAPTER_T *)&host->hostdata)

int hptmv_detect (Scsi_Host_Template *);
int hptmv_release (struct Scsi_Host *); 
int hptmv_queuecommand (Scsi_Cmnd *, void (*done) (Scsi_Cmnd *));
int hptmv_abort (Scsi_Cmnd *);
int hptmv_reset (Scsi_Cmnd *);
int hptmv_proc_info(char *buffer, char **start, off_t offset,
					int length, int inode, int inout);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
int hptmv_proc_info26(struct Scsi_Host *, char *, char **, off_t, int, int);
#endif

#if BITS_PER_LONG > 32
#define pci64_dma_hi32(a) (u32)((u64)(a)>>32)
#define pci64_dma_lo32(a) (u32)(u64)(a)
#else
#define pci64_dma_hi32(a) 0
#define pci64_dma_lo32(a) (a)
#endif	/* BITS_PER_LONG */

/* vmalloc have problem on some SMP kernels? */
#if 1
#define VMALLOC(size) kmalloc(size, GFP_ATOMIC)
#define VFREE(p) kfree(p)
#else
#define VMALLOC(size) vmalloc(size)
#define VFREE(p) vfree(p)
#endif

#ifndef DECLARE_MUTEX_LOCKED
#define DECLARE_MUTEX_LOCKED(sem) struct semaphore sem = MUTEX_LOCKED
#endif

#if LINUX_VERSION_CODE  < KERNEL_VERSION(2,4,4) || LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0) 
#define scsi_set_pci_device(pshost, pcidev)	do { ;} while (0)
#endif

#define DUPLICATE      0
#define INITIALIZE     1
#define REBUILD_PARITY 2
#define VERIFY_R         3
typedef void (*HPT_DPC)(void*, void*, UCHAR);

extern IAL_ADAPTER_T  *gIal_Adapter;

int hpt_queue_dpc(HPT_DPC dpc, void *arg1, void *arg2, UCHAR flags);
void hpt_rebuild_data_block(IAL_ADAPTER_T *pAdapter, PVDevice pArray, UCHAR flags);
PVOID AllocatePRDTable(IAL_ADAPTER_T *pAdapter);
void FreePRDTable(IAL_ADAPTER_T *pAdapter, PVOID PRDTable);
void init_vdev_params(IAL_ADAPTER_T *pAdapter, MV_U8 channel);
void fRegisterVdevice(IAL_ADAPTER_T *pAdapter);
void SetInquiryData(PINQUIRYDATA inquiryData, PVDevice pVDev);
int fMvResetBus(PVBus _vbus_p);
void FlushAdapter(IAL_ADAPTER_T *pAdapter);
void device_change(IAL_ADAPTER_T *pAdapter , MV_U8 channelIndex, int plugged);
int start_channel(IAL_ADAPTER_T *pAdapter, MV_U8 channelNum);
void hptmv_free_channel(IAL_ADAPTER_T *pAdapter, MV_U8 channelNum);
void Check_Idle_Call(IAL_ADAPTER_T *pAdapter);

#include "gui_lib.c"
#include "mv.c"
#include "mvlib.c"

