/*
 * RocketRAID 182x Controller Device Driver for Linux
 * Copyright (C) 2003-2004 HighPoint Technologies, Inc. All Rights Reserved.
 */

static int writeback = 1;
static int cachesize = 5;
static int flush_interval = 50;

#if defined(MODULE)
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
module_param(writeback, uint, 0);
module_param(cachesize, uint, 0);
module_param(flush_interval, uint, 0);
#else 
MODULE_PARM(writeback, "i");
MODULE_PARM(cachesize, "i");
MODULE_PARM(flush_interval, "i");
#endif
MODULE_PARM_DESC(writeback, "select RAID5 write-through/write-back");
MODULE_PARM_DESC(cachesize, "set cache memory size (MB)");
MODULE_PARM_DESC(flush_interval, "set cache flush interval");
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
static void hptmv_int_handler (int irq, void *dev_id, struct pt_regs *regs);
#else 
static irqreturn_t hptmv_int_handler (int irq, void *dev_id, struct pt_regs *regs);
#endif
static int hptmv_parse_identify_results(MV_SATA_CHANNEL *pMvSataChannel);
static int hptmv_allocate_edma_queues(IAL_ADAPTER_T *pAdapter);
static void hptmv_free_edma_queues(IAL_ADAPTER_T *pAdapter);
static void hptmv_handle_event_connect (unsigned long data);
static void hptmv_handle_event_disconnect(unsigned long data);
static int hptmv_init_channel(IAL_ADAPTER_T *pAdapter, MV_U8 channelNum);
static void hptmv_cleanup(void);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
#define sc_host(sc)    sc->device->host
#define sc_channel(sc) sc->device->channel
#define sc_target(sc)  sc->device->id
#define sc_lun(sc)     sc->device->lun
#else 
#define sc_host(sc)    sc->host
#define sc_channel(sc) sc->channel
#define sc_target(sc)  sc->target
#define sc_lun(sc)     sc->lun
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
/* notifier block to get a notify on system shutdown/halt/reboot */
static int hptmv_halt(struct notifier_block *nb, ulong event, void *buf);
static struct notifier_block mv_notifier = {
	hptmv_halt, NULL, 0
};
#endif

struct proc_dir_entry proc_scsi_hptmv = {
	0, 							/* inode number - ignore. */
	sizeof(PROC_DIR_NAME)-1, 	/* length of name */
	PROC_DIR_NAME,              /* file name */
	S_IFDIR|S_IRUGO|S_IXUGO,	/* */
	2							/* number of links */
};

typedef struct st_HPT_DPC {
	void (*dpc)(void *, void *, UCHAR);
	void *arg1;
	void *arg2;
	UCHAR flags;
} ST_HPT_DPC;

#define MAX_DPC 16

static spinlock_t DpcQueueLock = SPIN_LOCK_UNLOCKED;
static ST_HPT_DPC DpcQueue[MAX_DPC];
static int DpcQueue_First=0;
static int DpcQueue_Last = 0;

DECLARE_MUTEX_LOCKED(hpt_notify_wt);
struct semaphore * hpt_queue_ready = 0;
DECLARE_MUTEX_LOCKED(hpt_queue_wt);
struct task_struct *hpt_worker_thread_handle = 0;

volatile int stop_thread = 0;

IAL_ADAPTER_T  *gIal_Adapter = 0;
MV_SATA_CHANNEL	gMvSataChannels[MAX_VBUS][MV_SATA_CHANNELS_NUM];

#if DBG
int hpt_dbg_level = DEBUG_LEVEL;
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18) && defined(CONFIG_HIGHMEM)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
#define OS_KMAP_TYPE KM_BIO_SRC_IRQ
#else 
#define OS_KMAP_TYPE KM_BH_IRQ
#endif
void * HPTLIBAPI os_kmap_ptr(BUS_ADDR addr)
{
	struct page * page = (struct page *)(ULONG_PTR)(addr>>32);
	if (page)
		return (PageHighMem(page)? (char *)kmap_atomic(page, OS_KMAP_TYPE) : (char *)page_address(page))
			+ (addr & 0xffffffff);
	else
		return (void *)(ULONG_PTR)addr;
}

void HPTLIBAPI os_kunmap_ptr(void *ptr)
{
	if ((ULONG_PTR)ptr >= (ULONG_PTR)high_memory)
		kunmap_atomic(ptr, OS_KMAP_TYPE);
}
#else 
void * HPTLIBAPI os_kmap_ptr(BUS_ADDR addr) { return (void *)(ULONG_PTR)addr; }
void HPTLIBAPI os_kunmap_ptr(void *ptr) {}
#endif

static int scsicmd_buf_get(Scsi_Cmnd *cmd, void **pbuf)
{
	unsigned int buflen;

	if (cmd->use_sg) {
		struct scatterlist *sg = (struct scatterlist *) cmd->request_buffer;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
		*pbuf = kmap_atomic(sg->page, OS_KMAP_TYPE) + sg->offset;
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18) && defined(CONFIG_HIGHMEM)
		if (sg->page)
			sg->address = kmap_atomic(sg->page, OS_KMAP_TYPE) + sg->offset;
		*pbuf = sg->address;
#else 
		*pbuf = sg->address;
#endif
		buflen = sg->length;
	} else {
		*pbuf = cmd->request_buffer;
		buflen = cmd->request_bufflen;
	}

	return buflen;
}

static inline void scsicmd_buf_put(struct scsi_cmnd *cmd, void *buf)
{
	if (cmd->use_sg) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
		kunmap_atomic(buf, OS_KMAP_TYPE);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18) && defined(CONFIG_HIGHMEM)
		struct scatterlist *sg = (struct scatterlist *) cmd->request_buffer;
		if (sg->page)
			kunmap_atomic(buf, OS_KMAP_TYPE);
#endif
	}
}

/***************************************************************************
 * Function:     hpt_queue_dpc
 * Description:  Queue a DPC called by worker thread. It may be called in an ISR.
 * Returns:      None
 ***************************************************************************/
int hpt_queue_dpc(HPT_DPC dpc, void *arg1, void *arg2, UCHAR flags)
{
	int p;
	unsigned long Lock_Flag;
	/*
	 * driver_lock shall be locked before call this function
	 */
	spin_lock_irqsave(&DpcQueueLock, Lock_Flag);
	p = (DpcQueue_Last + 1) % MAX_DPC;
	if (p==DpcQueue_First) {
		KdPrintW(("DPC Queue full!\n"));
		return -1;
	}
	DpcQueue[DpcQueue_Last].dpc = dpc;
	DpcQueue[DpcQueue_Last].arg1 = arg1;
	DpcQueue[DpcQueue_Last].arg2 = arg2;
	DpcQueue[DpcQueue_Last].flags = flags;
	DpcQueue_Last = p;
	up(hpt_queue_ready);
	spin_unlock_irqrestore(&DpcQueueLock, Lock_Flag);
	return 0;
}

#ifdef SUPPORT_ARRAY
void hpt_worker_thread(void *arg);
static void launch_worker_thread(void)
{
	kernel_thread((int (*)(void *))hpt_worker_thread, (void*)0, 0);
	down(&hpt_notify_wt);
}
#endif

/****************************************************************
 *	Name: hptmv_detect
 *	Description:	Detect and initialize our boards.
 *	Parameters:    	tpnt - Pointer to SCSI host template structure.
 *	Returns:       	Number of adapters installed.
 ****************************************************************/
int hptmv_detect (Scsi_Host_Template *tpnt)
{
	struct Scsi_Host    *pshost;
	MV_SATA_ADAPTER     *pMvSataAdapter;
	IAL_ADAPTER_T       *pAdapter, *pAdapTemp = 0;
	int                 num_hosts=0, i;
	MV_U8               channel;
	struct pci_dev      *pcidev = NULL;
	MV_U16              mvSataDevices[3] = {MV_SATA_DEVICE_ID_5080, MV_SATA_DEVICE_ID_5081, 0};
	int                 index;
	PVBus               _vbus_p;
	PVDevice            pVDev;

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,3,18)
	DWORD v;
#endif

	printk("%s\nVersion %s(%s/%s)\n",
		DRIVER_NAME, DRIVER_VERSION,  __DATE__, __TIME__);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	if (!pci_present())
	{
		printk ("RR182x: PCI BIOS not present\n");
		return 0;
	}
#endif

	index = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	spin_unlock_irq (&io_request_lock);
#endif

	while (mvSataDevices[index]) {

		pcidev = NULL;
		while ((pcidev = pci_find_device (MV_SATA_VENDOR_ID,
										  mvSataDevices[index], pcidev)) != NULL)
		{
			MV_PRINTD("RR182x: PCI device found, BAR0=%lx\n",
					  pci_resource_start(pcidev,0));

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,18)
			if (pci_enable_device(pcidev))
			{
				MV_PRINTD("RR182x: failed to enable the pci device\n");
				continue;
			}
#endif
			pci_set_master(pcidev);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18)
			/* enable 64-bit DMA if possible */
			if (pci_set_dma_mask(pcidev, 0xffffffffffffffffULL))
				if (pci_set_dma_mask(pcidev, 0xffffffffUL)) {
					printk("RR182x: failed to set DMA mask\n");
					continue;
				}
#endif

			pshost = scsi_register(tpnt, sizeof(IAL_ADAPTER_T));
			if (pshost==NULL)
			{
				hpt_printk("RR182x: Scsi_Host allocation failed\n");
				continue;
			}

			pAdapter = HOSTDATA(pshost);
			pAdapter->driver_lock = SPIN_LOCK_UNLOCKED;
			atomic_set(&pAdapter->DPC_Request, 0);
			pAdapter->next = 0;

			_vbus_p  = &pAdapter->VBus;
			pAdapter->host = pshost;

			pAdapter->pci_dev = pcidev;
			pAdapter->outstandingCommands = 0;
			pAdapter->pending_Q = NULL;
			pMvSataAdapter = &(pAdapter->mvSataAdapter);
			_vbus_p->OsExt = (void *)pAdapter;
			pMvSataAdapter->IALData = pAdapter;

			if (hptmv_allocate_edma_queues(pAdapter))
			{
				hpt_printk("RR182x: Failed to allocate memory for EDMA queues\n");
				scsi_unregister(pshost);
				continue;
			}

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,3,18)
			pcibios_read_config_dword(pcidev->bus->number,pcidev->devfn,0x10,(unsigned int *)&v);
			pshost->base = (void*)(v & ~0xf);
#else 
			pshost->base = pci_resource_start(pcidev, 0);
#endif
			/* also map EPROM address */
			pMvSataAdapter->adapterIoBaseAddress =
				(MV_BUS_ADDR_T)ioremap((ULONG_PTR)pshost->base, MV_SATA_PCI_BAR0_SPACE_SIZE+0x40000);
			if (!pMvSataAdapter->adapterIoBaseAddress)
			{
				hpt_printk("RR182x: Failed to remap memory space\n");
				hptmv_free_edma_queues(pAdapter);
				scsi_unregister(pshost);
				continue;
			}
			else
			{
				MV_PRINTD("RR182x: io base address %p\n",
						  pMvSataAdapter->adapterIoBaseAddress);
			}

			pMvSataAdapter->adapterId = num_hosts;
			/* get the revision ID */
			if (pci_read_config_byte(pcidev, PCI_REVISION_ID, &pMvSataAdapter->pciConfigRevisionId))
			{
				printk(KERN_WARNING "RR182x: Failed to get revision id.\n");
unregister:
				iounmap((void *)pMvSataAdapter->adapterIoBaseAddress);
				hptmv_free_edma_queues(pAdapter);
				scsi_unregister(pshost);
				continue;
			}
			pMvSataAdapter->pciConfigDeviceId = mvSataDevices[index];

			pMvSataAdapter->intCoalThre[0]= 1;
			pMvSataAdapter->intCoalThre[1]= 1;
			pMvSataAdapter->intTimeThre[0] = 1;
			pMvSataAdapter->intTimeThre[1] = 1;
			pMvSataAdapter->pciCommand = 0x0107E371;
			pMvSataAdapter->pciSerrMask = 0xd77fe6ul;
			pMvSataAdapter->pciInterruptMask = 0xd77fe6ul;
			pMvSataAdapter->mvSataEventNotify = mvSataEventNotify;

			if (mvSataInitAdapter(pMvSataAdapter) == MV_FALSE)
			{
				hpt_printk("RR182x[%d]: core failed to initialize the adapter\n",
						 pMvSataAdapter->adapterId);
				goto unregister;
			}
			pAdapter->ver_601 = pMvSataAdapter->pcbVersion;

			set_fail_leds(pMvSataAdapter, 0);
			scsi_set_pci_device(pshost, pcidev);
			pshost->irq = pcidev->irq;

			/* each SATA channel will emulate scsi channel */
			pshost->max_id = MV_SATA_CHANNELS_NUM;
			pshost->max_lun = 1;
			pshost->max_channel = 0;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
			pshost->max_cmd_len = 16; /* for CONFIG_LBD */
#endif

			/* enable device interrupts */
			if (request_irq(pshost->irq, hptmv_int_handler,
							 SA_SHIRQ, PROC_DIR_NAME, pAdapter) < 0)
			{
				if (request_irq(pshost->irq, hptmv_int_handler,
							(SA_INTERRUPT | SA_SHIRQ), PROC_DIR_NAME,
							pAdapter) < 0)
				{
					hpt_printk("RR182x[%d]: unable to allocate IRQ for controler\n",
							 pMvSataAdapter->adapterId);
					goto unregister;
				}
			}

			init_timer(&pAdapter->event_timer_connect);
			pAdapter->event_timer_connect.data = (unsigned long)pAdapter;
			pAdapter->event_timer_connect.function = hptmv_handle_event_connect;

			init_timer(&pAdapter->event_timer_disconnect);
			pAdapter->event_timer_disconnect.data = (unsigned long)pAdapter;
			pAdapter->event_timer_disconnect.function = hptmv_handle_event_disconnect;

			/* setup command blocks */
			MV_PRINTD("Allocate command blocks\n");
			_vbus_(pFreeCommands) = 0;
			pAdapter->pCommandBlocks = VMALLOC(sizeof(struct _Command) * MAX_COMMAND_BLOCKS_FOR_EACH_VBUS);
			MV_PRINTD("pCommandBlocks:%p\n",pAdapter->pCommandBlocks);
			if (!pAdapter->pCommandBlocks) {
				hpt_printk("insufficient memory\n");
				/*return 0;//FIXME */
			}
			for (i=0; i<MAX_COMMAND_BLOCKS_FOR_EACH_VBUS; i++)
				FreeCommand(_VBUS_P &(pAdapter->pCommandBlocks[i]));

			/* setup PRD Tables */
			MV_PRINTD("Allocate PRD Tables\n");
			pAdapter->pFreePRDLink = 0;
			{
				PUCHAR PRDTable;
				for (i=0; i<PRD_TABLES_FOR_VBUS; i++) {
					PRDTable = kmalloc(PRD_ENTRIES_SIZE, GFP_ATOMIC);
					if (!PRDTable) {
						hpt_printk("insufficient PRD Tables\n");
						/*return 0;//FIXME */
					}
					FreePRDTable(pAdapter, PRDTable);
				}
			}

			/* configure and start the connected channels*/
			for (channel = 0; channel < MV_SATA_CHANNELS_NUM; channel++)
			{
				pAdapter->mvChannel[channel].online = MV_FALSE;
				if (mvSataIsStorageDeviceConnected(pMvSataAdapter, channel) == MV_TRUE)
				{
					MV_PRINTD("RR182x[%d]: channel %d is connected\n",
							  pMvSataAdapter->adapterId, channel);

					if (hptmv_init_channel(pAdapter, channel) == 0)
					{
						if (mvSataConfigureChannel(pMvSataAdapter, channel) == MV_FALSE)
						{
							hpt_printk("RR182x[%d]: Failed to configure channel"
									 " %d\n",pMvSataAdapter->adapterId, channel);
							hptmv_free_channel(pAdapter, channel);
						}
						else
						{
							if (start_channel(pAdapter, channel))
							{
								hpt_printk("RR182x[%d]: Failed to start channel,"
										 " channel=%d\n",pMvSataAdapter->adapterId,
										 channel);
								hptmv_free_channel(pAdapter, channel);
							}
							pAdapter->mvChannel[channel].online = MV_TRUE;
						}
					}
				}
				MV_PRINTD("pAdapter->mvChannel[channel].online:%x, channel:%d\n",
					pAdapter->mvChannel[channel].online, channel);
			}

#ifdef SUPPORT_ARRAY
			for(i = MAX_ARRAY_DEVICE - 1; i >= 0; i--) {
				pVDev = ArrayTables(i);
				mArFreeArrayTable(pVDev);
			}
#endif

			/* rxl 2003-7-14 Initialize Devices */
			MV_PRINTD("Initialize Devices\n");
			for (channel = 0; channel < MV_SATA_CHANNELS_NUM; channel++) {
				MV_SATA_CHANNEL *pMvSataChannel = pMvSataAdapter->sataChannel[channel];
				if (pMvSataChannel) {
					init_vdev_params(pAdapter, channel);
					fCheckArray(&pAdapter->VDevices[channel].u.disk);
				}
			}
#ifdef SUPPORT_ARRAY
			CheckArrayCritical(_VBUS_P0);
#endif
			_vbus_p->nInstances = 1;
			fRegisterVdevice(pAdapter);

			for (channel=0;channel<MV_SATA_CHANNELS_NUM;channel++) {
				pVDev = _vbus_p->pVDevice[channel];
				if (pVDev && pVDev->vf_online)
					fCheckBootable(pVDev);
			}

#if defined(SUPPORT_ARRAY) && defined(_RAID5N_)
			num_raid5_pages = 0x100000/CACHE_PAGE_SIZE * cachesize;
			init_raid5_memory(_VBUS_P0);
			_vbus_(r5).enable_write_back = writeback;
			printk("RR182x: RAID5 write-back %s\n",
				_vbus_(r5).enable_write_back? "enabled" : "disabled");
#endif

			/* enable the adapter interrupts */
			mvSataUnmaskAdapterInterrupt(pMvSataAdapter);

			if(gIal_Adapter == 0){
				gIal_Adapter = pAdapter;
				pAdapTemp = gIal_Adapter;
			}
			else {
				pAdapTemp->next = pAdapter;
				pAdapTemp = pAdapter;
			}
			num_hosts++;

			MV_PRINTD("RR182x: %d adapters are found\n", num_hosts);
		}
		index++;
	}
	MV_PRINTD("num_hosts:%d\n", num_hosts);

	if(num_hosts)
	{
	/* launch worker thread */
#ifdef SUPPORT_ARRAY
		launch_worker_thread();
#endif

#ifndef __BOOT_KERNEL_BOOT
#ifdef SUPPORT_ARRAY
		for (pAdapTemp = gIal_Adapter; pAdapTemp; pAdapTemp = pAdapTemp->next) {
			unsigned long flags;
			_vbus_p   = &pAdapTemp->VBus;
			spin_lock_irqsave(&pAdapTemp->driver_lock, flags);
			for(i = 0; i < MAX_ARRAY_PER_VBUS; i++) {
				if ((pVDev=ArrayTables(i))->u.array.dArStamp &&
					pVDev->u.array.rf_need_rebuild && !pVDev->u.array.rf_rebuilding)
						hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pAdapTemp, pVDev,
							(UCHAR)((pVDev->u.array.CriticalMembers || pVDev->VDeviceType == VD_RAID_1)?
								DUPLICATE : REBUILD_PARITY));
			}
			spin_unlock_irqrestore(&pAdapTemp->driver_lock, flags);
		}
#endif
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
		register_reboot_notifier(&mv_notifier);
#endif
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	spin_lock_irq (&io_request_lock);
#endif

	return num_hosts;
}

int HPTLIBAPI fOsBuildSgl(_VBUS_ARG PCommand pCmd, FPSCAT_GATH pSg, int logical)
{
	Scsi_Cmnd *SCpnt = (Scsi_Cmnd *)pCmd->pOrgCommand;
	int       idx;
	struct scatterlist *sgList = (struct scatterlist *)SCpnt->request_buffer;

	if (logical) {
		if(SCpnt->use_sg){
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18) && defined(CONFIG_HIGHMEM)
			for (idx = 0; idx < SCpnt->use_sg; idx++) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
				pSg[idx].dSgAddress = (ULONG_PTR)sgList[idx].address;
				if (pSg[idx].dSgAddress) {
					pSg[idx].wSgSize = sgList[idx].length;
				}
				else 
#endif
				{
					/*
					 *  highmem address needs special handling. 
					 *  we store the page in high part of BUS_ADDR and offset in low part
					 *  and map the address to a actual pointer when needed.
					 */
					unsigned int length = sgList[idx].length;
					struct page *page = sgList[idx].page;
					HPT_ASSERT(sizeof(pSg[idx].dSgAddress)>4 && sizeof(void *)==4);
					pSg[idx].dSgAddress = ((u64)(u32)page<<32) | sgList[idx].offset;
					if (sgList[idx].offset + length > PAGE_SIZE) return 0;
					pSg[idx].wSgSize = length;
				}
				pSg[idx].wSgFlag = (idx==SCpnt->use_sg-1)? SG_FLAG_EOT : 0;
			}
#else 
			for (idx = 0; idx < SCpnt->use_sg; idx++) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
				pSg[idx].dSgAddress = (ULONG_PTR)page_address(sgList[idx].page) + sgList[idx].offset;
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18)
				pSg[idx].dSgAddress = (ULONG_PTR)sgList[idx].address;
				if (!pSg[idx].dSgAddress)
					pSg[idx].dSgAddress = (ULONG_PTR)page_address(sgList[idx].page) + sgList[idx].offset;
#else 
				pSg[idx].dSgAddress = (ULONG_PTR)sgList[idx].address;
#endif
				pSg[idx].wSgSize = sgList[idx].length;
				pSg[idx].wSgFlag = (idx==SCpnt->use_sg-1)? SG_FLAG_EOT : 0;
			}
#endif
		}
		else{
			pSg->dSgAddress = (ULONG_PTR)SCpnt->request_buffer;
			pSg->wSgSize = (USHORT)SCpnt->request_bufflen;
			pSg->wSgFlag = SG_FLAG_EOT;
		}
	}
	else {
		if (SCpnt->use_sg) {
			BUS_ADDR addr, last=0;
			unsigned int length;
			if (!HPT_SCP(SCpnt)->mapped) {
				HPT_SCP(SCpnt)->sgcnt = pci_map_sg(((IAL_ADAPTER_T *)HOSTDATA(sc_host(SCpnt)))->pci_dev, 
							sgList, SCpnt->use_sg, scsi_to_pci_dma_dir(SCpnt->sc_data_direction));
				HPT_ASSERT(HPT_SCP(SCpnt)->sgcnt <= MAX_SG_DESCRIPTORS);
				HPT_SCP(SCpnt)->mapped = 1;
			}
			pSg--;
			for (idx = 0; idx < HPT_SCP(SCpnt)->sgcnt ; idx++) {
				addr = sg_dma_address(&sgList[idx]);
				length = sg_dma_len(&sgList[idx]);
				/* merge the sg elements if possible */
				if (idx && last==addr && pSg->wSgSize &&
					pSg->wSgSize+length<=0x10000 && (addr & 0xffffffff)!=0) {
					pSg->wSgSize += length;
					last += length;
				}
				else {
					if (addr & 1) return FALSE;
					pSg++;
					pSg->dSgAddress = addr;
					pSg->wSgSize = length;
					last = addr + length;
				}
				if (pSg->wSgSize & 1) return FALSE;
				pSg->wSgFlag = (idx==(HPT_SCP(SCpnt)->sgcnt)-1)? SG_FLAG_EOT : 0;
			}
		}
		else {
			if (!HPT_SCP(SCpnt)->mapped)  {
				HPT_SCP(SCpnt)->dma_handle = pci_map_single(
					((IAL_ADAPTER_T *)HOSTDATA(sc_host(SCpnt)))->pci_dev, 
					SCpnt->request_buffer,
					SCpnt->request_bufflen,
					scsi_to_pci_dma_dir(SCpnt->sc_data_direction));
				HPT_SCP(SCpnt)->mapped = 1;
			}
			pSg->dSgAddress = HPT_SCP(SCpnt)->dma_handle;
			if ((pSg->dSgAddress & 0x1) ||(!pSg->dSgAddress))return FALSE;
			pSg->wSgSize = (USHORT)SCpnt->request_bufflen;
			if (pSg->wSgSize & 0x1) return FALSE;
			pSg->wSgFlag = SG_FLAG_EOT;
		}
	}
	return TRUE;
}

void Queue_SC(Scsi_Cmnd **SC_Q,Scsi_Cmnd *SCpnt)
{
	if(*SC_Q==NULL)
		HPT_SCP(SCpnt)->ptr = (char *)SCpnt;
	else
	{
		HPT_SCP(SCpnt)->ptr = HPT_SCP(*SC_Q)->ptr;
		HPT_SCP(*SC_Q)->ptr = (char *)SCpnt;
	}
	*SC_Q = SCpnt;
}

void Release_SC(Scsi_Cmnd **SC_Q,Scsi_Cmnd *SCpnt)
{
	Scsi_Cmnd *TmpSCpnt;
	TmpSCpnt = *SC_Q;

	if(HPT_SCP(SCpnt)->ptr == (char *)SCpnt)/*it means SCpnt is the last one in CURRCMDs*/
		*SC_Q = NULL;
	else
	{
		while(HPT_SCP(TmpSCpnt)->ptr != (char *)SCpnt){
			TmpSCpnt = (Scsi_Cmnd *)HPT_SCP(TmpSCpnt)->ptr;
		}
		HPT_SCP(TmpSCpnt)->ptr = HPT_SCP(SCpnt)->ptr;
		if(*SC_Q == SCpnt)
			*SC_Q = TmpSCpnt;
	}
}

/* We want to use OsSendCommand in CallAfterReturn, so it must use below prototype */
void HPTLIBAPI OsSendCommand (_VBUS_ARG Scsi_Cmnd *SCpnt);

/* call this function inside driver_lock */
void Check_Idle_Call(IAL_ADAPTER_T *pAdapter)
{
	_VBUS_INST(&pAdapter->VBus)
	if (mWaitingForIdle(_VBUS_P0)) {
		CheckIdleCall(_VBUS_P0);
		up(&hpt_queue_wt);
#ifdef SUPPORT_ARRAY
		{
			int i;
			PVDevice pArray;
			for(i = 0; i < MAX_ARRAY_PER_VBUS; i++){
				if ((pArray=ArrayTables(i))->u.array.dArStamp==0)
					continue;
				else if (pArray->u.array.rf_auto_rebuild) {
						hpt_printk("auto rebuild.\n");
						pArray->u.array.rf_auto_rebuild = 0;
						hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pAdapter, pArray, DUPLICATE);
				}
			}
		}
#endif
	}
	/* check pending_Q */
	while(pAdapter->pending_Q)
	{
		Scsi_Cmnd *SCpnt = (Scsi_Cmnd *)HPT_SCP(pAdapter->pending_Q)->ptr;
		Release_SC(&pAdapter->pending_Q, SCpnt);
		/*printk("release pending scmd %x host %x \n",SCpnt, sc_host(SCpnt)); */
		CallAfterReturn(_VBUS_P (DPC_PROC)OsSendCommand, SCpnt);
	}
}

void ScsiDone(Scsi_Cmnd *SCpnt)
{
	IAL_ADAPTER_T *pAdapter = HOSTDATA(sc_host(SCpnt)) ;

	SCpnt->scsi_done(SCpnt);

	if (--pAdapter->outstandingCommands == 0)
	{
		if (atomic_read(&pAdapter->DPC_Request) == 0)
			Check_Idle_Call(pAdapter);
	}
}

void HPTLIBAPI fOsCommandDone(_VBUS_ARG PCommand pCmd)
{
	Scsi_Cmnd *SCpnt = (Scsi_Cmnd *)pCmd->pOrgCommand;

	if (HPT_SCP(SCpnt)->mapped) {
		if (SCpnt->use_sg)
			pci_unmap_sg(
				HOSTDATA(sc_host(SCpnt))->pci_dev,
				(struct scatterlist *)SCpnt->request_buffer,
				SCpnt->use_sg,
				scsi_to_pci_dma_dir(SCpnt->sc_data_direction)
			);
		else
				pci_unmap_single(
				HOSTDATA(sc_host(SCpnt))->pci_dev,
				HPT_SCP(SCpnt)->dma_handle, 
				SCpnt->request_bufflen,
				scsi_to_pci_dma_dir(SCpnt->sc_data_direction)
			);
	}

	switch(pCmd->Result) {
	case RETURN_SUCCESS:
		MV_PRINTD("RETURN_SUCCESS\n");
		SCpnt->result = (DID_OK<<16);
		break;
	case RETURN_BAD_DEVICE:
		MV_PRINTD("RETURN_BAD_DEVICE");
		SCpnt->result = (DID_BAD_TARGET<<16);
		break;
	case RETURN_DEVICE_BUSY:
		MV_PRINTD("RETURN_DEVICE_BUSY");
		SCpnt->result = (DID_BUS_BUSY<<16);
		break;
	case RETURN_SELECTION_TIMEOUT:
		MV_PRINTD("RETURN_SELECTION_TIMEOUT");
		SCpnt->result = (DID_NO_CONNECT<<16);
		break;
	case RETURN_BUS_RESET:
		MV_PRINTD("RETURN_BUS_RESET");
		SCpnt->result = (DID_RESET<<16);
		break;
	default:
		MV_PRINTD("RETURN_ERROR");
		SCpnt->result = ((DRIVER_INVALID|SUGGEST_ABORT)<<24) | (DID_ABORT<<16);
		break;
	}
	FreeCommand(_VBUS_P pCmd);
	ScsiDone(SCpnt);
}

static unsigned int fill_msense_caching(PVDevice pVDev, u8 *p)
{
	u8 page[7] = { 0xf, 0, 0x10, 0, 0x8, 0xa, 0x4 };
	memcpy(p, page, sizeof(page));
	return sizeof(page);
}

static unsigned int fill_msense_ctlmode(u8 *p)
{
	const u8 page[] = {0xa, 0xa, 2, 0, 0, 0, 0, 0, 0xff, 0xff, 0, 30};
	memcpy(p, page, sizeof(page));
	return sizeof(page);
}

static void do_mode_sense(PVDevice pVDev, Scsi_Cmnd *SCpnt)
{
	u8 *cdb = SCpnt->cmnd, *p;
	unsigned int page_control, output_len;

	/* only support saved and current values */
	page_control = cdb[2] >> 6;
	if ((page_control != 0) && (page_control != 3)) {
invalid:
		SCpnt->result = ((DRIVER_INVALID|SUGGEST_ABORT)<<24) | (DID_ABORT<<16);
		return;
	}
	
	if (SCpnt->use_sg) goto invalid;

	output_len = (cdb[0] == MODE_SENSE)? 4 : 8;
	p = (u8 *)SCpnt->request_buffer;
	memset(p, 0, output_len);

	switch(cdb[2] & 0x3f) {
	case 0x08: /* caching */
		output_len += fill_msense_caching(pVDev, p+output_len);
		break;

	case 0x0a: /* control mode */
		output_len += fill_msense_ctlmode(p+output_len);
		break;

	case 0x3f: /* all pages */
		output_len += fill_msense_caching(pVDev, p+output_len);
		output_len += fill_msense_ctlmode(p+output_len);
		break;

	default:		/* invalid page code */
		goto invalid;
	}

	if (cdb[0] == MODE_SENSE) {
		output_len--;
		p[0] = output_len;
	} else {
		output_len -= 2;
		p[0] = output_len >> 8;
		p[1] = output_len;
	}
	SCpnt->result = (DID_OK<<16);
}

/****************************************************************
 *	Name:	OsSendCommand
 *	Description:	execute a queued command from current_Q.
 *	Parameters:		SCpnt - Pointer to SCSI command structure.
 *	Returns:		viod
 ****************************************************************/
void HPTLIBAPI OsSendCommand (_VBUS_ARG Scsi_Cmnd *SCpnt)
{
	PCommand pCmd;
	struct Scsi_Host *phost = sc_host(SCpnt);
	IAL_ADAPTER_T    *pAdapter = HOSTDATA(phost);
	PVDevice pVDev = pAdapter->VBus.pVDevice[sc_target(SCpnt)];

	/*
	 * we dec pAdapter->outstandingCommands in ScsiDone, so ScsiDone can NOT
	 * be called before OsSendCommand.
	 */
	pAdapter->outstandingCommands++;

	if (pVDev == NULL || pVDev->vf_online == 0)
	{
		MV_PRINTD("in queuecommand, requested channel(%d) on host(%d) not connected\n",
				  sc_channel(SCpnt), phost->host_no);
		SCpnt->result = DID_NO_CONNECT;
		ScsiDone(SCpnt);
		goto out;
	}

	switch(SCpnt->cmnd[0]) {
	case START_STOP:
		SCpnt->result = (DID_OK<<16);
		break;

	case INQUIRY:
	{
		PINQUIRYDATA pInquiryData;
		if (scsicmd_buf_get(SCpnt, (void **)&pInquiryData)<36 )
		{
			scsicmd_buf_put(SCpnt, pInquiryData);
			SCpnt->result = ((DRIVER_INVALID|SUGGEST_ABORT)<<24) | (DID_ABORT<<16);
			break;
		}
		ZeroMemory(pInquiryData, 36);
		SetInquiryData(pInquiryData, pVDev);
		scsicmd_buf_put(SCpnt, pInquiryData);
		SCpnt->result = (DID_OK<<16);
		break;
	}
	case READ_CAPACITY:
	{
		UCHAR *rbuf;
		UINT cap;
		
		if (scsicmd_buf_get(SCpnt, (void **)&rbuf)<8) {
			scsicmd_buf_put(SCpnt, rbuf);
			SCpnt->result = ((DRIVER_INVALID|SUGGEST_ABORT)<<24) | (DID_ABORT<<16);
			break;
		}

		if (pVDev->VDeviceCapacity>0xfffffffful)
			cap = 0xffffffff;
		else
			cap = pVDev->VDeviceCapacity - 1;
		rbuf[0] = (u8)(cap>>24);
		rbuf[1] = (u8)(cap>>16);
		rbuf[2] = (u8)(cap>>8);
		rbuf[3] = (u8)cap;
		rbuf[4] = 0;
		rbuf[5] = 0;
		rbuf[6] = 2;
		rbuf[7] = 0;
		scsicmd_buf_put(SCpnt, rbuf);
		SCpnt->result = (DID_OK<<16);
		break;
	}
		
	case 0x9e: /* SERVICE_ACTION_IN */
		if ((SCpnt->cmnd[1] & 0x1f)==0x10 /* SAI_READ_CAPACITY_16 */) {
			u8 *rbuf;
			u64 cap = pVDev->VDeviceCapacity - 1;

			if (scsicmd_buf_get(SCpnt, (void **)&rbuf)<8) {
				scsicmd_buf_put(SCpnt, rbuf);
				SCpnt->result = ((DRIVER_INVALID|SUGGEST_ABORT)<<24) | (DID_ABORT<<16);
				break;
			}
			rbuf[0] = (u8)(cap>>56);
			rbuf[1] = (u8)(cap>>48);
			rbuf[2] = (u8)(cap>>40);
			rbuf[3] = (u8)(cap>>32);
			rbuf[4] = (u8)(cap>>24);
			rbuf[5] = (u8)(cap>>16);
			rbuf[6] = (u8)(cap>>8);
			rbuf[7] = (u8)cap;
			rbuf[8] = 0;
			rbuf[9] = 0;
			rbuf[10] = 2;
			rbuf[11] = 0;
			scsicmd_buf_put(SCpnt, rbuf);
			SCpnt->result = (DID_OK<<16);
		}
		else
			SCpnt->result = ((DRIVER_INVALID|SUGGEST_ABORT)<<24) | (DID_ABORT<<16);
		break;

	case READ_6:
	case WRITE_6:
	case 0x13: /*SCSIOP_VERIFY6*/
	case READ_10:
	case WRITE_10:
	case VERIFY:
	case 0x88: /* READ_16 */
	case 0x8a: /* WRITE_16 */
		{
			pCmd = AllocateCommand(_VBUS_P0);
			HPT_ASSERT(pCmd);

			switch (SCpnt->cmnd[0]) {
			case READ_6:
			case WRITE_6:
			case 0x13: /*SCSIOP_VERIFY6*/
				pCmd->uCmd.Ide.Lba = ((ULONG)SCpnt->cmnd[1] << 16) |
					((ULONG)SCpnt->cmnd[2] << 8) | (ULONG)SCpnt->cmnd[3];
				pCmd->uCmd.Ide.nSectors = SCpnt->cmnd[4];
				break;
			case 0x88: /* READ_16 */
			case 0x8a: /* WRITE_16 */
				{
					u64 block =
						((u64)SCpnt->cmnd[2]<<56) | 
						((u64)SCpnt->cmnd[3]<<48) | 
						((u64)SCpnt->cmnd[4]<<40) | 
						((u64)SCpnt->cmnd[5]<<32) | 
						((u64)SCpnt->cmnd[6]<<24) | 
						((u64)SCpnt->cmnd[7]<<16) | 
						((u64)SCpnt->cmnd[8]<<8) | 
						((u64)SCpnt->cmnd[9]);
					pCmd->uCmd.Ide.Lba = (LBA_T)block;
					pCmd->uCmd.Ide.nSectors = (USHORT)SCpnt->cmnd[13] | ((USHORT)SCpnt->cmnd[12]<<8);
				}
				break;
			default:
				pCmd->uCmd.Ide.Lba = (ULONG)SCpnt->cmnd[5] | ((ULONG)SCpnt->cmnd[4] << 8) |
					 ((ULONG)SCpnt->cmnd[3] << 16) | ((ULONG)SCpnt->cmnd[2] << 24);
				pCmd->uCmd.Ide.nSectors = (USHORT)SCpnt->cmnd[8] | ((USHORT)SCpnt->cmnd[7]<<8);
				break;
			}
			switch (SCpnt->cmnd[0]) {
			case READ_10:
			case READ_6:
			case 0x88: /* READ_16 */
				pCmd->uCmd.Ide.Command = IDE_COMMAND_READ;
				pCmd->cf_data_in = 1;
				break;
			case WRITE_10:
			case WRITE_6:
			case 0x8a: /* WRITE_16 */
				pCmd->uCmd.Ide.Command = IDE_COMMAND_WRITE;
				pCmd->cf_data_out = 1;
				break;
			case VERIFY:
			case 0x13: /*SCSIOP_VERIFY6*/
				pCmd->uCmd.Ide.Command = IDE_COMMAND_VERIFY;
				break;
			}
			MV_PRINTD("OsSendCommand: SCpnt=%p, pVDev=%p cmd=%x nSectors=%d lba=%x\n",
				SCpnt, pVDev, pCmd->uCmd.Ide.Command, pCmd->uCmd.Ide.nSectors, pCmd->uCmd.Ide.Lba);
			pCmd->pOrgCommand = SCpnt;
			pCmd->pVDevice = pVDev;
			pCmd->pfnBuildSgl = fOsBuildSgl;
			HPT_SCP(SCpnt)->mapped = 0;
			pCmd->pfnCompletion = fOsCommandDone;
			HPT_ASSERT(pVDev->pfnSendCommand);
			pAdapter->last_active_jiffy = jiffies;
			pVDev->pfnSendCommand(_VBUS_P pCmd);
			goto out;
		}

	case TEST_UNIT_READY:
		SCpnt->result = (DID_OK<<16);
		break;
		
	case MODE_SENSE:
	case MODE_SENSE_10:
		do_mode_sense(pVDev, SCpnt);
		break;

	default:
		SCpnt->result = ((DRIVER_INVALID|SUGGEST_ABORT)<<24) | (DID_ABORT<<16);
		break;
	}
	ScsiDone(SCpnt);
out:
	CheckPendingCall(_VBUS_P0);
}

/****************************************************************
 *	Name:   hptmv_queuecommand
 *	Description:	Process a queued command from the SCSI manager.
 *	Parameters:		SCpnt - Pointer to SCSI command structure.
 *					done  - Pointer to done function to call.
 *	Returns:		Status code.
 ****************************************************************/
int hptmv_queuecommand (Scsi_Cmnd * SCpnt, void (*done) (Scsi_Cmnd *))
{
	struct Scsi_Host *phost = sc_host(SCpnt);
	IAL_ADAPTER_T    *pAdapter = HOSTDATA(phost);
	PVBus             _vbus_p  = &pAdapter->VBus;
	unsigned long flags;

	MV_PRINTD("\nhptmv_queuecommand() host=%d %d/%d/%d cdb=(%x-%x-%x)\n",phost->host_no,
		sc_channel(SCpnt), sc_target(SCpnt), sc_lun(SCpnt),
			*((PULONG)&SCpnt->cmnd), *((PULONG)&SCpnt->cmnd+1), *((PULONG)&SCpnt->cmnd+2));

	SCpnt->result = 0;
	SCpnt->scsi_done = done;

#if DBG
	if (SCpnt->retries > 0)
	{
		hpt_printk("RR182x: retry command host=%d, target=%d\n", phost->host_no,
				 sc_target(SCpnt));
	}
#endif

	if (done == NULL)
	{
		hpt_printk("RR182x: in queuecommand, done function can't be NULL\n");
		return 0;
	}

	spin_lock_irqsave(&pAdapter->driver_lock, flags);

	if (pAdapter->outstandingCommands == 0 && atomic_read(&pAdapter->DPC_Request)==0)
		Check_Idle_Call(pAdapter);

	if (sc_channel(SCpnt) || sc_target(SCpnt) >= MV_SATA_CHANNELS_NUM || sc_lun(SCpnt))
	{
		hpt_printk("RR182x: in queuecommand, Bad target received,id=%d, "
				 "channel=%d target=%d, lun=%d\n", phost->host_no, sc_channel(SCpnt),
				 sc_target(SCpnt), sc_lun(SCpnt));
		SCpnt->result = DID_BAD_TARGET << 16;
		SCpnt->scsi_done(SCpnt);
		goto out;
	}

	if(mWaitingForIdle(_VBUS_P0))
	{
		Queue_SC(&pAdapter->pending_Q, SCpnt);
		/*printk("queue pending scmd %x host %x\n",SCpnt, sc_host(SCpnt) ); */
		goto out;;
	}

	OsSendCommand(_VBUS_P SCpnt);

out:
	spin_unlock_irqrestore (&pAdapter->driver_lock, flags);
	MV_PRINTD("----- Leaving queueCommand() -----\n");
	return 0;

}

int hptmv_scsi_ioctl_handle(Scsi_Device *dev, int cmd, void *arg);
/****************************************************************
 *	Name:   hptmv_scsi_ioctl
 *	Description:	Process a scsi ioctl from the application, just like the S.M.A.R.T app.
 *	Parameters:		dev - Pointer to SCSI dev, which will be accessed, such sda...
 *					cmd - ioctl cmd type.
 					arg - buffer from app.
 *	Returns:		Status code.
 ****************************************************************/
static int hptmv_scsi_ioctl(Scsi_Device *dev, int cmd, void *arg)
{
	return hptmv_scsi_ioctl_handle(dev, cmd, arg);
}

/****************************************************************
 *	Name:	hptmv_abort
 *	Description:	Process the Abort command from the SCSI manager.
 *	Parameters:		SCpnt - Pointer to SCSI command structure.
 *	Returns:		Always snooze.
 ****************************************************************/
int hptmv_abort (Scsi_Cmnd *SCpnt)
{
	hpt_printk("Abort called pid: %ld target: %x lun: %x\n",
		SCpnt->pid, sc_target(SCpnt), sc_lun(SCpnt));
	return FAILED;
}

/****************************************************************
 *	Name:   hptmv_reset
 *	Description:	reset given devise, all pending commands will be aborted
 *					with status DID_RESET.
 *	Parameters:		SCpnt - Pointer to SCSI command structure.
 *	Returns:		Status code.
 ****************************************************************/
int hptmv_reset (Scsi_Cmnd *SCpnt)
{
	IAL_ADAPTER_T   *pAdapter = HOSTDATA(sc_host(SCpnt));
	PVBus           _vbus_p   = &pAdapter->VBus;
	unsigned long   flags;

	hpt_printk("RR182x: Bus Reset: host=%d, channel=%d target=%d\n",
			 sc_host(SCpnt)->host_no, sc_channel(SCpnt), sc_target(SCpnt));

	/*
	 * Return immediately if reset is in progress.
	 */
	if (pAdapter->status == TRUE) {
		hpt_printk("reset: exit, already in reset.\n");
		return FAILED;
	}

	spin_lock_irqsave(&pAdapter->driver_lock,flags);

	pAdapter->status = TRUE;

	fMvResetBus(_vbus_p);

	HPT_ASSERT(HOSTDATA(sc_host(SCpnt))->outstandingCommands==0);
	HPT_ASSERT(pAdapter->pending_Q == NULL);

	pAdapter->status = FALSE;
	up(&hpt_queue_wt);

	spin_unlock_irqrestore(&pAdapter->driver_lock, flags);

	return SUCCESS;
}

/****************************************************************
 *	Name:   hptmv_release
 *	Description:   release scsi host
 *	Parameters:		SCpnt - Pointer to SCSI host structure.
 *	Returns:	      0 on success, otherwise of failure.
 ****************************************************************/
int hptmv_release (struct Scsi_Host *pHost)
{
	IAL_ADAPTER_T *pAdapter = HOSTDATA(pHost);
	MV_U8 channel;
	MV_SATA_ADAPTER * pMvSataAdapter = &pAdapter->mvSataAdapter;
	PUCHAR prdTable;
	_VBUS_INST(&pAdapter->VBus)

	MV_PRINTD("RR182x: release host %d\n", pHost->host_no);

		/*Stop ALL the rebuild thread, since we can not know which adapter in hpt_worker_thread*/
	hptmv_cleanup();
	
	if (pAdapter==gIal_Adapter)
		gIal_Adapter = gIal_Adapter->next;
	else {
		IAL_ADAPTER_T *pAdapTemp;
		for (pAdapTemp = gIal_Adapter; pAdapTemp; pAdapTemp = pAdapTemp->next) {
			if (pAdapter==pAdapTemp->next) {
				pAdapTemp->next = pAdapter->next;
				break;
			}
		}
	}

	if (gIal_Adapter==0) {
		unregister_reboot_notifier(&mv_notifier);
	}

	FlushAdapter(pAdapter);

#ifdef _RAID5N_
	free_raid5_memory(_VBUS_P0);
#endif

	if(pAdapter->pCommandBlocks) VFREE(pAdapter->pCommandBlocks);
	while ((prdTable=AllocatePRDTable(pAdapter)))
		kfree(prdTable);

	/* Delete any pending timers */
	del_timer (&pAdapter->event_timer_connect);
	del_timer (&pAdapter->event_timer_disconnect);

	for (channel =0 ; channel < MV_SATA_CHANNELS_NUM; channel++)
	{
		if (pMvSataAdapter->sataChannel[channel] &&
			(HOSTDATA(pHost)->mvChannel[channel].online == MV_TRUE))
		{
			mvSataDisableChannelDma(pMvSataAdapter, channel);

			mvSataFlushDmaQueue(pMvSataAdapter, channel,
								MV_FLUSH_TYPE_CALLBACK);
			hptmv_free_channel(HOSTDATA(pHost), channel);
		}
	}
	mvSataShutdownAdapter(pMvSataAdapter);
	free_irq (pHost->irq, pAdapter);
	iounmap((void *)pMvSataAdapter->adapterIoBaseAddress);
	hptmv_free_edma_queues(HOSTDATA(pHost));
	
	scsi_unregister(pHost);
	return 0;
}

/***************************************************************************
 * Function:     hpt_worker_thread
 * Description:  Do background rebuilding. Execute in kernel thread context.
 * Returns:      None
 ***************************************************************************/
#ifdef SUPPORT_ARRAY
static void thread_timeout(unsigned long data)
{
	up((struct semaphore *)data);
}

static void check_flush(void)
{
	IAL_ADAPTER_T *pAdapter;
	unsigned long flags;
	
	for (pAdapter = gIal_Adapter; pAdapter; pAdapter = pAdapter->next) {
		spin_lock_irqsave(&pAdapter->driver_lock, flags);
		if (time_before(pAdapter->last_active_jiffy + flush_interval, jiffies)) {
			_VBUS_INST(&pAdapter->VBus)
			flush_stripe_cache(_VBUS_P0);
			pAdapter->last_active_jiffy = jiffies;
		}
		spin_unlock_irqrestore(&pAdapter->driver_lock, flags);
	}
}

void hpt_worker_thread(void *arg)
{
	DECLARE_MUTEX_LOCKED(semQueue);
	DECLARE_MUTEX_LOCKED(sem);
	struct timer_list timer;

	lock_kernel();

	/*
	 * Detach thread
	 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
	daemonize("hpt_wt");
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	daemonize();
#else 
	{
		struct fs_struct *fs;
		exit_mm(current);

		current->session = 1;
		current->pgrp = 1;

		exit_fs(current);
		fs = init_task.fs;

		if( !fs){
			hpt_printk("fs is NULL! Is this an SMP kernel on uniproc machine?\n ");
		}
		else{
			current->fs = fs;
			atomic_inc(&fs->count);
		}
	}
#endif

	/*
	 * Set the name of this process.
	 */
	sprintf(current->comm, "hpt_wt");

	hpt_queue_ready = &semQueue;

	hpt_worker_thread_handle = current;

	unlock_kernel();

	up(&hpt_notify_wt);

	init_timer(&timer);
	timer.data = (unsigned long)hpt_queue_ready;
	timer.function = thread_timeout;
	flush_interval = flush_interval * HZ / 100;

	while(1) {
		unsigned long flags;
		
		if (flush_interval) {
			timer.expires = flush_interval + jiffies + 1;
			add_timer(&timer);
		}
		
		down_interruptible(hpt_queue_ready);
		
		if (flush_interval) del_timer_sync(&timer);
		
		check_flush();

		while (DpcQueue_First!=DpcQueue_Last) {
			ST_HPT_DPC p;
			IAL_ADAPTER_T *pAdapter;

			p = DpcQueue[DpcQueue_First];
			pAdapter = (IAL_ADAPTER_T *)p.arg1;

			if(mWaitingForIdle(&pAdapter->VBus) || (pAdapter->status == TRUE)) {
				MV_PRINTD("hpt_wt: waiting for idle\n");
				down(&hpt_queue_wt);
			}

			spin_lock_irqsave(&DpcQueueLock, flags);
			DpcQueue_First++;
			DpcQueue_First %= MAX_DPC;
			spin_unlock_irqrestore(&DpcQueueLock, flags);

			atomic_inc(&pAdapter->DPC_Request);
			p.dpc(p.arg1, p.arg2, p.flags);

			/* since we may have prevented Check_Idle_Call, do it here */
			if (atomic_dec_and_test(&pAdapter->DPC_Request)) {
				spin_lock_irqsave(&pAdapter->driver_lock, flags);
				if (pAdapter->outstandingCommands == 0) {
					_VBUS_INST(&pAdapter->VBus);
					Check_Idle_Call(pAdapter);
					CheckPendingCall(_VBUS_P0);
				}
				spin_unlock_irqrestore(&pAdapter->driver_lock, flags);
			}

			/* see also lock_driver_idle */
			do { 
				schedule();
				check_flush();
			} while (atomic_read(&pAdapter->DPC_Request));
			
			/*
			 * we should finish all pending DPCs before terminate the thread.
			 * Abort all rebuild process.
			 */
			if (stop_thread || signal_pending(current)) {
				IAL_ADAPTER_T *pAdapter;
				PVDevice      pArray;
				PVBus         _vbus_p;
				int i;
				pAdapter = gIal_Adapter;

				while(pAdapter != 0){

					_vbus_p = &pAdapter->VBus;

					for (i=0;i<MAX_VDEVICE_PER_VBUS;i++)
					{
						if ((pArray=ArrayTables(i))->u.array.dArStamp==0)
							continue;
						else if (pArray->u.array.rf_rebuilding ||
								pArray->u.array.rf_verifying ||
								pArray->u.array.rf_initializing)
							{
								pArray->u.array.rf_abort_rebuild = 1;
							}
					}
					pAdapter = pAdapter->next;
				}
			}
		}

		if (stop_thread || signal_pending(current)) break;
	}

	hpt_worker_thread_handle = 0;
	up(&hpt_notify_wt);
}
#endif

/****************************************************************
 *	Name:	hptmv_cleanup
 *	Description:	cleanup routine
 ****************************************************************/
static void hptmv_cleanup(void)
{
#ifdef SUPPORT_ARRAY
	/* kill kernel thread */
	if (hpt_worker_thread_handle)
	{
		stop_thread = 1;
		up(hpt_queue_ready);
		down(&hpt_notify_wt);
	}
#endif
}

/****************************************************************
 *	Name:	hptmv_halt
 *	Description:	shutdown routine
 ****************************************************************/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
static int hptmv_halt(struct notifier_block *nb, ulong event, void *buf)
{
	IAL_ADAPTER_T *pAdapter;

	if (event != SYS_RESTART && event != SYS_HALT && event != SYS_POWER_OFF)
		return NOTIFY_DONE;

	hptmv_cleanup();
	unregister_reboot_notifier(&mv_notifier);

	for (pAdapter = gIal_Adapter; pAdapter; pAdapter = pAdapter->next)
		FlushAdapter(pAdapter);

	return NOTIFY_OK;
}
#endif

int hptmv_init_channel(IAL_ADAPTER_T *pAdapter, MV_U8 channelNum)
{
	MV_SATA_CHANNEL *pMvSataChannel;
	dma_addr_t    req_dma_addr;
	dma_addr_t    rsp_dma_addr;

	if (channelNum >= MV_SATA_CHANNELS_NUM)
	{
		hpt_printk("RR182x[%d]: Bad channelNum=%d",
				 pAdapter->mvSataAdapter.adapterId, channelNum);
		return -1;
	}
	pMvSataChannel = &gMvSataChannels[pAdapter->mvSataAdapter.adapterId][channelNum];
	pAdapter->mvSataAdapter.sataChannel[channelNum] = pMvSataChannel;
	pMvSataChannel->channelNumber = channelNum;
	pMvSataChannel->lba48Address = MV_FALSE;
	pMvSataChannel->maxReadTransfer = MV_FALSE;

	pMvSataChannel->requestQueue = (struct mvDmaRequestQueueEntry *)
		   (pAdapter->requestsArrayBaseAlignedAddr + (channelNum * MV_EDMA_REQUEST_QUEUE_SIZE));
	req_dma_addr = pAdapter->requestsArrayBaseDmaAlignedAddr + (channelNum * MV_EDMA_REQUEST_QUEUE_SIZE);

	/* check the 1K alignment of the request queue*/
	if (((u64)req_dma_addr) & 0x3ff)
	{
		hpt_printk("RR182x[%d]: request queue allocated isn't 1 K aligned,"
				 " dma_addr=%llx channel=%d\n", pAdapter->mvSataAdapter.adapterId,
				 (u64)(ULONG_PTR)req_dma_addr, channelNum);
		return -1;
	}
	pMvSataChannel->requestQueuePciLowAddress = pci64_dma_lo32(req_dma_addr);
	pMvSataChannel->requestQueuePciHiAddress = pci64_dma_hi32(req_dma_addr);
	MV_PRINTD("RR182x[%d,%d]: request queue allocated: 0x%p\n",
			  pAdapter->mvSataAdapter.adapterId, channelNum,
			  pMvSataChannel->requestQueue);
	pMvSataChannel->responseQueue = (struct mvDmaResponseQueueEntry *)
			(pAdapter->responsesArrayBaseAlignedAddr + (channelNum * MV_EDMA_RESPONSE_QUEUE_SIZE));
	rsp_dma_addr = pAdapter->responsesArrayBaseDmaAlignedAddr + (channelNum * MV_EDMA_RESPONSE_QUEUE_SIZE);

	/* check the 256 alignment of the response queue*/
	if (((u64)rsp_dma_addr) & 0xff)
	{
		hpt_printk("RR182x[%d,%d]: response queue allocated isn't 256 byte "
				 "aligned, dma_addr=%llx\n",
				 pAdapter->mvSataAdapter.adapterId, channelNum, (u64)(ULONG_PTR)rsp_dma_addr);
		return -1;
	}
	pMvSataChannel->responseQueuePciLowAddress = pci64_dma_lo32(rsp_dma_addr);
	pMvSataChannel->responseQueuePciHiAddress = pci64_dma_hi32(rsp_dma_addr);
	MV_PRINTD("RR182x[%d,%d]: response queue allocated: 0x%p\n",
			  pAdapter->mvSataAdapter.adapterId, channelNum,
			  pMvSataChannel->responseQueue);
	pAdapter->mvChannel[channelNum].online = MV_TRUE;
	return 0;
}

/****************************************************************
 *	Name:	hptmv_int_handler
 *	Description:	Interrupt handler.
 ****************************************************************/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
static void hptmv_int_handler (int irq, void *dev_id, struct pt_regs *regs)
#else 
static irqreturn_t hptmv_int_handler (int irq, void *dev_id, struct pt_regs *regs)
#endif
{
	IAL_ADAPTER_T       *pAdapter;
	int handled;
	unsigned long f2;
	unsigned long flags;

	pAdapter = (IAL_ADAPTER_T *)dev_id;
	if(pAdapter->host->irq != irq) return
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
		IRQ_NONE;
#endif
	;

	/*
	 * Disable interrupts, if they aren't already disabled and acquire
	 * the I/O spinlock.
	 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	spin_lock_irqsave (&io_request_lock, flags);
#else 
	spin_lock_irqsave(pAdapter->host->host_lock, flags);
#endif
	spin_lock_irqsave(&pAdapter->driver_lock, f2);

	if ((handled=mvSataInterruptServiceRoutine(&pAdapter->mvSataAdapter)) == MV_TRUE)
	{
		_VBUS_INST(&pAdapter->VBus)
		CheckPendingCall(_VBUS_P0);
		if (pAdapter->outstandingCommands == 0 && atomic_read(&pAdapter->DPC_Request)==0)
			Check_Idle_Call(pAdapter);
	}

	/*
	 * Release the I/O spinlock and restore the original flags
	 * which will enable interrupts if and only if they were
	 * enabled on entry.
	 */
	spin_unlock_irqrestore(&pAdapter->driver_lock, f2);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	spin_unlock_irqrestore (&io_request_lock, flags);
#else 
	spin_unlock_irqrestore (pAdapter->host->host_lock, flags);
	return IRQ_RETVAL(handled);
#endif
}

/****************************************************************
 *	Name: mvSataEventNotify
 *	Description:	this function called by the low level to notify a certain event
 *	Parameters:     pMvSataAdapter, pointer to the Device data structure.
 *	Returns:       	MV_TRUE on success, MV_FALSE on failure.
 ****************************************************************/
#define EVENT_CONNECT					1
#define EVENT_DISCONNECT				0
#define EVENT_CONNECT_TIMER 10*HZ /* 10 seconds */
#define EVENT_DISCONNECT_TIMER 1 /* Immediate execution */
MV_BOOLEAN HPTLIBAPI mvSataEventNotify( MV_SATA_ADAPTER *pMvSataAdapter,
									MV_EVENT_TYPE eventType,
									MV_U32 param1, MV_U32 param2)
{
	IAL_ADAPTER_T   *pAdapter = pMvSataAdapter->IALData;

	switch (eventType)
	{
	case MV_EVENT_TYPE_SATA_CABLE:
		{
			MV_U8   channel = param2;

			if (param1 == EVENT_CONNECT)
			{
				if (pAdapter->sataEvents[channel] == SATA_EVENT_NO_CHANGE) {
					pAdapter->sataEvents[channel] = SATA_EVENT_CHANNEL_CONNECTED;
					hpt_printk("RR182x [%d,%d]: device connected event received\n",
							 pMvSataAdapter->adapterId, channel);
					/* Delete previous timers (if multiple drives connected in the same time */
					del_timer (&pAdapter->event_timer_connect);
					pAdapter->event_timer_connect.expires = jiffies + EVENT_CONNECT_TIMER;
					add_timer (&pAdapter->event_timer_connect);
				}
			}
			else if (param1 == EVENT_DISCONNECT)
			{
				pAdapter->sataEvents[channel] = SATA_EVENT_CHANNEL_DISCONNECTED;
				hpt_printk("RR182x [%d,%d]: device disconnected event received \n",
						 pMvSataAdapter->adapterId, channel);
				device_change(pAdapter, channel, FALSE);
				/* Delete previous timers (if multiple drives disconnected in the same time */
				del_timer(&pAdapter->event_timer_disconnect);
				pAdapter->event_timer_disconnect.expires = jiffies + EVENT_DISCONNECT_TIMER;
				add_timer(&pAdapter->event_timer_disconnect);
			}
			else
			{
				hpt_printk("RR182x: illigal value for param1(%d) at "
						 "connect/disconect event, host=%d\n", param1,
						 pMvSataAdapter->adapterId );
			}
		}
		break;
	case MV_EVENT_TYPE_ADAPTER_ERROR:
		MV_PRINTD("RR182x: DEVICE error event received, pci cause "
				  "reg=%x,  don't how to handle this\n", param1);
		return MV_TRUE;
	default:
		hpt_printk("RR182x[%d]: unknown event type (%d)\n",
				 pMvSataAdapter->adapterId, eventType);
		return MV_FALSE;
	}
	return MV_TRUE;
}

static void hptmv_handle_event(IAL_ADAPTER_T *pAdapter, int flag)
{
	MV_SATA_ADAPTER *pMvSataAdapter = &pAdapter->mvSataAdapter;
	MV_U8           channelIndex;

	for (channelIndex = 0; channelIndex < MV_SATA_CHANNELS_NUM; channelIndex++)
	{
		switch(pAdapter->sataEvents[channelIndex])
		{
			case SATA_EVENT_CHANNEL_CONNECTED:
				/* Handle only connects */
				if (flag == 1)
					break;
				if(pMvSataAdapter->sataChannel[channelIndex]) {
					hpt_printk("RR182x [%d,%d]: received connect event on exising device\n",
							 pMvSataAdapter->adapterId, channelIndex);
				}
				else {
					hpt_printk("RR182x [%d,%d]: new device connected\n",
							 pMvSataAdapter->adapterId, channelIndex);
					hptmv_init_channel(pAdapter, channelIndex);
					if (mvSataConfigureChannel( pMvSataAdapter, channelIndex) == MV_FALSE)
					{
						hpt_printk("RR182x [%d,%d] Failed to configure\n",
								 pMvSataAdapter->adapterId, channelIndex);
						hptmv_free_channel(pAdapter, channelIndex);
					}
					else
					{
						/*mvSataChannelHardReset(pMvSataAdapter, channel);*/
						if (start_channel( pAdapter, channelIndex))
						{
							hpt_printk("RR182x [%d,%d]Failed to start channel\n",
									 pMvSataAdapter->adapterId, channelIndex);
							hptmv_free_channel(pAdapter, channelIndex);
						}
						/*zwy */
						else
						{
							device_change(pAdapter, channelIndex, TRUE);
						}
					}
				}

				pAdapter->sataEvents[channelIndex] = SATA_EVENT_NO_CHANGE;
			    break;

			case SATA_EVENT_CHANNEL_DISCONNECTED:
				/* Handle only disconnects */
				if (flag == 0)
					break;
				hpt_printk("RR182x [%d,%d]: device disconnected\n",
						 pMvSataAdapter->adapterId, channelIndex);
					/* Flush pending commands */
				if(pMvSataAdapter->sataChannel[channelIndex])
				{
					_VBUS_INST(&pAdapter->VBus)
					mvSataFlushDmaQueue (pMvSataAdapter, channelIndex,
										 MV_FLUSH_TYPE_CALLBACK);
					CheckPendingCall(_VBUS_P0);
					mvSataRemoveChannel(pMvSataAdapter,channelIndex);
					hptmv_free_channel(pAdapter, channelIndex);
					hpt_printk("RR182x [%d,%d]: channel removed\n",
						 pMvSataAdapter->adapterId, channelIndex);
					if (pAdapter->outstandingCommands == 0 && atomic_read(&pAdapter->DPC_Request)==0)
						Check_Idle_Call(pAdapter);
				}
				else
				{
					hpt_printk("RR182x [%d,%d]: channel already removed!!\n",
							 pMvSataAdapter->adapterId, channelIndex);
				}
				pAdapter->sataEvents[channelIndex] = SATA_EVENT_NO_CHANGE;
				break;

			default:
				break;
		}
	}
}
static void hptmv_handle_event_connect (unsigned long data)
{
	IAL_ADAPTER_T *pAdapter = (IAL_ADAPTER_T *)data;
	unsigned long flags;
	_VBUS_INST(&pAdapter->VBus)
	
	spin_lock_irqsave(&pAdapter->driver_lock, flags);
	if (mWaitingForIdle(_VBUS_P0)) {
		pAdapter->event_timer_connect.expires = jiffies + HZ;
		add_timer(&pAdapter->event_timer_connect);
	}
	else
		hptmv_handle_event(pAdapter, 0);
	spin_unlock_irqrestore (&pAdapter->driver_lock, flags);
}
static void hptmv_handle_event_disconnect(unsigned long data)
{
	IAL_ADAPTER_T *pAdapter = (IAL_ADAPTER_T *)data;
	unsigned long flags;
	spin_lock_irqsave(&pAdapter->driver_lock, flags);
	hptmv_handle_event(pAdapter, 1);
	spin_unlock_irqrestore (&pAdapter->driver_lock, flags);
}

static int hptmv_allocate_edma_queues(IAL_ADAPTER_T *pAdapter)
{
	struct pci_dev *dev = pAdapter->pci_dev;

	pAdapter->requestsArrayBaseAddr = (u8 *)
									  pci_alloc_consistent(dev, REQUESTS_ARRAY_SIZE,
														   &(pAdapter->requestsArrayBaseDmaAddr));
	if (pAdapter->requestsArrayBaseAddr == NULL)
	{
		hpt_printk("RR182x[%d]: Failed to allocate memory for EDMA request"
				 " queues\n", pAdapter->mvSataAdapter.adapterId);
		return -1;
	}
	pAdapter->requestsArrayBaseAlignedAddr = pAdapter->requestsArrayBaseAddr;
	pAdapter->requestsArrayBaseAlignedAddr += MV_EDMA_REQUEST_QUEUE_SIZE;
	pAdapter->requestsArrayBaseAlignedAddr  = (u8 *)
		((ULONG_PTR)pAdapter->requestsArrayBaseAlignedAddr & ~(ULONG_PTR)(MV_EDMA_REQUEST_QUEUE_SIZE - 1));
	pAdapter->requestsArrayBaseDmaAlignedAddr = pAdapter->requestsArrayBaseDmaAddr;
	pAdapter->requestsArrayBaseDmaAlignedAddr += MV_EDMA_REQUEST_QUEUE_SIZE;
	pAdapter->requestsArrayBaseDmaAlignedAddr &= ~(ULONG_PTR)(MV_EDMA_REQUEST_QUEUE_SIZE - 1);

	if ((pAdapter->requestsArrayBaseDmaAlignedAddr - pAdapter->requestsArrayBaseDmaAddr) !=
		(pAdapter->requestsArrayBaseAlignedAddr - pAdapter->requestsArrayBaseAddr))
	{
		hpt_printk("RR182x[%d]: Error in Request Quueues Alignment\n",
				 pAdapter->mvSataAdapter.adapterId);
		pci_free_consistent(dev, REQUESTS_ARRAY_SIZE,
							pAdapter->requestsArrayBaseAddr,
							pAdapter->requestsArrayBaseDmaAddr);
		return -1;
	}
	/* response queues */
	pAdapter->responsesArrayBaseAddr = (u8 *)
									   pci_alloc_consistent(dev, RESPONSES_ARRAY_SIZE,
															&(pAdapter->responsesArrayBaseDmaAddr));
	if (pAdapter->responsesArrayBaseAddr == NULL)
	{
		hpt_printk("RR182x[%d]: Failed to allocate memory for EDMA response"
				 " queues\n", pAdapter->mvSataAdapter.adapterId);
		pci_free_consistent(dev, REQUESTS_ARRAY_SIZE,
							pAdapter->requestsArrayBaseAddr,
							pAdapter->requestsArrayBaseDmaAddr);
		return -1;
	}
	pAdapter->responsesArrayBaseAlignedAddr = pAdapter->responsesArrayBaseAddr;
	pAdapter->responsesArrayBaseAlignedAddr += MV_EDMA_RESPONSE_QUEUE_SIZE;
	pAdapter->responsesArrayBaseAlignedAddr  = (u8 *)
		((ULONG_PTR)pAdapter->responsesArrayBaseAlignedAddr & ~(ULONG_PTR)(MV_EDMA_RESPONSE_QUEUE_SIZE - 1));
	pAdapter->responsesArrayBaseDmaAlignedAddr = pAdapter->responsesArrayBaseDmaAddr;
	pAdapter->responsesArrayBaseDmaAlignedAddr += MV_EDMA_RESPONSE_QUEUE_SIZE;
	pAdapter->responsesArrayBaseDmaAlignedAddr &= ~(ULONG_PTR)(MV_EDMA_RESPONSE_QUEUE_SIZE - 1);

	if ((pAdapter->responsesArrayBaseDmaAlignedAddr - pAdapter->responsesArrayBaseDmaAddr) !=
		(pAdapter->responsesArrayBaseAlignedAddr - pAdapter->responsesArrayBaseAddr))
	{
		hpt_printk("RR182x[%d]: Error in Response Quueues Alignment\n",
				 pAdapter->mvSataAdapter.adapterId);
		pci_free_consistent(dev, RESPONSES_ARRAY_SIZE,
							pAdapter->responsesArrayBaseAddr,
							pAdapter->responsesArrayBaseDmaAddr);
		pci_free_consistent(dev, REQUESTS_ARRAY_SIZE,
							pAdapter->requestsArrayBaseAddr,
							pAdapter->requestsArrayBaseDmaAddr);
		return -1;
	}
	return 0;
}

static void hptmv_free_edma_queues(IAL_ADAPTER_T *pAdapter)
{
	struct pci_dev *dev = pAdapter->pci_dev;

	pci_free_consistent(dev, RESPONSES_ARRAY_SIZE,
						pAdapter->responsesArrayBaseAddr,
						pAdapter->responsesArrayBaseDmaAddr);
	pci_free_consistent(dev, REQUESTS_ARRAY_SIZE,
						pAdapter->requestsArrayBaseAddr,
						pAdapter->requestsArrayBaseDmaAddr);
}

/*
 * Host template
 */
static Scsi_Host_Template driver_template = {
	name:				     PROC_DIR_NAME,
	detect:				     hptmv_detect,
	release:			     hptmv_release,
	queuecommand:		     	hptmv_queuecommand,
	ioctl:			 hptmv_scsi_ioctl,
	eh_abort_handler:        hptmv_abort,
	eh_device_reset_handler: hptmv_reset,
	eh_bus_reset_handler:    hptmv_reset,
	bios_param:			     NULL,
	can_queue:			     MV_EDMA_QUEUE_LENGTH,
	sg_tablesize:		     16,
	/* larger cmd_per_lun will cause Edma queue full, and hang up RAID5 case. */
	cmd_per_lun:		     8,
	unchecked_isa_dma:	     0,
	emulated:       	     0,
	/* ENABLE_CLUSTERING will cause problem when we handle PIO for highmem_io */
	use_clustering: 	     DISABLE_CLUSTERING, 
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)   /* 2.2.x */
	use_new_eh_code:	     1,
	proc_dir:			     &proc_scsi_hptmv,
	proc_info:			     hptmv_proc_info,
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0) /* 2.4.x */
	use_new_eh_code:	     1,
	proc_name:			     PROC_DIR_NAME,
	proc_info:			     hptmv_proc_info,
	select_queue_depths:     NULL,
	max_sectors:		     128,

	#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18))
		highmem_io:              1,
	#endif
#else /* 2.6.x */
	proc_name:			     PROC_DIR_NAME,
	proc_info:			     hptmv_proc_info26,
	max_sectors:		     128,
#endif
	this_id:			     -1
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)

#include "scsi_module.c"

EXPORT_NO_SYMBOLS;

#else /* 2.6.x */

/*scsi_module.c will be deprecated since 2.6.x*/
static int __init init_this_scsi_driver(void)
{
	struct scsi_host_template *sht = &driver_template;
	struct Scsi_Host *shost;
	struct list_head *l;
	int error;

	if (!sht->release) {
		printk(KERN_ERR
		    "scsi HBA driver %s didn't set a release method.\n",
		    sht->name);
		return -EINVAL;
	}

	sht->module = THIS_MODULE;
	INIT_LIST_HEAD(&sht->legacy_hosts);

	sht->detect(sht);
	if (list_empty(&sht->legacy_hosts))
		return -ENODEV;

	list_for_each_entry(shost, &sht->legacy_hosts, sht_legacy_list) {
		error = scsi_add_host(shost, &((HOSTDATA(shost))->pci_dev->dev));
		if (error)
			goto fail;
		scsi_scan_host(shost);
	}
	return 0;
 fail:
	l = &shost->sht_legacy_list;
	while ((l = l->prev) != &sht->legacy_hosts)
		scsi_remove_host(list_entry(l, struct Scsi_Host, sht_legacy_list));
	return error;
}

static void __exit exit_this_scsi_driver(void)
{
	struct scsi_host_template *sht = &driver_template;
	struct Scsi_Host *shost, *s;

	list_for_each_entry(shost, &sht->legacy_hosts, sht_legacy_list)
		scsi_remove_host(shost);
	list_for_each_entry_safe(shost, s, &sht->legacy_hosts, sht_legacy_list)
		sht->release(shost);

	if (list_empty(&sht->legacy_hosts))
		return;

	printk(KERN_WARNING "%s did not call scsi_unregister\n", sht->name);
	dump_stack();

	list_for_each_entry_safe(shost, s, &sht->legacy_hosts, sht_legacy_list)
		scsi_unregister(shost);
}

module_init(init_this_scsi_driver);
module_exit(exit_this_scsi_driver);

#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,10)
MODULE_LICENSE("Proprietary");
#endif
