#ifndef _MVOS_H_
#define _MVOS_H_

#if DBG
#define MV_DEBUG_LOG
#endif

/* HDD features */
#define ENABLE_READ_AHEAD
#define ENABLE_WRITE_CACHE

/* Typedefs    */
#define FAR
#define SS_SEG
#ifdef FASTCALL
#undef FASTCALL
#endif
#ifndef HPTLIBAPI
#define HPTLIBAPI __attribute__((regparm(0)))
#endif
#define FASTCALL HPTLIBAPI
#define PASCAL   HPTLIBAPI

typedef unsigned short USHORT;
typedef unsigned char  UCHAR;
typedef unsigned char *PUCHAR;
typedef unsigned short *PUSHORT;
typedef unsigned char  BOOLEAN;
typedef unsigned short WORD;
typedef unsigned int   UINT, BOOL;
typedef unsigned char  BYTE;
typedef void *PVOID, *LPVOID;
typedef void *ADDRESS;

typedef int  LONG;
typedef unsigned int ULONG, *PULONG;
#ifndef LBA_T
typedef unsigned long long LBA_T;
#endif
typedef unsigned long long BUS_ADDR;
typedef unsigned int DWORD, *LPDWORD, *PDWORD;
typedef unsigned long ULONG_PTR, UINT_PTR;

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

typedef enum mvBoolean { MV_FALSE, MV_TRUE } MV_BOOLEAN;

/* System dependant typedefs */
typedef void			MV_VOID;
typedef unsigned int    MV_U32;
typedef unsigned short	MV_U16;
typedef unsigned char	MV_U8;
typedef void			*MV_VOID_PTR;
typedef MV_U32			*MV_U32_PTR;
typedef MV_U16			*MV_U16_PTR;
typedef MV_U8			*MV_U8_PTR;
typedef char			*MV_CHAR_PTR;
typedef void *          MV_BUS_ADDR_T;
typedef unsigned long	MV_CPU_FLAGS;

/* Structures  */
#pragma pack(1)
typedef struct _INQUIRYDATA {
#ifdef __BIG_ENDIAN_BITFIELD
	UCHAR DeviceTypeQualifier : 3;
	UCHAR DeviceType : 5;
	UCHAR RemovableMedia : 1;
	UCHAR DeviceTypeModifier : 7;
#else 
	UCHAR DeviceType : 5;
	UCHAR DeviceTypeQualifier : 3;
	UCHAR DeviceTypeModifier : 7;
	UCHAR RemovableMedia : 1;
#endif
	UCHAR Versions;
	UCHAR ResponseDataFormat;
	UCHAR AdditionalLength;
	UCHAR Reserved[2];
#ifdef __BIG_ENDIAN_BITFIELD
	UCHAR RelativeAddressing : 1;
	UCHAR Wide32Bit : 1;
	UCHAR Wide16Bit : 1;
	UCHAR Synchronous : 1;
	UCHAR LinkedCommands : 1;
	UCHAR Reserved2 : 1;
	UCHAR CommandQueue : 1;
	UCHAR SoftReset : 1;
#else 
	UCHAR SoftReset : 1;
	UCHAR CommandQueue : 1;
	UCHAR Reserved2 : 1;
	UCHAR LinkedCommands : 1;
	UCHAR Synchronous : 1;
	UCHAR Wide16Bit : 1;
	UCHAR Wide32Bit : 1;
	UCHAR RelativeAddressing : 1;
#endif
	UCHAR VendorId[8];
	UCHAR ProductId[16];
	UCHAR ProductRevisionLevel[4];
	UCHAR VendorSpecific[20];
	UCHAR Reserved3[40];
} __attribute__((packed)) INQUIRYDATA, *PINQUIRYDATA;

#pragma pack()

/* Functions (User implemented)*/
void HPTLIBAPI mvMicroSecondsDelay(MV_U32);
#define StallExec(x) mvMicroSecondsDelay(x)
ULONG_PTR HPTLIBAPI fOsPhysicalAddress(void *addr);

extern MV_U16 HPTLIBAPI MV_CPU_TO_LE16(MV_U16 x);
extern MV_U32 HPTLIBAPI MV_CPU_TO_LE32(MV_U32 x);
extern MV_U16 HPTLIBAPI MV_LE16_TO_CPU(MV_U16 x);
extern MV_U32 HPTLIBAPI MV_LE32_TO_CPU(MV_U32 x);

extern void HPTLIBAPI MV_REG_WRITE_BYTE(MV_BUS_ADDR_T base, MV_U32 offset, MV_U8 val);
extern void HPTLIBAPI MV_REG_WRITE_WORD(MV_BUS_ADDR_T base, MV_U32 offset, MV_U16 val);
extern void HPTLIBAPI MV_REG_WRITE_DWORD(MV_BUS_ADDR_T base, MV_U32 offset, MV_U32 val);
extern MV_U8  HPTLIBAPI MV_REG_READ_BYTE(MV_BUS_ADDR_T base, MV_U32 offset);
extern MV_U16 HPTLIBAPI MV_REG_READ_WORD(MV_BUS_ADDR_T base, MV_U32 offset);
extern MV_U32 HPTLIBAPI MV_REG_READ_DWORD(MV_BUS_ADDR_T base, MV_U32 offset);
extern void HPTLIBAPI MV_CPU_WRITE_BUFFER_FLUSH(void);

typedef void * MV_OS_SEMAPHORE;
#define mvOsSemInit(p)		(MV_TRUE)
#define mvOsSemTake(p)		(MV_TRUE)
#define mvOsSemRelease(p)	(MV_TRUE)

/*************************************************************************
 * C library
 *************************************************************************/
#ifdef NO_LIBC
int HPTLIBAPI os_memcmp(const void *cs, const void *ct, unsigned len);
void HPTLIBAPI os_memcpy(void *to, const void *from, unsigned len);
void HPTLIBAPI os_memset(void *s, char c, unsigned len);
unsigned HPTLIBAPI os_strlen(const char *s);
#define memcmp os_memcmp
#define memcpy os_memcpy
#define memset os_memset
#define strlen os_strlen
#endif
#define ZeroMemory(a, b)  	memset((char *)a, 0, b)
#define MemoryCopy(a,b,c) 	memcpy((char *)(a), (char *)(b), (UINT)(c))
#define StrLen            	strlen
#define farMemoryCopy(a,b,c) memcpy((char *)(a), (char *)(b), (UINT)c)

void * HPTLIBAPI os_kmap_ptr(BUS_ADDR addr);
void HPTLIBAPI os_kunmap_ptr(void *ptr);

/* System logging function */
#ifdef MV_DEBUG_LOG
int mvLogMsg(MV_U8, MV_CHAR_PTR, ...);
#define _mvLogMsg(x) mvLogMsg x
#else 
#define mvLogMsg(x...)
#define _mvLogMsg(x)
#endif

#if DBG
extern int hpt_dbg_level;
#define KdPrint(_x_)  printk _x_
#define KdPrintI(_x_) do{ if (hpt_dbg_level>2) KdPrint(_x_); }while(0)
#define KdPrintW(_x_) do{ if (hpt_dbg_level>1) KdPrint(_x_); }while(0)
#define KdPrintE(_x_) do{ if (hpt_dbg_level>0) KdPrint(_x_); }while(0)
#define HPT_ASSERT(x) \
	do { if (!(x)) printk("Internal error in file %s, line %d.\n", __FILE__, __LINE__); } while (0)
#else 
#define KdPrint(x)
#define KdPrintI(x)
#define KdPrintW(x)
#define KdPrintE(x)
#define HPT_ASSERT(x)
#endif

#endif
