/*
 * mvlib.c
 * Copyright (c) 2003-2004 HighPoint Technologies, Inc. All rights reserved.
 */

static int use_tcq = 0;

#if defined(MODULE)
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
module_param(use_tcq, uint,0);
#else 
MODULE_PARM(use_tcq, "i");
#endif
MODULE_PARM_DESC(use_tcq, "enable/disable TCQ feature");
#endif

void HPTLIBAPI MV_REG_WRITE_BYTE(MV_BUS_ADDR_T base, MV_U32 offset, MV_U8 val) { writeb(val, base + offset); }
void HPTLIBAPI MV_REG_WRITE_WORD(MV_BUS_ADDR_T base, MV_U32 offset, MV_U16 val) { writew(val, base + offset); }
void HPTLIBAPI MV_REG_WRITE_DWORD(MV_BUS_ADDR_T base, MV_U32 offset, MV_U32 val) { writel(val, base + offset); }
MV_U8  HPTLIBAPI MV_REG_READ_BYTE(MV_BUS_ADDR_T base, MV_U32 offset) { return readb(base + offset); }
MV_U16 HPTLIBAPI MV_REG_READ_WORD(MV_BUS_ADDR_T base, MV_U32 offset) { return readw(base + offset); }
MV_U32 HPTLIBAPI MV_REG_READ_DWORD(MV_BUS_ADDR_T base, MV_U32 offset) { return readl(base + offset); }

void HPTLIBAPI MV_CPU_WRITE_BUFFER_FLUSH(void) { wmb(); }
MV_U16 HPTLIBAPI MV_CPU_TO_LE16(MV_U16 x) { return cpu_to_le16(x); }
MV_U32 HPTLIBAPI MV_CPU_TO_LE32(MV_U32 x) { return cpu_to_le32(x); }
MV_U16 HPTLIBAPI MV_LE16_TO_CPU(MV_U16 x) { return le16_to_cpu(x); }
MV_U32 HPTLIBAPI MV_LE32_TO_CPU(MV_U32 x) { return le32_to_cpu(x); }

ULONG_PTR HPTLIBAPI fOsPhysicalAddress(void *addr) { return virt_to_bus(addr); }

void * HPTLIBAPI os_alloc_page(_VBUS_ARG0) { return VMALLOC(CACHE_PAGE_SIZE); }
void * HPTLIBAPI os_alloc_dma_page(_VBUS_ARG0) { return kmalloc(CACHE_PAGE_SIZE, GFP_ATOMIC); }
void HPTLIBAPI os_free_page(_VBUS_ARG void *p) { VFREE(p); }
void HPTLIBAPI os_free_dma_page(_VBUS_ARG void *p) { kfree(p); }
int HPTLIBAPI os_memcmp(const void *cs, const void *ct, unsigned len) { return memcmp(cs, ct, len); }
void HPTLIBAPI os_memcpy(void *to, const void *from, unsigned len) { memcpy(to, from, len); }
void HPTLIBAPI os_memset(void *s, char c, unsigned len) { memset(s, c, len); }
unsigned HPTLIBAPI os_strlen(const char *s) { return strlen(s); }

void HPTLIBAPI DoXor1(ULONG *p0, ULONG *p1, ULONG *p2, UINT nBytes)
{
	UINT i;
	register ULONG_PTR *_p0 = (ULONG_PTR *)p0;
	register ULONG_PTR *_p1 = (ULONG_PTR *)p1;
	register ULONG_PTR *_p2 = (ULONG_PTR *)p2;
	for (i = 0; i < nBytes / sizeof(ULONG_PTR); i++) *_p0++ = *_p1++ ^ *_p2++;
}

void HPTLIBAPI DoXor2(ULONG *p0, ULONG *p2, UINT nBytes)
{
	UINT i;
	register ULONG_PTR *_p0 = (ULONG_PTR *)p0;
	register ULONG_PTR *_p2 = (ULONG_PTR *)p2;
	for (i = 0; i < nBytes / sizeof(ULONG_PTR); i++) *_p0++ ^= *_p2++;
}

void HPTLIBAPI mvMicroSecondsDelay(MV_U32 msecs)
{
	if (msecs > 1000) {
		mdelay(msecs / 1000);
		udelay(msecs % 1000);
	}
	else
		udelay(msecs);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0) && defined(__x86_64__)		
	touch_nmi_watchdog();
#endif
}

ULONG HPTLIBAPI GetStamp(void)
{
	static ULONG LastStamp = 0;
	ULONG stamp;
	stamp =	jiffies;
	if(stamp == LastStamp)
		stamp++;
	LastStamp = stamp;
	return LastStamp;
}

PVDevice HPTLIBAPI GetSpareDisk(_VBUS_ARG PVDevice pArray)
{
	IAL_ADAPTER_T *pAdapter = (IAL_ADAPTER_T *)pArray->pVBus->OsExt;
	LBA_T capacity, thiscap, maxcap = MAX_LBA_T;
	PVDevice pVDevice, pFind = NULL;
	int i;
	
	capacity = pArray->VDeviceCapacity;
	do_div(capacity, (pArray->u.array.bArnMember-1));

	for(i=0;i<MV_SATA_CHANNELS_NUM;i++)
	{
		pVDevice = &pAdapter->VDevices[i];
		if(!pVDevice) 
			continue;
		thiscap = pArray->vf_format_v2? pVDevice->u.disk.dDeRealCapacity : pVDevice->VDeviceCapacity;
		/* find the smallest usable spare disk */
		if (pVDevice->VDeviceType==VD_SPARE &&
			pVDevice->u.disk.df_on_line &&
			thiscap < maxcap &&
			thiscap >= capacity)
		{						
				maxcap = pVDevice->VDeviceCapacity;
				pFind = pVDevice;			
		}
	}
	return pFind;
}

/******************************************************************
 * IO ATA Command
 *******************************************************************/
int HPTLIBAPI fDeReadWrite(PDevice pDev, ULONG Lba, UCHAR Cmd, void *tmpBuffer)
{
	if (pDev->df_on_line)
		return mvReadWrite(pDev->mv, Lba, Cmd, tmpBuffer);
	else
		return 0;
}

void HPTLIBAPI fDeSelectMode(PDevice pDev, UCHAR NewMode)
{
	MV_SATA_CHANNEL *pSataChannel = pDev->mv;
	MV_SATA_ADAPTER *pSataAdapter = pSataChannel->mvSataAdapter;	
	MV_U8 channelIndex = pSataChannel->channelNumber;
	UCHAR mvMode;
	/* 508x don't use MW-DMA? */
	if (NewMode>4 && NewMode<8) NewMode = 4;
	pDev->bDeModeSetting = NewMode;
	if (NewMode<=4)
		mvMode = MV_ATA_TRANSFER_PIO_0 + NewMode;
	else
		mvMode = MV_ATA_TRANSFER_UDMA_0 + (NewMode-8);

	/*To fix 88i8030 bug*/
	if (mvMode > MV_ATA_TRANSFER_UDMA_0 && mvMode < MV_ATA_TRANSFER_UDMA_4)
		mvMode = MV_ATA_TRANSFER_UDMA_0;

	mvSataDisableChannelDma(pSataAdapter, channelIndex);
	/* Flush pending commands */
	mvSataFlushDmaQueue (pSataAdapter, channelIndex, MV_FLUSH_TYPE_NONE);

	if (mvStorageDevATASetFeatures(pSataAdapter, channelIndex,
								   MV_ATA_SET_FEATURES_TRANSFER,
								   mvMode, 0, 0, 0) == MV_FALSE)
	{
		KdPrint(("channel %d: Set Features failed\n", channelIndex)); 
	}
	/* Enable EDMA */
	if (mvSataEnableChannelDma(pSataAdapter, channelIndex) == MV_FALSE)
		KdPrint(("Failed to enable DMA, channel=%d", channelIndex));
}

extern PVDevice HPTLIBAPI fGetFirstChild(PVDevice pLogical);
extern void HPTLIBAPI fResetBootMark(PVDevice pLogical);
void fRegisterVdevice(IAL_ADAPTER_T *pAdapter)
{
	PVDevice pPhysical, pLogical;
	PVBus  pVBus;
	int i,j;

	for(i=0;i<MV_SATA_CHANNELS_NUM;i++) {
		pPhysical = &(pAdapter->VDevices[i]);
		pLogical = pPhysical;
		while (pLogical->pParent) pLogical = pLogical->pParent;
		if (pLogical->vf_online==0) {
			pPhysical->vf_bootmark = pLogical->vf_bootmark = 0;
			continue;
		}
		if (pLogical->VDeviceType==VD_SPARE || pPhysical!=fGetFirstChild(pLogical)) 
			continue;

		pVBus = &pAdapter->VBus;
		if(pVBus)
		{
			j=0;
			while(j<MAX_VDEVICE_PER_VBUS && pVBus->pVDevice[j]) j++;
			if(j<MAX_VDEVICE_PER_VBUS){
				pVBus->pVDevice[j] = pLogical; 
				pLogical->pVBus = pVBus;

				if (j>0 && pLogical->vf_bootmark) {
					if (pVBus->pVDevice[0]->vf_bootmark) {
						fResetBootMark(pLogical);
					}
					else {
						do { pVBus->pVDevice[j] = pVBus->pVDevice[j-1]; } while (--j);
						pVBus->pVDevice[0] = pLogical;
					}
				}
			}
		}
	}
}

PVOID AllocatePRDTable(IAL_ADAPTER_T *pAdapter)
{
	PVOID ret;
	if (pAdapter->pFreePRDLink) {
		ret = (PVOID)pAdapter->pFreePRDLink;
		pAdapter->pFreePRDLink = *(void**)ret;
		return ret;
	}
	return NULL;
}

void FreePRDTable(IAL_ADAPTER_T *pAdapter, PVOID PRDTable)
{
	*(void**)PRDTable = pAdapter->pFreePRDLink;
	pAdapter->pFreePRDLink = PRDTable;
}

/******************************************************************************
 *	Name: hptmv_parse_identify_results
 *	Description:	this functions parses the identify command results, 
 *					and updates the channel stucture accordingly.
 *	Parameters:     pMvSataChannel, pointer to the channel data structure.
 *	Returns:       	=0 ->success, < 0 ->failure.
 ******************************************************************************/
static int hptmv_parse_identify_results(MV_SATA_CHANNEL *pMvSataChannel)
{
	MV_U16  *iden = pMvSataChannel->identifyDevice;

	/*LBA addressing*/
	if (! (iden[IDEN_CAPACITY_1_OFFSET] & le16_to_cpu(0x200)))
	{
		MV_PRINTD("IAL Error in IDENTIFY info: LBA not supported\n");
		return -1;
	}
	/*DMA support*/
	if (! (iden[IDEN_CAPACITY_1_OFFSET] & le16_to_cpu(0x100)))
	{
		MV_PRINTD("IAL Error in IDENTIFY info: DMA not supported\n");
		return -1;
	}
	/* PIO */
	if ((iden[IDEN_VALID] & le16_to_cpu(2)) == 0)
	{
		MV_PRINTD("IAL Error in IDENTIFY info: not able to find PIO mode\n");
		return -1;
	}
	/*UDMA*/
	if ((iden[IDEN_VALID] & le16_to_cpu(4)) == 0)
	{
		MV_PRINTD("IAL Error in IDENTIFY info: not able to find UDMA mode\n");
		return -1;
	}

	/* 48 bit address */
	if ((iden[IDEN_SUPPORTED_COMMANDS2] & le16_to_cpu(0x400)))
	{
		MV_PRINTD("%25s - %s\n", "LBA48 addressing", "supported");
		pMvSataChannel->lba48Address = MV_TRUE;
	}
	else
	{
		MV_PRINTD("%25s - %s\n", "LBA48 addressing", "Not supported");
		pMvSataChannel->lba48Address = MV_FALSE;
	}
	return 0;
}

/****************************************************************
 *	Name: start_channel
 *
 *	Description:	this functions do the following:
 *			1. identifies the channel.
 *			2. check if the storage device supports the required ATA commands
 *			3. updates the channel configuration params (lba48Address, queuedDma,..)
 *			4. updates the device features by sending ATA SET FEATURES commands
 *			5. activates the RR182x channel
 *
 *	Parameters:     pMvSataAdapter, pointer to the Device data structure.
 *		        channelNum - te chanel number.
 *
 *	Returns:       	=0 ->success, < 0 ->failure.
 *
 ****************************************************************/
int start_channel(IAL_ADAPTER_T *pAdapter, MV_U8 channelNum)
{
	MV_SATA_ADAPTER *pMvSataAdapter = &pAdapter->mvSataAdapter;
	MV_SATA_CHANNEL *pMvSataChannel = pMvSataAdapter->sataChannel[channelNum];
	MV_CHANNEL		*pChannelInfo = &(pAdapter->mvChannel[channelNum]);
	MV_U32          udmaMode,pioMode;

	hpt_printk("RR182x [%d]: start channel (%d)\n", pMvSataAdapter->adapterId, 
			 channelNum);

	/* Software reset channel */
	if (mvStorageDevATASoftResetDevice(pMvSataAdapter, channelNum) == MV_FALSE)
	{
		hpt_printk("RR182x [%d,%d]: failed to perform Software reset\n",
				 pMvSataAdapter->adapterId, channelNum);
		return -1;

	}
	/* Hardware reset channel */
	if (mvSataChannelHardReset(pMvSataAdapter, channelNum) == MV_FALSE)
	{
		/* If failed, try again - this is when trying to hardreset a channel */
		/* when drive is just spinning up */
		mdelay(5000); /* wait 5 sec before trying again */
		if (mvSataChannelHardReset(pMvSataAdapter, channelNum) == MV_FALSE)
		{
			hpt_printk("RR182x [%d,%d]: failed to perform Hard reset\n",
					 pMvSataAdapter->adapterId, channelNum);
			return -1;
		}
	}

	/* identify device*/
	if (mvStorageDevATAIdentifyDevice(pMvSataAdapter, channelNum) == MV_FALSE)
	{
		hpt_printk("RR182x [%d,%d]: failed to perform ATA Identify command\n"
				 , pMvSataAdapter->adapterId, channelNum);
		return -1;
	}
	if (hptmv_parse_identify_results(pMvSataChannel))
	{
		hpt_printk("RR182x [%d,%d]: Error in parsing ATA Identify message\n"
				 , pMvSataAdapter->adapterId, channelNum);
		return -1;
	}

	/* mvStorageDevATASetFeatures */
	/* Disable 8 bit PIO in case CFA enabled */
	if (pMvSataChannel->identifyDevice[86] & le16_to_cpu(4))
	{
		MV_PRINTD("RR182x [%d]: Disable 8 bit PIO (CFA enabled) \n",
				  pMvSataAdapter->adapterId);
		if (mvStorageDevATASetFeatures(pMvSataAdapter, channelNum,
									   MV_ATA_SET_FEATURES_DISABLE_8_BIT_PIO, 0,
									   0, 0, 0) == MV_FALSE)
		{
			hpt_printk("RR182x [%d]: channel %d: mvStorageDevATASetFeatures"
					 " failed\n", pMvSataAdapter->adapterId, channelNum); 
			return -1;
		}
	}
	/* Write cache */
#ifdef ENABLE_WRITE_CACHE
	if (pMvSataChannel->identifyDevice[82] & le16_to_cpu(0x20))
	{
		pChannelInfo->writeCacheSupported = MV_TRUE;
		if (!(pMvSataChannel->identifyDevice[85] & le16_to_cpu(0x20))) /* if not enabled by default */
		{
			if (mvStorageDevATASetFeatures(pMvSataAdapter, channelNum,
										   MV_ATA_SET_FEATURES_ENABLE_WCACHE, 0,
										   0, 0, 0) == MV_FALSE)
			{
				hpt_printk("RR182x [%d]: channel %d: mvStorageDevATASetFeatures failed\n",
						 pMvSataAdapter->adapterId, channelNum); 
				return -1;
			}
			pChannelInfo->writeCacheEnabled = MV_TRUE;
		}
		else {
			pChannelInfo->writeCacheEnabled = MV_TRUE;
		}
	}
	else
	{
		pChannelInfo->writeCacheSupported = MV_FALSE;
		pChannelInfo->writeCacheEnabled = MV_FALSE;
		MV_PRINTD("RR182x [%d]: channel %d, write cache not supported\n",
				  pMvSataAdapter->adapterId, channelNum);
	}
#else /* disable write cache */
	if (pMvSataChannel->identifyDevice[82] & le16_to_cpu(0x20))
	{
		if (!(pMvSataChannel->identifyDevice[85] & le16_to_cpu(0x20)))
		{
			MV_PRINTD("RR182x [%d]: channel =%d, disable write cache\n",
					  pMvSataAdapter->adapterId, channelNum);
			if (mvStorageDevATASetFeatures(pMvSataAdapter, channelNum,
										   MV_ATA_SET_FEATURES_DISABLE_WCACHE, 0,
										   0, 0, 0) == MV_FALSE)
			{
				hpt_printk("RR182x [%d]: channel %d: mvStorageDevATASetFeatures failed\n",
						 pMvSataAdapter->adapterId, channelNum); 
				return -1;
			}
		}
		MV_PRINTD("RR182x [%d]: channel=%d, write cache disabled\n",
				  pMvSataAdapter->adapterId, channelNum);
	}
#endif
	/* Set transfer mode */
	MV_PRINTD("RR182x [%d] Set transfer mode XFER_PIO_SLOW\n",
			  pMvSataAdapter->adapterId);
	if (mvStorageDevATASetFeatures(pMvSataAdapter, channelNum,
								   MV_ATA_SET_FEATURES_TRANSFER,
								   MV_ATA_TRANSFER_PIO_SLOW, 0, 0, 0) == 
		MV_FALSE)
	{
		hpt_printk("RR182x [%d] channel %d: Set Features failed\n",
				 pMvSataAdapter->adapterId, channelNum); 
		return -1;
	}

	if (pMvSataChannel->identifyDevice[IDEN_PIO_MODE_SPPORTED] & le16_to_cpu(1))
	{
		pioMode = MV_ATA_TRANSFER_PIO_4;
	}
	else if (pMvSataChannel->identifyDevice[IDEN_PIO_MODE_SPPORTED] & le16_to_cpu(2))
	{
		pioMode = MV_ATA_TRANSFER_PIO_3;
	}
	else
	{
		hpt_printk("IAL Error in IDENTIFY info: PIO modes 3 and 4 not supported\n");
		pioMode = MV_ATA_TRANSFER_PIO_SLOW;
	}

	MV_PRINTD("RR182x [%d] Set transfer mode XFER_PIO_4\n",
			  pMvSataAdapter->adapterId);
	pAdapter->mvChannel[channelNum].maxPioModeSupported = pioMode;
	if (mvStorageDevATASetFeatures(pMvSataAdapter, channelNum,
								   MV_ATA_SET_FEATURES_TRANSFER,
								   pioMode, 0, 0, 0) == MV_FALSE)
	{
		hpt_printk("RR182x [%d] channel %d: Set Features failed\n",
				 pMvSataAdapter->adapterId, channelNum); 
		return -1;
	}

	udmaMode = MV_ATA_TRANSFER_UDMA_0;
	if (pMvSataChannel->identifyDevice[IDEN_UDMA_MODE] & le16_to_cpu(0x40))
	{
		udmaMode =  MV_ATA_TRANSFER_UDMA_6;
	}
	else if (pMvSataChannel->identifyDevice[IDEN_UDMA_MODE] & le16_to_cpu(0x20))
	{
		udmaMode =  MV_ATA_TRANSFER_UDMA_5;
	}
	else if (pMvSataChannel->identifyDevice[IDEN_UDMA_MODE] & le16_to_cpu(0x10))
	{
		udmaMode =  MV_ATA_TRANSFER_UDMA_4;
	}
	else if (pMvSataChannel->identifyDevice[IDEN_UDMA_MODE] & le16_to_cpu(0x8))
	{
		udmaMode =  MV_ATA_TRANSFER_UDMA_3;
	}
	else if (pMvSataChannel->identifyDevice[IDEN_UDMA_MODE] & le16_to_cpu(0x4))
	{
		udmaMode =  MV_ATA_TRANSFER_UDMA_2;
	}

	MV_PRINTD("RR182x [%d] Set transfer mode XFER_UDMA_%d\n",
			  pMvSataAdapter->adapterId, udmaMode & 0xf);
	pChannelInfo->maxUltraDmaModeSupported = udmaMode;

	/*if (mvStorageDevATASetFeatures(pMvSataAdapter, channelNum,
								   MV_ATA_SET_FEATURES_TRANSFER, udmaMode,
								   0, 0, 0) == MV_FALSE)
	{
		hpt_printk("RR182x [%d] channel %d: Set Features failed\n",
				 pMvSataAdapter->adapterId, channelNum); 
		return -1;
	}*/
	if (pChannelInfo->maxUltraDmaModeSupported == 0xFF) 
		return TRUE;
	else 
		do
		{
			if (mvStorageDevATASetFeatures(pMvSataAdapter, channelNum,
								   MV_ATA_SET_FEATURES_TRANSFER, 
								   pChannelInfo->maxUltraDmaModeSupported,
								   0, 0, 0) == MV_FALSE)
			{
				if (pChannelInfo->maxUltraDmaModeSupported > MV_ATA_TRANSFER_UDMA_0)
				{
					if (mvStorageDevATASoftResetDevice(pMvSataAdapter, channelNum) == MV_FALSE)
					{
						MV_REG_WRITE_BYTE(pMvSataAdapter->adapterIoBaseAddress,
										  pMvSataChannel->eDmaRegsOffset +
										  0x11c, /* command reg */
										  MV_ATA_COMMAND_IDLE_IMMEDIATE); 
						mvMicroSecondsDelay(10000);
						mvSataChannelHardReset(pMvSataAdapter, channelNum);
						if (mvStorageDevATASoftResetDevice(pMvSataAdapter, channelNum) == MV_FALSE)
							return FALSE;
					}
					if (mvSataChannelHardReset(pMvSataAdapter, channelNum) == MV_FALSE)
						return FALSE;
					pChannelInfo->maxUltraDmaModeSupported--;
					continue;
				}
				else   return FALSE;
			}
			break;
		}while (1);

	/* Read look ahead */
	pChannelInfo->readAheadEnabled = MV_FALSE;
#ifdef ENABLE_READ_AHEAD
	if (pMvSataChannel->identifyDevice[82] & le16_to_cpu(0x40))
	{
		pChannelInfo->readAheadSupported = MV_TRUE;
		if (!(pMvSataChannel->identifyDevice[85] & le16_to_cpu(0x40))) /* if not enabled by default */
		{
			if (mvStorageDevATASetFeatures(pMvSataAdapter, channelNum,
										   MV_ATA_SET_FEATURES_ENABLE_RLA, 0, 0,
										   0, 0) == MV_FALSE)
			{
				hpt_printk("RR182x [%d] channel %d: Set Features failed\n",
						 pMvSataAdapter->adapterId, channelNum); 
				return -1;
			}
		}
		pChannelInfo->readAheadEnabled = MV_TRUE;
		MV_PRINTD("RR182x [%d]: channel=%d, read look ahead enabled\n", 
				  pMvSataAdapter->adapterId, channelNum);
	}
	else
	{
		pChannelInfo->readAheadSupported = MV_FALSE;
		MV_PRINTD("RR182x [%d]: channel %d, Read Look Ahead not supported\n",
				  pMvSataAdapter->adapterId, channelNum);
	}
#else 
	{
		if (pMvSataChannel->identifyDevice[86] & le16_to_cpu(0x20))
		{
			MV_PRINTD("RR182x [%d]:channel %d, disable read look ahead\n",
					  pMvSataAdapter->adapterId, channelNum);
			if (mvStorageDevATASetFeatures(pMvSataAdapter, channelNum,
										   MV_ATA_SET_FEATURES_DISABLE_RLA, 0, 0,
										   0, 0) == MV_FALSE)
			{
				hpt_printk("RR182x [%d]:channel %d:  ATA Set Features failed\n",
						 pMvSataAdapter->adapterId, channelNum); 
				return -1;
			}
		}
		MV_PRINTD("RR182x [%d]:channel %d, read look ahead disabled\n",
				  pMvSataAdapter->adapterId, channelNum);           
	}    
#endif


	if (use_tcq && (pMvSataChannel->identifyDevice[IDEN_SUPPORTED_COMMANDS2] & le16_to_cpu(2)))
	{
		MV_U8       depth;
		MV_BOOLEAN  result;

		depth = (le16_to_cpu(pMvSataChannel->identifyDevice[IDEN_QUEUE_DEPTH]) & 0x1f) + 1;
		MV_PRINTD("RR182x [%d]: channel %d config EDMA, Queued Mode, queue depth %d\n",
				  pMvSataAdapter->adapterId, channelNum, depth);
		result = mvSataConfigEdmaMode(pMvSataAdapter, channelNum,
									  MV_EDMA_MODE_QUEUED, depth);
		if (result == MV_FALSE)
		{
			hpt_printk("RR182x [%d] Error: mvSataConfigEdmaMode failed\n",
					 pMvSataAdapter->adapterId);
			return -1;
		}
	}
	else
	{
		MV_PRINTD("RR182x [%d]: channel %d config EDMA, Non Queued Mode\n",
				  pMvSataAdapter->adapterId, 
				  channelNum);
		if (mvSataConfigEdmaMode(pMvSataAdapter, channelNum,
								 MV_EDMA_MODE_NOT_QUEUED, 0) == MV_FALSE)
		{
			hpt_printk("RR182x [%d] channel %d Error: mvSataConfigEdmaMode failed\n",
					 pMvSataAdapter->adapterId, channelNum);
			return -1;
		}
	}
	/* Enable EDMA */
	if (mvSataEnableChannelDma(pMvSataAdapter, channelNum) == MV_FALSE)
	{
		hpt_printk("RR182x [%d] Failed to enable DMA, channel=%d\n",
				 pMvSataAdapter->adapterId, channelNum);
		return -1;
	}
	hpt_printk("RR182x [%d,%d]: channel started successfully\n",
			 pMvSataAdapter->adapterId, channelNum);
	set_fail_led(pMvSataAdapter, channelNum, 0);
	return 0;
}

void init_vdev_params(IAL_ADAPTER_T *pAdapter, MV_U8 channel)
{
	PVDevice pVDev = &pAdapter->VDevices[channel];
	MV_SATA_CHANNEL *pMvSataChannel = pAdapter->mvSataAdapter.sataChannel[channel];
	MV_U16_PTR IdentifyData = pMvSataChannel->identifyDevice;

	pMvSataChannel->outstandingCommands = 0;

	pVDev->u.disk.mv         = pMvSataChannel;
	pVDev->u.disk.df_on_line = 1;
	pVDev->u.disk.resetCount = 0;
	pVDev->u.disk.retryCount = 0;
	pVDev->u.disk.busyCount = 0;
	pVDev->u.disk.pVBus      = &pAdapter->VBus;
	pVDev->pVBus             = &pAdapter->VBus;

#ifdef SUPPORT_48BIT_LBA
	if (pMvSataChannel->lba48Address == MV_TRUE)
		pVDev->u.disk.dDeRealCapacity = ((le16_to_cpu(IdentifyData[101])<<16) | le16_to_cpu(IdentifyData[100])) - 1;
	else
#endif
	if(IdentifyData[53] & le16_to_cpu(1)) {
	pVDev->u.disk.dDeRealCapacity = 
	  (((le16_to_cpu(IdentifyData[58])<<16 | le16_to_cpu(IdentifyData[57])) < 
	  	(le16_to_cpu(IdentifyData[61])<<16 | le16_to_cpu(IdentifyData[60]))) ? 
		  (le16_to_cpu(IdentifyData[61])<<16 | le16_to_cpu(IdentifyData[60])) :
				(le16_to_cpu(IdentifyData[58])<<16 | le16_to_cpu(IdentifyData[57]))) - 1;
	} else
		pVDev->u.disk.dDeRealCapacity = 
				 (le16_to_cpu(IdentifyData[61])<<16 | le16_to_cpu(IdentifyData[60])) - 1;

	pVDev->u.disk.bDeUsable_Mode = pVDev->u.disk.bDeModeSetting = 
		pAdapter->mvChannel[channel].maxPioModeSupported - MV_ATA_TRANSFER_PIO_0;

	if (pAdapter->mvChannel[channel].maxUltraDmaModeSupported!=0xFF) {
		pVDev->u.disk.bDeUsable_Mode = pVDev->u.disk.bDeModeSetting = 
			pAdapter->mvChannel[channel].maxUltraDmaModeSupported - MV_ATA_TRANSFER_UDMA_0 + 8;
	}
}

void SetInquiryData(PINQUIRYDATA inquiryData, PVDevice pVDev)
{
	int i;
	IDENTIFY_DATA2 *pIdentify = (IDENTIFY_DATA2*)pVDev->u.disk.mv->identifyDevice;

	inquiryData->DeviceType = TYPE_DISK; /*DIRECT_ACCESS_DEVICE*/
	inquiryData->AdditionalLength = (UCHAR)(sizeof(INQUIRYDATA) - 5);
	inquiryData->CommandQueue = 1;

	switch(pVDev->VDeviceType) {
	case VD_SINGLE_DISK:
	case VD_ATAPI:
	case VD_REMOVABLE:
		if ((pVDev->u.disk.df_removable_drive) || (pIdentify->GeneralConfiguration & cpu_to_le16(0x80)))
			inquiryData->RemovableMedia = 1;
		for (i = 0; i < 20; i += 2)	{				
			inquiryData->VendorId[i] 	= ((PUCHAR)pIdentify->ModelNumber)[i + 1];
			inquiryData->VendorId[i+1] 	= ((PUCHAR)pIdentify->ModelNumber)[i];

		}
		for (i = 0; i < 4; i++) inquiryData->ProductId[12+i] = ' ';
		for (i = 0; i < 4; i += 2) {				
			inquiryData->ProductRevisionLevel[i] 	= ((PUCHAR)pIdentify->FirmwareRevision)[i+1];
			inquiryData->ProductRevisionLevel[i+1] 	= ((PUCHAR)pIdentify->FirmwareRevision)[i];
		}
		break;
		
	default:
		memcpy(&inquiryData->VendorId, "RR182x  ", 8);
#ifdef SUPPORT_ARRAY
		switch(pVDev->VDeviceType){
		case VD_RAID_0:
			if ((pVDev->u.array.pMember[0] && mIsArray(pVDev->u.array.pMember[0])) ||
				(pVDev->u.array.pMember[1] && mIsArray(pVDev->u.array.pMember[1])))
				memcpy(&inquiryData->ProductId, "RAID 1/0 Array  ", 16);
			else
				memcpy(&inquiryData->ProductId, "RAID 0 Array    ", 16);
			break;
		case VD_RAID_1:
			if ((pVDev->u.array.pMember[0] && mIsArray(pVDev->u.array.pMember[0])) ||
				(pVDev->u.array.pMember[1] && mIsArray(pVDev->u.array.pMember[1])))
				memcpy(&inquiryData->ProductId, "RAID 0/1 Array  ", 16);
			else
				memcpy(&inquiryData->ProductId, "RAID 1 Array    ", 16);
			break;
		case VD_RAID_5:
			memcpy(&inquiryData->ProductId, "RAID 5 Array    ", 16);
			break;
		case VD_JBOD:
			memcpy(&inquiryData->ProductId, "JBOD Array      ", 16);
			break;
		}
#endif
		memcpy(&inquiryData->ProductRevisionLevel, "3.00", 4);
		break;
	}
}

static void failDevice(PVDevice pVDev)
{
	PVBus _vbus_p = pVDev->pVBus;
	IAL_ADAPTER_T *pAdapter = (IAL_ADAPTER_T *)_vbus_p->OsExt;
	
	if (pVDev->vf_online) {
		pVDev->u.disk.df_on_line = 0;
		pVDev->vf_online = 0;
		if (pVDev->pfnDeviceFailed) 
			CallWhenIdle(_VBUS_P (DPC_PROC)pVDev->pfnDeviceFailed, pVDev);
	
		fNotifyGUI(ET_DEVICE_REMOVED, pVDev);

		if (pAdapter->ver_601==2 && !pAdapter->beeping) {
			pAdapter->beeping = 1;
			BeepOn(pAdapter->mvSataAdapter.adapterIoBaseAddress);
			set_fail_led(&pAdapter->mvSataAdapter, pVDev->u.disk.mv->channelNumber, 1);
		}
	}
}

int MvSataResetChannel(MV_SATA_ADAPTER *pMvSataAdapter, MV_U8 channel);
void HPTLIBAPI fDeviceSendCommand(_VBUS_ARG PCommand pCmd);
static void HPTLIBAPI handleEdmaError(_VBUS_ARG PCommand pCmd)
{
	PDevice pDevice = &pCmd->pVDevice->u.disk;
	MV_SATA_ADAPTER * pSataAdapter = pDevice->mv->mvSataAdapter;

	if (!pDevice->df_on_line) {
		KdPrint(("Device is offline"));
		pCmd->Result = RETURN_BAD_DEVICE;
		CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);	
		return;
	}

	if (pCmd->RetryCount++>5) {
		hpt_printk("too many retries on channel(%d)\n", pDevice->mv->channelNumber);
failed:
		failDevice(pCmd->pVDevice);
		pCmd->Result = RETURN_IDE_ERROR;
		CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);	
		return;
	}

	/* reset the channel and retry the command */
	if (MvSataResetChannel(pSataAdapter, pDevice->mv->channelNumber)==FAILED)
		goto failed;

	fNotifyGUI(ET_DEVICE_ERROR, Map2pVDevice(pDevice));

	hpt_printk("Retry on channel(%d)\n", pDevice->mv->channelNumber);
	fDeviceSendCommand(_VBUS_P pCmd);
}

MV_BOOLEAN HPTLIBAPI 
CommandCompletionCB(MV_SATA_ADAPTER *pMvSataAdapter,
					MV_U8 channelNum,
					MV_COMPLETION_TYPE comp_type,
					MV_VOID_PTR commandId,
					MV_U16 responseFlags,
					MV_U32 timeStamp,
					MV_STORAGE_DEVICE_REGISTERS *registerStruct)
{
	PCommand pCmd = (PCommand) commandId;
	PDevice	pDevice = 
		&(((IAL_ADAPTER_T *)(pMvSataAdapter->IALData))->VDevices[channelNum].u.disk);
	_VBUS_INST(pCmd->pVDevice->pVBus)

	if (pCmd->uScratch.sata_param.prdAddr) 
		FreePRDTable(pMvSataAdapter->IALData,pCmd->uScratch.sata_param.prdAddr);

	switch (comp_type)
	{
	case MV_COMPLETION_TYPE_NORMAL:
		pDevice->busyCount=pDevice->retryCount=pDevice->resetCount=0;    				
		pCmd->Result = RETURN_SUCCESS;
		break;
	case MV_COMPLETION_TYPE_ABORT:
		pCmd->Result = RETURN_BUS_RESET;
		break;
	case MV_COMPLETION_TYPE_ERROR:
		hpt_printk("IAL: COMPLETION ERROR, adapter %d, channel %d, flags=%x\n",
				 pMvSataAdapter->adapterId, channelNum, responseFlags);

		if (responseFlags & 4) {
			hpt_printk("ATA regs: error %x, sector count %x, LBA low %x, LBA mid %x,"
				" LBA high %x, device %x, status %x\n",
				registerStruct->errorRegister,
				registerStruct->sectorCountRegister,
				registerStruct->lbaLowRegister,
				registerStruct->lbaMidRegister,
				registerStruct->lbaHighRegister,
				registerStruct->deviceRegister,
				registerStruct->statusRegister);
		}
		
		if (responseFlags & 0x60) {
			MV_SATA_EDMA_PRD_ENTRY  *pPRDTable = pCmd->uScratch.sata_param.prdAddr;
			hpt_printk("EDMA overrun/underrun: sectors %x, PRD table:", pCmd->uCmd.Ide.nSectors);
			do {
				printk("(%x%x,%x)", pPRDTable->highBaseAddr, pPRDTable->lowBaseAddr, pPRDTable->byteCount);
			}
			while (pPRDTable++->flags==0);
		}
	
		/*We can't do handleEdmaError directly here, because CommandCompletionCB is called by 
		 * mv's ISR, if we retry the command, than the internel data structure may be destroyed*/
		pCmd->uScratch.sata_param.responseFlags = responseFlags;
		pCmd->uScratch.sata_param.bIdeStatus = registerStruct->statusRegister;
		pCmd->uScratch.sata_param.errorRegister = registerStruct->errorRegister;
		pCmd->pVDevice->u.disk.QueueLength--;
		CallAfterReturn(_VBUS_P (DPC_PROC)handleEdmaError,pCmd);
		return TRUE;
		
	default:
		hpt_printk(" Unknown completion type (%d)\n", comp_type);
		return MV_FALSE;
	}
	
	if (pCmd->uCmd.Ide.Command == IDE_COMMAND_VERIFY && pCmd->uScratch.sata_param.cmd_priv > 1) {
		pCmd->uScratch.sata_param.cmd_priv --;
		return TRUE;
	}
	pCmd->pVDevice->u.disk.QueueLength--;
	CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
	return TRUE;
}

MV_BOOLEAN mvAdapterPassthroughCommand(
								MV_SATA_ADAPTER *pSataAdapter, 
								MV_U8 channelIndex,
								PCommand pCmd
								);

void HPTLIBAPI fDeviceSendCommand(_VBUS_ARG PCommand pCmd)
{
	MV_SATA_EDMA_PRD_ENTRY  *pPRDTable = 0;
	MV_SATA_ADAPTER *pMvSataAdapter;
	MV_SATA_CHANNEL *pMvSataChannel;
	PVDevice pVDevice = pCmd->pVDevice;
	PDevice  pDevice = &pVDevice->u.disk;
	LBA_T    Lba = pCmd->uCmd.Ide.Lba;
	USHORT   nSector = pCmd->uCmd.Ide.nSectors;

	MV_QUEUE_COMMAND_RESULT result;
	MV_QUEUE_COMMAND_INFO commandInfo;	
	MV_UDMA_COMMAND_PARAMS  *pUdmaParams = &commandInfo.commandParams.udmaCommand;
	MV_NONE_UDMA_COMMAND_PARAMS *pNoUdmaParams = &commandInfo.commandParams.NoneUdmaCommand;

	MV_BOOLEAN is48bit = MV_FALSE;
	MV_U8      channel;
	int        i=0;
	
	DECLARE_BUFFER(FPSCAT_GATH, tmpSg);

	if (!pDevice->df_on_line) {
		hpt_printk("Device is offline");
		pCmd->Result = RETURN_BAD_DEVICE;
		CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
		return;
	}

	pDevice->HeadPosition = pCmd->uCmd.Ide.Lba + pCmd->uCmd.Ide.nSectors;
	pMvSataChannel = pDevice->mv;
	pMvSataAdapter = pMvSataChannel->mvSataAdapter;
	channel = pMvSataChannel->channelNumber;

	if(pCmd->cf_ide_passthrough) {
		mvAdapterPassthroughCommand(pMvSataAdapter, channel, pCmd);
		return;
	}

	/* old RAID0 has hidden lba. Remember to clear dDeHiddenLba when delete array! */
	Lba += pDevice->dDeHiddenLba;
	/* check LBA */
	if (Lba+nSector-1 > pDevice->dDeRealCapacity) {
		pCmd->Result = RETURN_INVALID_REQUEST;
		CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
		return;
	}
	
	if(Lba & 0xF0000000){
		is48bit = MV_TRUE;
	}

	switch (pCmd->uCmd.Ide.Command)
	{
	case IDE_COMMAND_READ:
	case IDE_COMMAND_WRITE:
		if (pDevice->bDeModeSetting<8) goto pio;
		
		commandInfo.type = MV_QUEUED_COMMAND_TYPE_UDMA;
		pUdmaParams->isEXT = is48bit;
		pUdmaParams->numOfSectors = nSector;
		pUdmaParams->lowLBAAddress = Lba;
		pUdmaParams->highLBAAddress = sizeof(Lba)>4? (Lba>>32) : 0;
		pUdmaParams->callBack = CommandCompletionCB;
		pUdmaParams->commandId = (MV_VOID_PTR )pCmd;
		if(pCmd->uCmd.Ide.Command == IDE_COMMAND_READ)
			pUdmaParams->readWrite = MV_UDMA_TYPE_READ;
		else 
			pUdmaParams->readWrite = MV_UDMA_TYPE_WRITE;
		
		if (pCmd->pSgTable && pCmd->cf_physical_sg) {
			FPSCAT_GATH sg1=tmpSg, sg2=pCmd->pSgTable;
			do { *sg1++=*sg2; } while ((sg2++->wSgFlag & SG_FLAG_EOT)==0);
		}
		else {
			if (!pCmd->pfnBuildSgl || !pCmd->pfnBuildSgl(_VBUS_P pCmd, tmpSg, 0)) {
pio:				
				mvSataDisableChannelDma(pMvSataAdapter, channel);
				mvSataFlushDmaQueue(pMvSataAdapter, channel, MV_FLUSH_TYPE_CALLBACK);
	
				if (pCmd->pSgTable && pCmd->cf_physical_sg==0) {
					FPSCAT_GATH sg1=tmpSg, sg2=pCmd->pSgTable;
					do { *sg1++=*sg2; } while ((sg2++->wSgFlag & SG_FLAG_EOT)==0);
				}
				else {
					if (!pCmd->pfnBuildSgl || !pCmd->pfnBuildSgl(_VBUS_P pCmd, tmpSg, 1)) {
						pCmd->Result = RETURN_NEED_LOGICAL_SG;
						goto finish_cmd;
					}
				}
										
				do {
					ULONG size = tmpSg->wSgSize? tmpSg->wSgSize : 0x10000;
					void *addr;

					if (size & 0x1ff) {
						pCmd->Result = RETURN_INVALID_REQUEST;
						goto finish_cmd;
					}
					
					addr = os_kmap_ptr(tmpSg->dSgAddress);

					if (mvStorageDevATAExecuteNonUDMACommand(pMvSataAdapter, channel,
						(pCmd->cf_data_out)?MV_NON_UDMA_PROTOCOL_PIO_DATA_OUT:MV_NON_UDMA_PROTOCOL_PIO_DATA_IN,
						is48bit,
						(MV_U16_PTR)addr, 
						size >> 1,	/* count       */
						0,		/* features  N/A  */
						(MV_U16)(size>>9),	/*sector count*/
						(MV_U16)(  (is48bit? (MV_U16)((Lba >> 16) & 0xFF00) : 0 )  | (UCHAR)(Lba & 0xFF) ), /*lbalow*/
						(MV_U16)((Lba >> 8) & 0xFF), /* lbaMid      */
						(MV_U16)((Lba >> 16) & 0xFF),/* lbaHigh     */
						(MV_U8)(0x40 | (is48bit ? 0 : (UCHAR)(Lba >> 24) & 0xFF )),/* device      */
						(MV_U8)(is48bit ? (pCmd->cf_data_in?IDE_COMMAND_READ_EXT:IDE_COMMAND_WRITE_EXT):pCmd->uCmd.Ide.Command)
					)==MV_FALSE)
					{
						pCmd->Result = RETURN_IDE_ERROR;
						os_kunmap_ptr(addr);
						goto finish_cmd;
					}
					
					os_kunmap_ptr(addr);
					
					Lba += size>>9;
					if(Lba & 0xF0000000) is48bit = MV_TRUE;
				}
				while ((tmpSg++->wSgFlag & SG_FLAG_EOT)==0);
				pCmd->Result = RETURN_SUCCESS;
finish_cmd:
				mvSataEnableChannelDma(pMvSataAdapter,channel);
				CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
				return;
			}
		}
		
		pPRDTable = (MV_SATA_EDMA_PRD_ENTRY *) AllocatePRDTable(pMvSataAdapter->IALData);
		MV_PRINTD("pPRDTable:%p\n",pPRDTable);
		if (!pPRDTable) {
			pCmd->Result = RETURN_DEVICE_BUSY;
			CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
			HPT_ASSERT(0);
			return;
		}
		/* Convert tmpSg to PRD Table */
		do{
			pPRDTable[i].flags = cpu_to_le16(tmpSg->wSgFlag);
			pPRDTable[i].byteCount = cpu_to_le16(tmpSg->wSgSize);
			pPRDTable[i].lowBaseAddr = cpu_to_le32((MV_U32)tmpSg->dSgAddress);
			if (sizeof(dma_addr_t)>4)
				pPRDTable[i].highBaseAddr = cpu_to_le32((MV_U32)(tmpSg->dSgAddress>>32));
			else
				pPRDTable[i].highBaseAddr = 0;
			pPRDTable[i].reserved = 0;
			i++;
		}while((tmpSg++->wSgFlag & SG_FLAG_EOT)==0);
		
		if (sizeof(ULONG_PTR)>4) {
			ULONG_PTR addr = fOsPhysicalAddress(pPRDTable);
			pUdmaParams->prdLowAddr = (MV_U32)addr;
			pUdmaParams->prdHighAddr = (MV_U32)((u64)addr>>32);
		}
		else {
			pUdmaParams->prdLowAddr = fOsPhysicalAddress(pPRDTable);
			pUdmaParams->prdHighAddr = 0;
		}

		if ((pUdmaParams->numOfSectors == 256) && (pMvSataChannel->lba48Address == MV_FALSE)) {
			pUdmaParams->numOfSectors = 0;
		}
		
		pCmd->uScratch.sata_param.prdAddr = (PVOID)pPRDTable;

		result = mvSataQueueCommand(pMvSataAdapter, channel, &commandInfo);

		if (result != MV_QUEUE_COMMAND_RESULT_OK)
		{
queue_failed:
			switch (result)
			{
			case MV_QUEUE_COMMAND_RESULT_BAD_LBA_ADDRESS:
				hpt_printk("IAL Error: Edma Queue command failed. Bad LBA "
						 "LBA[31:0](0x%08x)\n", pUdmaParams->lowLBAAddress);
				pCmd->Result = RETURN_IDE_ERROR;
				break;
			case MV_QUEUE_COMMAND_RESULT_QUEUED_MODE_DISABLED:
				hpt_printk("IAL Error: Edma Queue command failed. EDMA"
						 " disabled adapter %d channel %d\n",
						 pMvSataAdapter->adapterId, channel);
				mvSataEnableChannelDma(pMvSataAdapter,channel);
				pCmd->Result = RETURN_IDE_ERROR;
				break;
			case MV_QUEUE_COMMAND_RESULT_FULL:
				hpt_printk("IAL Error: Edma Queue command failed. Queue is"
						 " Full adapter %d channel %d\n",
						 pMvSataAdapter->adapterId, channel);
				pCmd->Result = RETURN_DEVICE_BUSY;
				break;
			case MV_QUEUE_COMMAND_RESULT_BAD_PARAMS:
				hpt_printk("IAL Error: Edma Queue command failed. (Bad "
						 "Params), pMvSataAdapter: %p,  pSataChannel: %p.\n",
						 pMvSataAdapter, pMvSataAdapter->sataChannel[channel]);
				pCmd->Result = RETURN_IDE_ERROR;
				break;
			default:
				hpt_printk("IAL Error: Bad result value (%d) from queue"
						 " command\n", result);
				pCmd->Result = RETURN_IDE_ERROR;
			}
			if(pPRDTable) 
				FreePRDTable(pMvSataAdapter->IALData,pPRDTable);
			CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
		}
		pDevice->QueueLength++;
		return;
		
	case IDE_COMMAND_VERIFY:
		commandInfo.type = MV_QUEUED_COMMAND_TYPE_NONE_UDMA;
		pNoUdmaParams->bufPtr = NULL;
		pNoUdmaParams->callBack = CommandCompletionCB;
		pNoUdmaParams->commandId = (MV_VOID_PTR)pCmd;
		pNoUdmaParams->count = 0;
		pNoUdmaParams->features = 0;
		pNoUdmaParams->protocolType = MV_NON_UDMA_PROTOCOL_NON_DATA;
		
		pCmd->uScratch.sata_param.cmd_priv = 1;
		if (pMvSataChannel->lba48Address == MV_TRUE){
			pNoUdmaParams->command = MV_ATA_COMMAND_READ_VERIFY_SECTORS_EXT;
			pNoUdmaParams->isEXT = MV_TRUE;
			pNoUdmaParams->lbaHigh = (MV_U16)((Lba & 0xff0000) >> 16);
			pNoUdmaParams->lbaMid = (MV_U16)((Lba & 0xff00) >> 8);   
			pNoUdmaParams->lbaLow = 
				(MV_U16)(((Lba & 0xff000000) >> 16)| (Lba & 0xff));
			pNoUdmaParams->sectorCount = nSector;
			pNoUdmaParams->device = 0x40;
			result = mvSataQueueCommand(pMvSataAdapter, channel, &commandInfo);
			if (result != MV_QUEUE_COMMAND_RESULT_OK){
				goto queue_failed;
			}
			return;
		}
		else{
			pNoUdmaParams->command = MV_ATA_COMMAND_READ_VERIFY_SECTORS;
			pNoUdmaParams->isEXT = MV_FALSE;
			pNoUdmaParams->lbaHigh = (MV_U16)((Lba & 0xff0000) >> 16);
			pNoUdmaParams->lbaMid = (MV_U16)((Lba & 0xff00) >> 8);   
			pNoUdmaParams->lbaLow = (MV_U16)(Lba & 0xff);
			pNoUdmaParams->sectorCount = 0xff & nSector;
			pNoUdmaParams->device = (MV_U8)(0x40 |
				((Lba & 0xf000000) >> 24));
			pNoUdmaParams->callBack = CommandCompletionCB;
			result = mvSataQueueCommand(pMvSataAdapter, channel, &commandInfo);
			/*FIXME: how about the commands already queued? but marvel also forgets to consider this*/
			if (result != MV_QUEUE_COMMAND_RESULT_OK){
				goto queue_failed;
			}
		}
		break;
	default:
		pCmd->Result = RETURN_INVALID_REQUEST;
		CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
		break;
	}
}

void hptmv_free_channel(IAL_ADAPTER_T *pAdapter, MV_U8 channelNum)
{
	HPT_ASSERT(channelNum < MV_SATA_CHANNELS_NUM);
	pAdapter->mvSataAdapter.sataChannel[channelNum] = NULL;
}

int MvSataResetChannel(MV_SATA_ADAPTER *pMvSataAdapter, MV_U8 channel)
{
	IAL_ADAPTER_T   *pAdapter = (IAL_ADAPTER_T *)pMvSataAdapter->IALData;
	PDevice	pdev = &(pAdapter->VDevices[channel]).u.disk;

	mvSataDisableChannelDma(pMvSataAdapter, channel);
	/* Flush pending commands */
	mvSataFlushDmaQueue (pMvSataAdapter, channel, MV_FLUSH_TYPE_CALLBACK);

	if (++pdev->resetCount > 3) {
		hpt_printk("RR182x [%d,%d]: Reset more than 3 times, disconnect it\n", 
			pMvSataAdapter->adapterId, channel);		
			hptmv_free_channel(pAdapter, channel);
			return FAILED;
	}
		
	/* Software reset channel */
	if (mvStorageDevATASoftResetDevice(pMvSataAdapter, channel) == MV_FALSE)
	{
		hpt_printk("RR182x [%d,%d]: failed to perform Software reset\n",
				 pMvSataAdapter->adapterId, channel);
		hptmv_free_channel(pAdapter, channel);
		return FAILED;
	}
	
	/* Hardware reset channel */
	if (mvSataChannelHardReset(pMvSataAdapter, channel)== MV_FALSE)
	{
		hpt_printk("RR182x [%d,%d] Failed to Hard reset the SATA channel\n",
				 pMvSataAdapter->adapterId, channel);
		hptmv_free_channel(pAdapter, channel);
		return FAILED;
	}

	if (mvSataIsStorageDeviceConnected(pMvSataAdapter, channel) == MV_FALSE)
	{
		 hpt_printk("RR182x [%d,%d] Failed to Connect Device\n",
				 pMvSataAdapter->adapterId, channel);
		hptmv_free_channel(pAdapter, channel);
		return FAILED;
	}else
	{
		hpt_printk("channel %d: perform recalibrate command", channel);
		if (!mvStorageDevATAExecuteNonUDMACommand(pMvSataAdapter, channel,
												MV_NON_UDMA_PROTOCOL_NON_DATA,
												MV_FALSE,
												NULL,	 /* pBuffer*/
												0,		 /* count  */
												0,		/*features*/
												/* sectorCount */
												0,
												0,	/* lbaLow */
												0,	/* lbaMid */
												/* lbaHigh */
												0,
												0,		/* device */
												/* command */
												0x10))
			hpt_printk("channel %d: recalibrate failed", channel);

		/* Set transfer mode */
		if((mvStorageDevATASetFeatures(pMvSataAdapter, channel,
						MV_ATA_SET_FEATURES_TRANSFER,
						MV_ATA_TRANSFER_PIO_SLOW, 0, 0, 0) == MV_FALSE) || 
			(mvStorageDevATASetFeatures(pMvSataAdapter, channel,
						MV_ATA_SET_FEATURES_TRANSFER,
						pAdapter->mvChannel[channel].maxPioModeSupported, 0, 0, 0) == MV_FALSE) ||
			(mvStorageDevATASetFeatures(pMvSataAdapter, channel,
						MV_ATA_SET_FEATURES_TRANSFER,
						pAdapter->mvChannel[channel].maxUltraDmaModeSupported, 0, 0, 0) == MV_FALSE) )
		{
			hpt_printk("channel %d: Set Features failed", channel);
			hptmv_free_channel(pAdapter, channel);
			return FAILED;
		}
		/* Enable EDMA */
		if (mvSataEnableChannelDma(pMvSataAdapter, channel) == MV_FALSE)
		{
			hpt_printk("Failed to enable DMA, channel=%d", channel);
			hptmv_free_channel(pAdapter, channel);
			return FAILED;
		}
	}
	return SUCCESS;
}

int fResetActiveCommands(PVBus _vbus_p)
{
	MV_SATA_ADAPTER *pMvSataAdapter = &((IAL_ADAPTER_T *)_vbus_p->OsExt)->mvSataAdapter;
	MV_U8 channel;
	for (channel=0;channel< MV_SATA_CHANNELS_NUM;channel++) {
		if (pMvSataAdapter->sataChannel[channel] && pMvSataAdapter->sataChannel[channel]->outstandingCommands) 
			if (MvSataResetChannel(pMvSataAdapter, channel) == FAILED)
				failDevice(&((IAL_ADAPTER_T *)_vbus_p->OsExt)->VDevices[channel]);
	}
	return SUCCESS;
}

void HPTLIBAPI fCompleteAllCommandsSynchronously(PVBus _vbus_p)
{
	UINT cont;
	ULONG ticks = 0;
	MV_U8 channel;
	MV_SATA_ADAPTER *pMvSataAdapter = &((IAL_ADAPTER_T *)_vbus_p->OsExt)->mvSataAdapter;
	MV_SATA_CHANNEL *pMvSataChannel;

	do {
check_cmds:
		cont = 0;
		CheckPendingCall(_VBUS_P0);
#ifdef _RAID5N_
		dataxfer_poll();
		xor_poll();
#endif
		for (channel=0;channel< MV_SATA_CHANNELS_NUM;channel++) {
			pMvSataChannel = pMvSataAdapter->sataChannel[channel];
			if (pMvSataChannel && pMvSataChannel->outstandingCommands) 
			{
				while (pMvSataChannel->outstandingCommands) {
					if (!mvSataInterruptServiceRoutine(pMvSataAdapter)) {
						mvMicroSecondsDelay(1000);
						if (ticks++ > 3000) {
							if(MvSataResetChannel(pMvSataAdapter,channel) == FAILED)
								failDevice(&((IAL_ADAPTER_T *)_vbus_p->OsExt)->VDevices[channel]);
							goto check_cmds;
						}
					}
					else 
						ticks = 0;
				}
				cont = 1;
			}
		}
	} while (cont);
}

int fMvResetBus(PVBus _vbus_p)
{
	int rtn = SUCCESS;

	KdPrint(("fMvResetBus(%p)", _vbus_p));

	/* some commands may already finished. */
	CheckPendingCall(_VBUS_P0);

	fResetActiveCommands(_vbus_p);
	/* 
	 * the other pending commands may still be finished successfully.
	 */
	fCompleteAllCommandsSynchronously(_vbus_p);

	/* Now there should be no pending commands. No more action needed. */
	CheckIdleCall(_VBUS_P0);

	KdPrint(("fMvResetBus() done"));
	
	return rtn;
}

void FlushAdapter(IAL_ADAPTER_T *pAdapter)
{
	int i;

	MV_PRINTD("RR182x: flush all devices\n");
	
	/* flush all devices */
	for (i=0; i<MAX_VDEVICE_PER_VBUS; i++) {
		PVDevice pVDev = pAdapter->VBus.pVDevice[i];
		if(pVDev) fFlushVDev(pVDev);
	}
}

void device_change(IAL_ADAPTER_T *pAdapter , MV_U8 channelIndex, int plugged)
{
	PVDevice pVDev;
	MV_SATA_ADAPTER  *pMvSataAdapter = &pAdapter->mvSataAdapter;
	MV_SATA_CHANNEL  *pMvSataChannel = pMvSataAdapter->sataChannel[channelIndex];
	DECLARE_BUFFER(PArrayDescript, pArrayDesc);
	
	if (!pMvSataChannel) return;

	if (plugged)
	{
		pVDev = &(pAdapter->VDevices[channelIndex]);
		init_vdev_params(pAdapter, channelIndex);

		pVDev->VDeviceType = pVDev->u.disk.df_atapi? VD_ATAPI : 
			pVDev->u.disk.df_removable_drive? VD_REMOVABLE : VD_SINGLE_DISK;

		pVDev->VDeviceCapacity = pVDev->u.disk.dDeRealCapacity - SAVE_FOR_RAID_INFO;
		pVDev->pfnSendCommand = pfnSendCommand[pVDev->VDeviceType];
		pVDev->pfnDeviceFailed = pfnDeviceFailed[pVDev->VDeviceType];
		pVDev->vf_online = 1;

#ifdef SUPPORT_ARRAY
		if(pVDev->pParent) 
		{
			int iMember;
			for(iMember = 0; iMember < 	pVDev->pParent->u.array.bArnMember; iMember++)
				if((PVDevice)pVDev->pParent->u.array.pMember[iMember] == pVDev)
					pVDev->pParent->u.array.pMember[iMember] = NULL;
			pVDev->pParent = NULL;
		}
#endif
		fNotifyGUI(ET_DEVICE_PLUGGED,pVDev);
		fDeReadWrite(&pVDev->u.disk, pVDev->VDeviceCapacity, IDE_COMMAND_READ, pArrayDesc);
		if((pArrayDesc->Signature!=HPT_ARRAY_V3)
			|| ((pArrayDesc->Signature==HPT_ARRAY_V3) && (pArrayDesc->dArStamp == 0)))
			fCheckBootable(pVDev);
		RegisterVDevice(pVDev);
		if (pAdapter->beeping) {
			pAdapter->beeping = 0;
			BeepOff(pAdapter->mvSataAdapter.adapterIoBaseAddress);
		}
	}
	else
	{
		pVDev  = &(pAdapter->VDevices[channelIndex]);
		failDevice(pVDev);
	}
}

int HPTLIBAPI fDeSetTCQ(PDevice pDev, int enable, int depth)
{
	MV_SATA_CHANNEL *pSataChannel = pDev->mv;
	MV_SATA_ADAPTER *pSataAdapter = pSataChannel->mvSataAdapter;	
	MV_U8 channelIndex = pSataChannel->channelNumber;
	IAL_ADAPTER_T *pAdapter = pSataAdapter->IALData;
	MV_CHANNEL		*channelInfo = &(pAdapter->mvChannel[channelIndex]);
	int dmaActive = pSataChannel->queueCommandsEnabled;
	int ret = 0;

	if (dmaActive) {
		mvSataDisableChannelDma(pSataAdapter, channelIndex);
		mvSataFlushDmaQueue(pSataAdapter,channelIndex,MV_FLUSH_TYPE_CALLBACK);
	}

	if (enable) {
		if (pSataChannel->queuedDMA == MV_EDMA_MODE_NOT_QUEUED &&
			(pSataChannel->identifyDevice[IDEN_SUPPORTED_COMMANDS2] & le16_to_cpu(0x2))) {
			UCHAR depth = (le16_to_cpu(pSataChannel->identifyDevice[IDEN_QUEUE_DEPTH]) & 0x1f) + 1;
			channelInfo->queueDepth = (depth==32)? 31 : depth;
			mvSataConfigEdmaMode(pSataAdapter, channelIndex, MV_EDMA_MODE_QUEUED, depth);
			ret = 1;
		}
	}
	else
	{
		if (pSataChannel->queuedDMA != MV_EDMA_MODE_NOT_QUEUED) {
			channelInfo->queueDepth = 2;
			mvSataConfigEdmaMode(pSataAdapter, channelIndex, MV_EDMA_MODE_NOT_QUEUED, 0);
			ret = 1;
		}
	}

	if (dmaActive)									
		mvSataEnableChannelDma(pSataAdapter,channelIndex);
	return ret;
}

int HPTLIBAPI fDeSetNCQ(PDevice pDev, int enable, int depth)
{
	return 0;
}

int HPTLIBAPI fDeSetWriteCache(PDevice pDev, int enable)
{
	MV_SATA_CHANNEL *pSataChannel = pDev->mv;
	MV_SATA_ADAPTER *pSataAdapter = pSataChannel->mvSataAdapter;	
	MV_U8 channelIndex = pSataChannel->channelNumber;
	IAL_ADAPTER_T *pAdapter = pSataAdapter->IALData;
	MV_CHANNEL		*channelInfo = &(pAdapter->mvChannel[channelIndex]);
	int dmaActive = pSataChannel->queueCommandsEnabled;
	int ret = 0;

	if (dmaActive) {
		mvSataDisableChannelDma(pSataAdapter, channelIndex);
		mvSataFlushDmaQueue(pSataAdapter,channelIndex,MV_FLUSH_TYPE_CALLBACK);
	}
	
	if ((pSataChannel->identifyDevice[82] & le16_to_cpu(0x20))) {
		if (enable) {
			if (mvStorageDevATASetFeatures(pSataAdapter, channelIndex,
				MV_ATA_SET_FEATURES_ENABLE_WCACHE, 0, 0, 0, 0))
			{
				channelInfo->writeCacheEnabled = MV_TRUE;
				ret = 1;
			}
		}
		else {
			if (mvStorageDevATASetFeatures(pSataAdapter, channelIndex,
				MV_ATA_SET_FEATURES_DISABLE_WCACHE, 0, 0, 0, 0))
			{
				channelInfo->writeCacheEnabled = MV_FALSE;
				ret = 1;
			}
		}
	}

	if (dmaActive)									
		mvSataEnableChannelDma(pSataAdapter,channelIndex);
	return ret;
}

int HPTLIBAPI fDeSetReadAhead(PDevice pDev, int enable)
{
	MV_SATA_CHANNEL *pSataChannel = pDev->mv;
	MV_SATA_ADAPTER *pSataAdapter = pSataChannel->mvSataAdapter;	
	MV_U8 channelIndex = pSataChannel->channelNumber;
	IAL_ADAPTER_T *pAdapter = pSataAdapter->IALData;
	MV_CHANNEL		*channelInfo = &(pAdapter->mvChannel[channelIndex]);
	int dmaActive = pSataChannel->queueCommandsEnabled;
	int ret = 0;

	if (dmaActive) {
		mvSataDisableChannelDma(pSataAdapter, channelIndex);
		mvSataFlushDmaQueue(pSataAdapter,channelIndex,MV_FLUSH_TYPE_CALLBACK);
	}
	
	if ((pSataChannel->identifyDevice[82] & le16_to_cpu(0x40))) {
		if (enable) {
			if (mvStorageDevATASetFeatures(pSataAdapter, channelIndex,
				MV_ATA_SET_FEATURES_ENABLE_RLA, 0, 0, 0, 0))
			{
				channelInfo->readAheadEnabled = MV_TRUE;
				ret = 1;
			}
		}
		else {
			if (mvStorageDevATASetFeatures(pSataAdapter, channelIndex,
				MV_ATA_SET_FEATURES_DISABLE_RLA, 0, 0, 0, 0))
			{
				channelInfo->readAheadEnabled = MV_FALSE;
				ret = 1;
			}
		}
	}

	if (dmaActive)									
		mvSataEnableChannelDma(pSataAdapter,channelIndex);
	return ret;
}

static MV_BOOLEAN HPTLIBAPI PassthroughCommandCompletionCB(
								MV_SATA_ADAPTER *pSataAdapter,
								MV_U8 channelIndex,
								MV_COMPLETION_TYPE comp_type,
								MV_VOID_PTR commandId,
								MV_U16 responseFlags,
								MV_U32 timeStamp,
								MV_STORAGE_DEVICE_REGISTERS *registerStruct)
{
	PCommand pCmd;
	PVDevice pVDev;
	PVBus _vbus_p;
		
	if (commandId == NULL)
	{
		KdPrint(("Unable to handle CommandId = NULL, adapterId=%d,"
				 " channel=%d \n", pSataAdapter->adapterId, channelIndex));
		return MV_FALSE;
	}

	pCmd = (PCommand)commandId;
	KdPrint(("PassthroughCommandCompletionCB(%p, %d)", pCmd, comp_type));

	pVDev = pCmd->pVDevice;
	_vbus_p = pVDev->pVBus;

	/* copy back register values */
	pCmd->uCmd.Passthrough.bCommandReg = registerStruct->statusRegister;
	pCmd->uCmd.Passthrough.bLbaLowReg = (UCHAR)registerStruct->lbaLowRegister;
	pCmd->uCmd.Passthrough.bLbaMidReg = (UCHAR)registerStruct->lbaMidRegister;
	pCmd->uCmd.Passthrough.bLbaHighReg = (UCHAR)registerStruct->lbaHighRegister;
	pCmd->uCmd.Passthrough.bDriveHeadReg = registerStruct->deviceRegister;
	pCmd->uCmd.Passthrough.bFeaturesReg = registerStruct->errorRegister;
	pCmd->uCmd.Passthrough.bSectorCountReg = (UCHAR)registerStruct->sectorCountRegister;

	switch (comp_type) {
	case MV_COMPLETION_TYPE_NORMAL:
		pCmd->Result = RETURN_SUCCESS;
		break;
	default:
		pCmd->Result = RETURN_IDE_ERROR;
	}

	CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
	return MV_TRUE;       
}

MV_BOOLEAN mvAdapterPassthroughCommand(
								MV_SATA_ADAPTER *pSataAdapter, 
								MV_U8 channelIndex,
								PCommand pCmd
								)
{
	PVDevice pVDev = pCmd->pVDevice;
	_VBUS_INST(pVDev->pVBus)
		
	MV_QUEUE_COMMAND_INFO commandParams;
	MV_QUEUE_COMMAND_RESULT result;    

	memset(&commandParams, 0, sizeof(commandParams));

	commandParams.type = MV_QUEUED_COMMAND_TYPE_NONE_UDMA;
	commandParams.commandParams.NoneUdmaCommand.isEXT = MV_FALSE;

	if (pCmd->cf_data_in)
		commandParams.commandParams.NoneUdmaCommand.protocolType = MV_NON_UDMA_PROTOCOL_PIO_DATA_IN;
	else if (pCmd->cf_data_out)
		commandParams.commandParams.NoneUdmaCommand.protocolType = MV_NON_UDMA_PROTOCOL_PIO_DATA_OUT;
	else
		commandParams.commandParams.NoneUdmaCommand.protocolType = MV_NON_UDMA_PROTOCOL_NON_DATA;

/*	commandParams.commandParams.NoneUdmaCommand.isSgPtr = MV_FALSE; */
	commandParams.commandParams.NoneUdmaCommand.bufPtr = (USHORT *)pCmd->uCmd.Passthrough.pDataBuffer;
	commandParams.commandParams.NoneUdmaCommand.count = (UINT)pCmd->uCmd.Passthrough.nSectors<<8; /* WORDs */
	commandParams.commandParams.NoneUdmaCommand.features = pCmd->uCmd.Passthrough.bFeaturesReg;
	commandParams.commandParams.NoneUdmaCommand.sectorCount = pCmd->uCmd.Passthrough.bSectorCountReg;
	commandParams.commandParams.NoneUdmaCommand.lbaLow = pCmd->uCmd.Passthrough.bLbaLowReg;
	commandParams.commandParams.NoneUdmaCommand.lbaMid = pCmd->uCmd.Passthrough.bLbaMidReg;
	commandParams.commandParams.NoneUdmaCommand.lbaHigh = pCmd->uCmd.Passthrough.bLbaHighReg;
	commandParams.commandParams.NoneUdmaCommand.device = (MV_U8)0x00000040;
	commandParams.commandParams.NoneUdmaCommand.command = pCmd->uCmd.Passthrough.bCommandReg;
	commandParams.commandParams.NoneUdmaCommand.callBack = PassthroughCommandCompletionCB;
	commandParams.commandParams.NoneUdmaCommand.commandId = pCmd;

	result = mvSataQueueCommand(pSataAdapter, channelIndex, &commandParams);

	if (result != MV_QUEUE_COMMAND_RESULT_OK) {
		KdPrint(("mvSataQueueCommand failed result(%d)", result));
		pCmd->Result = RETURN_IDE_ERROR;
		CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
		return MV_FALSE;
	}

	return MV_TRUE;
}
