#!/bin/sh

#check to see if it is really a src dir
if [ ! -d ${KERNELDIR}/Documentation ] ; then
	echo "The target directory ${KERNELDIR} is not a full kernel source tree."
	exit 1
fi

echo "The target kernel version is ${KERNEL_VER}"

#echo "Make directory to hold the open source driver code"
rm -rf ${KERNELDIR}/drivers/scsi/rr182x
mkdir ${KERNELDIR}/drivers/scsi/rr182x
cp *.* ${KERNELDIR}/drivers/scsi/rr182x/

case "$KERNEL_VER" in 
	2.4 )
	cd $KERNELDIR/drivers/scsi/rr182x
	echo "#
# RR182x driver Makefile for linux kernel 2.4
# Copyright (C) 2003-2004 HighPoint Technologies, Inc.
# All Rights Reserved.
#

DRIVER_VERSION = ${DRIVER_VERSION}

ifeq (\$(CONFIG_X86_64),y)
HPT_RAID_O = raid-x86_64.obj
else
HPT_RAID_O = raid.obj
endif

O_TARGET := hptmv.o

obj-m := \$(O_TARGET)
obj-y := hpt.o \$(HPT_RAID_O)

EXTRA_CFLAGS += -DDRIVER_VERSION=\"\\\"\$(DRIVER_VERSION)\\\"\" -DLIST_H_INCLUDED -I\$(TOPDIR)/drivers/scsi -DLINUX -D_LINUX_ -D__KERNEL__=1

include \$(TOPDIR)/Rules.make
" > Makefile
	rm -f Makefile.* *.o *.sh *~
	grep "CONFIG_SCSI_RR182X" ../Config.in >> /dev/null
	if test $? -eq 1 ; then
		LINENUM=$(grep -n "SCSI\ low-level" ../Config.in | cut -f1 -d:)
		let LINENUM=$LINENUM+1
		mv ../Config.in ../Config.in.bak
		sed -e "
${LINENUM}a\\
dep_tristate \'HPT RocketRAID 182x support\' CONFIG_SCSI_RR182X \$CONFIG_SCSI \$CONFIG_PCI" ../Config.in.bak > ../Config.in
	fi
	grep "CONFIG_SCSI_RR182X" ../Makefile >> /dev/null
	if test $? -eq 1 ; then
		LINENUM=$(grep -n "obj-\$(CONFIG_SCSI)" ../Makefile | cut -f1 -d:)
		let LINENUM=$LINENUM-2
		let LINENUM2=$LINENUM+3
		mv ../Makefile ../Makefile.bak
		sed -e "
${LINENUM}a\\
subdir-\$(CONFIG_SCSI_RR182X)	+= rr182x
${LINENUM2}a\\
ifeq (\$(CONFIG_SCSI_RR182X),y)
${LINENUM2}a\\
  obj-\$(CONFIG_SCSI_RR182X)	+= rr182x/hptmv.o
${LINENUM2}a\\
endif
" ../Makefile.bak > ../Makefile
	fi
	grep "CONFIG_SCSI_RR182X" ../../../Documentation/Configure.help >> /dev/null
	if test $? -eq 1 ; then
		LINENUM=$(grep -n "#\ A\ couple\ of" ../../../Documentation/Configure.help | cut -f1 -d:)
		let LINENUM=$LINENUM-2
		mv ../../../Documentation/Configure.help ../../../Documentation/Configure.help.bak
		sed "
${LINENUM}a\\

${LINENUM}a\\
HPT RocketRAID 182x support
${LINENUM}a\\
CONFIG_SCSI_RR182X
${LINENUM}a\\
  This is support for HighPoint RocketRAID 182x RAID
${LINENUM}a\\
  Host Adapter
${LINENUM}a\\

${LINENUM}a\\
  This also could be compiled as module called hptmv.o.
" ../../../Documentation/Configure.help.bak > ../../../Documentation/Configure.help
	fi
	if [ -d ../../../arch/i386 ];then
		grep "CONFIG_SCSI_RR182X" ../../../arch/i386/defconfig >> /dev/null
		if test $? -eq 1 ; then
			LINENUM=$(grep -n "#\ SCSI\ low-level" ../../../arch/i386/defconfig | cut -f1 -d:)
			let LINENUM=$LINENUM+1
			mv ../../../arch/i386/defconfig ../../../arch/i386/defconfig.bak
			sed "
${LINENUM}a\\
CONFIG_SCSI_RR182X=y
" ../../../arch/i386/defconfig.bak > ../../../arch/i386/defconfig
		fi
	fi
	if [ -d ../../../arch/x86_64 ]; then
		grep "CONFIG_SCSI_RR182X" ../../../arch/x86_64/defconfig >> /dev/null
		if test $? -eq 1 ; then
			LINENUM=$(grep -n "#\ SCSI\ low-level" ../../../arch/i386/defconfig | cut -f1 -d:)
			let LINENUM=$LINENUM+1
			mv ../../../arch/x86_64/defconfig ../../../arch/x86_64/defconfig.bak
			sed "
${LINENUM}a\\
CONFIG_SCSI_RR182X=y
" ../../../arch/x86_64/defconfig.bak > ../../../arch/x86_64/defconfig
		fi
	fi
	if [ -f ../../../.config ]; then
		grep "CONFIG_SCSI_RR182X" ../../../.config >> /dev/null
		if test $? -eq 1 ; then
			LINENUM=$(grep -n "#\ SCSI\ low-level" ../../../.config | cut -f1 -d:)
			let LINENUM=$LINENUM+1
			mv ../../../.config ../../../.config.bak
			sed "
${LINENUM}a\\
CONFIG_SCSI_RR182X=y
" ../../../.config.bak > ../../../.config
		fi
	fi
	;;
	2.6 )
	cd $KERNELDIR/drivers/scsi/rr182x
	echo "#
# RR182x driver Makefile for linux kernel 2.6
# Copyright (C) 2003-2004 HighPoint Technologies, Inc.
# All Rights Reserved.
#

DRIVER_VERSION = ${DRIVER_VERSION}

ifeq (\$(CONFIG_X86_64),y)
HPT_RAID_O = raid-x86_64.obj
else
HPT_RAID_O = raid.obj
endif

obj-\$(CONFIG_SCSI_RR182X) := hptmv.o

hptmv-objs := hpt.o \$(HPT_RAID_O)

EXTRA_CFLAGS += -DDRIVER_VERSION=\"\\\"\$(DRIVER_VERSION)\\\"\" -DLIST_H_INCLUDED -I\$(TOPDIR)/drivers/scsi -DLINUX -D_LINUX_ -D__LINUX__=1
" > Makefile
	rm -f Makefile.* *.o *.sh *~
	grep -n "SCSI_RR182X" ../Kconfig >> /dev/null
	if test $? -eq 1 ; then
		mv ../Kconfig ../Kconfig.bak
		LINENUM=$(grep -n "SCSI\ low-level" ../Kconfig.bak | cut -f1 -d:)
		let LINENUM=$LINENUM+1
		sed "
${LINENUM}a\\

${LINENUM}a\\
config SCSI_RR182X
${LINENUM}a\\
	tristate \"HPT RocketRAID 182x support\"
${LINENUM}a\\
	depends on PCI && SCSI
${LINENUM}a\\
	help
${LINENUM}a\\
	This is support for HighPoint RocketRAID 182x RAID
${LINENUM}a\\
	Host Adapter
${LINENUM}a\\

${LINENUM}a\\
	You can compile it as a module called hptmve.
" ../Kconfig.bak > ../Kconfig
	fi
	grep -n "CONFIG_SCSI_RR182X" ../Makefile >> /dev/null
	if test $? -eq 1 ; then
		mv ../Makefile ../Makefile.bak
		LINENUM=$(grep -n "obj-\$(CONFIG_SCSI_FC_ATTRS)" ../Makefile.bak | cut -f1 -d:)
		let LINENUM=$LINENUM+2
		sed "
${LINENUM}a\\
obj-\$(CONFIG_SCSI_RR182X)  += rr182x/
" ../Makefile.bak > ../Makefile
	fi
	if [ -d ../../../arch/i386 ];then
		grep "CONFIG_SCSI_RR182X" ../../../arch/i386/defconfig >> /dev/null
		if test $? -eq 1 ; then
			LINENUM=$(grep -n "#\ SCSI\ low-level" ../../../arch/i386/defconfig | cut -f1 -d:)
			let LINENUM=$LINENUM+1
			mv ../../../arch/i386/defconfig ../../../arch/i386/defconfig.bak
			sed "
${LINENUM}a\\
CONFIG_SCSI_RR182X=y
" ../../../arch/i386/defconfig.bak > ../../../arch/i386/defconfig
		fi
	fi
	if [ -d ../../arch/x86_64 ];then
		grep "CONFIG_SCSI_RR182X" ../../../arch/x86_64/defconfig >> /dev/null
		if test $? -eq 1 ; then
			LINENUM=$(grep -n "#\ SCSI\ low-level" ../../../arch/x86_64/defconfig | cut -f1 -d:)
			let LINENUM=$LINENUM+1
			mv ../../../arch/x86_64/defconfig ../../../arch/x86_64/defconfig.bak
			sed "
${LINENUM}a\\
CONFIG_SCSI_RR182X=y
" ../../../arch/x86_64/defconfig.bak > ../../../arch/x86_64/defconfig
		fi
	fi
	if [ -f ../../../.config ]; then
		grep "CONFIG_SCSI_RR182X" ../../../.config >> /dev/null
		if test $? -eq 1 ; then
			LINENUM=$(grep -n "#\ SCSI\ low-level" ../../../.config | cut -f1 -d:)
			let LINENUM=$LINENUM+1
			mv ../../../.config ../../../.config.bak
			sed "
${LINENUM}a\\
CONFIG_SCSI_RR182X=y
" ../../../.config.bak > ../../../.config
		fi
	fi
	;;
esac
echo "Kernel source has been patched successfully."
exit 0
